// MG4interface.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2017 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Wrappers for the MadGraph F77 matrix elements

#include "Vincia/MG4interface.h"

namespace Vincia {

//==========================================================================

  // Wrappers for common blocks, ported from coupl.inc
  
  extern "C" {
    
    extern struct {
      double          g;        
      complex         gg[2];        
    } coupl_qcd_;

    extern struct {
      double          hmass, wmass, zmass, amass, tmass, bmass, lmass, 
        cmass, dmass, umass, smass;
    } coupl_mass_;
    
    extern struct {
      double          hwidth, wwidth, zwidth, twidth, lwidth, awidth;
    } coupl_width_;
    
    extern struct {
      complex         gal[2], gad[2], gau[2], gwf[2], gzn[2], gzl[2], gzd[2], gzu[2];
      double          gw, gwwa, gwwz;      
    } coupl_gauge_;
    
    extern struct {
      complex gwfud[2],  gwfcd[2], gwftd[2],
        gwfus[2],  gwfcs[2], gwfts[2],
        gwfub[2],  gwfcb[2], gwftb[2];
    } coupl_ckm_;
    
    // Uncomment for MadGraph with diagonal CKM:
    /*
    extern struct {
      complex gwfc[2],  gwfs[2], gwfm[2];    
    } coupl_ckm_;
    */

    extern struct {
      complex gwwh, gzzh, gwwhh, gzzhh, ghhh, ghhhh;
    } coupl_scal_;
    
    extern struct {
      complex ghtop[2], ghbot[2], ghcha[2], ghstr[2], ghtau[2];
    } coupl_yuk_;

    //Add coupling of higgs to gluons for PS/ME comparison
    extern struct {
      complex gh[2];
    } coupl_heft_;
      
    // Wrappers for MG decay matrix elements
    
    // MEs for IF helicitiy ants checks
    // For QQemitIF and QXsplitIF with ++/-- QQ
    extern void sde2de_(int *cMode, int *ICOL, double P1[3][4], int HEL[4], double ANS[1]);
    extern void sdxe2dxe_(int *cMode, int *ICOL, double P1[3][4], int HEL[4], double ANS[1]);
    extern void sde2dge_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void sdxe2dxge_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void sge2ddxe_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void sge2dxde_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    // For GGemitIF and GXconvIF and XGsplitIF with +-/-+ GG
    extern void sgta2gta_(int *cMode, int *ICOL, double P1[3][4], int HEL[4], double ANS[1]);
    extern void sgta2ggta_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void sgta2dxdta_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void sgta2ddxta_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void sdta2gdta_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void sdxta2gdxta_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    // For QQemitIF and QXsplitIF with +-/-+ QQ
    extern void sbta2bta_(int *cMode, int *ICOL, double P1[3][4], int HEL[4], double ANS[1]);
    extern void sbxta2bxta_(int *cMode, int *ICOL, double P1[3][4], int HEL[4], double ANS[1]);
    extern void sbta2bgta_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void sbxta2bxgta_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void sgta2bbxta_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void sgta2bxbta_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    
    // Z -> jets

    // Born
    extern void sz2ddx_(int *cMode, int *ICOL, double P1[3][4], int HEL[4], double ANS[1]);

    // 1st Order
    extern void sz2dgdx_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);

    // 2nd Order
    extern void sz2dggdx_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void sz2duxudx_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void sz2dsxsdx_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void sz2ddxddx_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);

    // 3rd Order
    extern void sz2dgggdx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sz2duxugdx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sz2dsxsgdx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sz2ddxdgdx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);

    // 4th Order
    extern void sz2dggggdx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sz2duxuggdx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sz2dsxsggdx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sz2ddxdggdx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sz2dsxsuxudx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sz2dsxsbxbdx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sz2duxuuxudx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sz2dsxssxsdx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sz2ddxddxddx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    
    // W -> jets
    
    // Born
    extern void swx2udx_(int *cMode, int *ICOL, double P1[3][4], int HEL[4], double ANS[1]);

    // 1st Order
    extern void swx2ugdx_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);

    // 2nd Order
    extern void swx2uggdx_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void swx2uuxudx_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void swx2udxddx_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void swx2usxsdx_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);

    // 3rd Order
    extern void swx2ugggdx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void swx2uuxugdx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void swx2udxdgdx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void swx2usxsgdx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);

    // 4th Order
    extern void swx2uggggdx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void swx2uuxuggdx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void swx2udxdggdx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void swx2usxsggdx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void swx2uuxuuxudx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void swx2udxddxddx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void swx2uuxudxddx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void swx2uuxusxsdx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void swx2udxdsxsdx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void swx2usxssxsdx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void swx2usxscxcdx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    
    // H -> jets

    // Born
    extern void sh2gg_(int *cMode, int *ICOL, double P1[3][4], int HEL[4], double ANS[1]);
    extern void sh2bbx_(int *cMode, int *ICOL, double P1[3][4], int HEL[4], double ANS[1]);

    // 1st Order
    extern void sh2ggg_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void sh2bgbx_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);

    // 2nd Order
    extern void sh2gggg_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void sh2bggbx_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void sh2bcxcbx_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void sh2bbxbbx_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);

    // 3rd Order
    extern void sh2ggggg_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sh2bgggbx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sh2bcxcgbx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sh2bbxbgbx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);

    // 4th Order
    extern void sh2gggggg_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sh2bggggbx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sh2bcxcggbx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sh2bbxbggbx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sh2bcxcsxsbx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sh2bcxccxcbx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sh2bbxbbxbbx_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    
    // pp -> Z + jets

    // Born
    extern void sddx2z_(int *cMode, int *ICOL, double P1[3][4], int HEL[12], double ANS[1]);

    // 1st Order
    extern void sddx2gz_(int *cMode, int *ICOL, double P1[4][4], int HEL[12], double ANS[1]);
    extern void sgd2dz_(int *cMode, int *ICOL, double P1[4][4], int HEL[12], double ANS[1]);
    extern void sgdx2dxz_(int *cMode, int *ICOL, double P1[4][4], int HEL[12], double ANS[1]);

    // 2nd Order
    extern void sddx2ggz_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sgg2ddxz_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sgd2gdz_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sdd2ddz_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sdu2duz_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sds2dsz_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sddx2ddxz_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sdux2duxz_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sdsx2dsxz_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    //extern void sddx2ddxz_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sddx2uuxz_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sddx2ssxz_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);

    // 3rd Order
    extern void sddx2gggz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sgg2ddxgz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdg2dggz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdg2dddxz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdg2dssxz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdg2duuxz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdxg2dxddxz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdxg2dxssxz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdxg2dxuuxz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sddx2ddxgz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sddx2ssxgz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sddx2uuxgz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdd2ddgz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sds2dsgz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdu2dugz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    //extern void sddx2ddxgz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdsx2dsxgz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdux2duxgz_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    
    // pp -> W + jets

    // Born
    extern void sudx2wx_(int *cMode, int *ICOL, double P1[3][4], int HEL[12], double ANS[1]);

    // 1st Order
    extern void sudx2gwx_(int *cMode, int *ICOL, double P1[4][4], int HEL[12], double ANS[1]);
    extern void sug2dwx_(int *cMode, int *ICOL, double P1[4][4], int HEL[12], double ANS[1]);
    extern void sgdx2uxwx_(int *cMode, int *ICOL, double P1[4][4], int HEL[12], double ANS[1]);

    // 2nd Order
    extern void sudx2ggwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sug2gdwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sgdx2guxwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sgg2uxdwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void suux2duxwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sdxux2uxuxwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sudx2uuxwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void suu2udwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sddx2duxwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sdxdx2dxuxwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sdxu2ddxwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sdu2ddwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void scxdx2cxuxwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sucx2dcxwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void scdx2cuxwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void scu2cdwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sudx2ccxwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sccx2uxdwx_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);

    // 3rd Order
    extern void sudx2gggwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sug2ggdwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sgdx2gguxwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sgg2guxdwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sug2ccxdwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdxu2ccxgwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdxg2ccxuxwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void scxu2cxdgwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void scxg2cxduxwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void scxdx2cxuxgwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void scu2cdgwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void scg2cduxwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void scdx2cuxgwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sddx2duxgwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdg2dduxwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdu2ddgwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdxdx2dxuxgwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdxg2ddxuxwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sudx2ddxgwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sug2dddxwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sdxux2uxuxgwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sgdx2uuxuxwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sgux2uxuxdwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sudx2uuxgwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sug2duuxwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void suu2dugwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void suux2duxgwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sccx2uxdgwx_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    
    // pp -> H + jets

    // Born
    extern void sgg2h_(int *cMode, int *ICOL, double P1[3][4], int HEL[12], double ANS[1]);
    extern void sbbx2h_(int *cMode, int *ICOL, double P1[3][4], int HEL[12], double ANS[1]);

    // 1st Order
    extern void sgg2gh_(int *cMode, int *ICOL, double P1[4][4], int HEL[12], double ANS[1]);
    extern void sgb2bh_(int *cMode, int *ICOL, double P1[4][4], int HEL[12], double ANS[1]);
    extern void sgbx2bxh_(int *cMode, int *ICOL, double P1[4][4], int HEL[12], double ANS[1]);
    extern void sbbx2gh_(int *cMode, int *ICOL, double P1[4][4], int HEL[12], double ANS[1]);

    // 2nd Order
    extern void sgg2ggh_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sgg2bbxh_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sgb2bgh_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sgbx2bxgh_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sbbx2ggh_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sbb2bbh_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sbbx2bbxh_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sbxbx2bxbxh_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sbc2bch_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sbcx2bcxh_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sbxcx2bxcxh_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);
    extern void sbbx2ccxh_(int *cMode, int *ICOL, double P1[5][4], int HEL[12], double ANS[1]);

    // 3rd Order
    extern void sgg2gggh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sgg2bbxgh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sgb2bggh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sgbx2bxggh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sbbx2gggh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sbb2bbgh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sbbx2bbxgh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sbxbx2bxbxgh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sbg2bbbxh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sbxg2bxbbxh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sbc2bcgh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sbcx2bcxgh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sbxcx2bxcxgh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sbbx2ccxgh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sbg2bccxh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    extern void sbxg2bxccxh_(int *cMode, int *ICOL, double P1[6][4], int HEL[12], double ANS[1]);
    
    // QCD 2->n

    // Born
    extern void sgg2gg_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void sgg2uux_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void sug2ug_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void suux2gg_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void suux2uux_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void suu2uu_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void suux2ddx_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void sud2ud_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    extern void sudx2udx_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    
    // 1st order
    extern void sgg2ggg_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void sgg2uuxg_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void sug2ugg_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void suux2ggg_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void suux2uuxg_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void suu2uug_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void sug2uuux_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void suux2ddxg_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void sud2udg_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void sudx2udxg_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    extern void sug2uddx_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);

    // 2nd order
    extern void sgg2gggg_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sgg2uuxgg_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sug2uggg_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void suux2gggg_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void suux2uuxgg_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void suu2uugg_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sug2uuuxg_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sgg2uuxuux_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void suux2ddxgg_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sud2udgg_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sudx2udxgg_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sug2uddxg_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sgg2uuxddx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void suux2uuxuux_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void suu2uuuux_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void suux2ddxddx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void suux2uuxddx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void suu2uuddx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sud2udddx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sudx2udxddx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sudx2udxuux_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void suux2ddxssx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sudx2udxssx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    extern void sud2udssx_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);

    // 3rd order
    //extern void sgg2ggggg_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sgg2uuxggg_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void sug2ugggg_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    extern void suux2ggggg_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    
    //e+e- -> jets
    
    //Born
    extern void sexe2uux_(int *cMode, int *ICOL, double P1[4][4], int HEL[4], double ANS[1]);
    
    //1st order
    extern void sexe2ugux_(int *cMode, int *ICOL, double P1[5][4], int HEL[4], double ANS[1]);
    
    //2nd order
    extern void sexe2uggux_(int *cMode, int *ICOL, double P1[6][4], int HEL[4], double ANS[1]);
    
    //3rd order
    extern void sexe2ugggux_(int *cMode, int *ICOL, double P1[7][4], int HEL[4], double ANS[1]);
    
    //4th order
    extern void sexe2uggggux_(int *cMode, int *ICOL, double P1[8][4], int HEL[4], double ANS[1]);
  }

  //*********

  // Function to swap parameters (eg to use Z->bb matrix element for Z->uu)
  // IMPORTANT NOTE: always remember to swap back after such operations

  void MG4interface::swapZud() {
    complex tmp0 = coupl_gauge_.gzu[0];
    complex tmp1 = coupl_gauge_.gzu[1];
    coupl_gauge_.gzu[0] = coupl_gauge_.gzd[0];
    coupl_gauge_.gzu[1] = coupl_gauge_.gzd[1];
    coupl_gauge_.gzd[0] = tmp0;
    coupl_gauge_.gzd[1] = tmp1;
  }

  // Function to swap parameters (eg to use e+e-->bb matrix element for e+e-->uu)
  // IMPORTANT NOTE: always remember to swap back after such operations

  void MG4interface::swapAud() {
    complex tmp0 = coupl_gauge_.gau[0];
    complex tmp1 = coupl_gauge_.gau[1];
    coupl_gauge_.gau[0] = coupl_gauge_.gad[0];
    coupl_gauge_.gau[1] = coupl_gauge_.gad[1];
    coupl_gauge_.gad[0] = tmp0;
    coupl_gauge_.gad[1] = tmp1;
  }
  
  //*********
    
  // Function to set the coupling for W
  
  void MG4interface::setWcouplings(int id1, int id2) {
    // MEs are using gwf
    if (id1 > id2) {
      int idTmp = id1;
      id1 = id2;
      id2 = idTmp;
    }
    if (id1 == 1) {
      if (id2 == 2) {
        coupl_gauge_.gwf[0] = Vud * complex( gwf1 , 0.0 );
        coupl_gauge_.gwf[1] = Vud * complex( gwf2 , 0.0 );
      } else if (id2 == 4) {
        coupl_gauge_.gwf[0] = Vcd * complex( gwf1 , 0.0 );
        coupl_gauge_.gwf[1] = Vcd * complex( gwf2 , 0.0 );
      } else if (id2 == 6) {
        coupl_gauge_.gwf[0] = Vtd * complex( gwf1 , 0.0 );
        coupl_gauge_.gwf[1] = Vtd * complex( gwf2 , 0.0 );
      }
    } else if (id1 == 2) {
      if (id2 == 3) {
        coupl_gauge_.gwf[0] = Vus * complex( gwf1 , 0.0 );
        coupl_gauge_.gwf[1] = Vus * complex( gwf2 , 0.0 );
      } else if (id2 == 5) {
        coupl_gauge_.gwf[0] = Vub * complex( gwf1 , 0.0 );
        coupl_gauge_.gwf[1] = Vub * complex( gwf2 , 0.0 );
      }
    } else if (id1 == 3) {
      if (id2 == 4) {
        coupl_gauge_.gwf[0] = Vcs * complex( gwf1 , 0.0 );
        coupl_gauge_.gwf[1] = Vcs * complex( gwf2 , 0.0 );
      } else if (id2 == 6) {
        coupl_gauge_.gwf[0] = Vts * complex( gwf1 , 0.0 );
        coupl_gauge_.gwf[1] = Vts * complex( gwf2 , 0.0 );
      }
    } else if (id1 == 4) {
      if (id2 == 5) {
        coupl_gauge_.gwf[0] = Vcb * complex( gwf1 , 0.0 );
        coupl_gauge_.gwf[1] = Vcb * complex( gwf2 , 0.0 );
      }
    } else if (id1 == 5) {
      if (id2 == 6) {
        coupl_gauge_.gwf[0] = Vtb * complex( gwf1 , 0.0 );
        coupl_gauge_.gwf[1] = Vtb * complex( gwf2 , 0.0 );
      }
    }
  }

  //*********
  
  // Functions to set order for q qbar' q' qbar'' q'' qbar with one distinct pair
  vector<int> MG4interface::orderQQBqqbqqb(pair<int,int> posQQB, bool
    firstLast = true) {
   vector<int> order;
   order.clear(); 
   order.resize(0);
   order.push_back(0);
   if (firstLast) {
     order.push_back(posQQB.first);
     order.push_back(posQQB.first+1);
   }
        if (posQQB==make_pair(1,4)) { order.push_back(5); order.push_back(6); }
   else if (posQQB==make_pair(1,6)) { order.push_back(3); order.push_back(4); }
   else if (posQQB==make_pair(3,2)) { order.push_back(5); order.push_back(6); }
   else if (posQQB==make_pair(3,6)) { order.push_back(1); order.push_back(2); }
   else if (posQQB==make_pair(5,2)) { order.push_back(3); order.push_back(4); }
   else if (posQQB==make_pair(5,4)) { order.push_back(1); order.push_back(2); }
   order.push_back(posQQB.second-1);
   order.push_back(posQQB.second);
   if (!firstLast) {
     order.push_back(posQQB.first);
     order.push_back(posQQB.first+1);
   }
   return order;
  }
  vector<int> MG4interface::orderQQBqqbqqbQCD(pair<int,int> posQQB,
   bool firstLast = true) {
   vector<int> order;
   order.clear(); 
   order.resize(0);
   if (firstLast) {
     order.push_back(posQQB.first);
     order.push_back(posQQB.first+1);
   }
        if (posQQB==make_pair(0,3)) { order.push_back(4); order.push_back(5); }
   else if (posQQB==make_pair(0,5)) { order.push_back(2); order.push_back(3); }
   else if (posQQB==make_pair(2,1)) { order.push_back(4); order.push_back(5); }
   else if (posQQB==make_pair(2,5)) { order.push_back(0); order.push_back(1); }
   else if (posQQB==make_pair(4,1)) { order.push_back(2); order.push_back(3); }
   else if (posQQB==make_pair(4,3)) { order.push_back(0); order.push_back(1); }
   order.push_back(posQQB.second-1);
   order.push_back(posQQB.second);
   if (!firstLast) {
     order.push_back(posQQB.first);
     order.push_back(posQQB.first+1);
   }
   return order;
  }

  //*********
    
  // Functions to get order for W decays
  vector<int> MG4interface::orderWfourQuarks(vector<int> ids, bool& gSwap,
    bool& gg) {
    vector<int> order;
    order.push_back(0);
    gSwap = false;
    gg    = false;
    if (ids.size() == 5) {
      // W -> q qbar q qbar
      if (abs(ids[2]) == abs(ids[3])) {
        // q qbar' q' qbar''
        for (int i=1; i<=4; i++) order.push_back(i);
      } else {
        // Swap q' qbar'' q qbar' to q qbar' q' qbar''
        order.push_back(3);
        order.push_back(4);
        order.push_back(1);
        order.push_back(2);
      }
    } else if (ids.size() == 6) {
      // W -> q qbar q g qbar
      if (abs(ids[2]) == abs(ids[3])) {
        // q qbar' q' g qbar''
        for (int i=1; i<=5; i++) order.push_back(i);
      } else {
        // Swap q' qbar'' q g qbar' to q qbar' q' g qbar''
        order.push_back(3);
        order.push_back(5);
        order.push_back(1);
        order.push_back(4);
        order.push_back(2);
        gSwap = true;
      }
    } else if (ids.size() == 7) {
      if (ids[4] == 21 && ids[5] == 21) {
        // W -> q qbar q g g qbar
        gg = true;
        if (abs(ids[2]) == abs(ids[3])) {
          // q qbar' q' g g qbar''
          for (int i=1; i<=6; i++) order.push_back(i);
        } else {
          // Swap q' qbar'' q g g qbar' to q qbar' q' g g qbar''
          order.push_back(3);
          order.push_back(6);
          order.push_back(1);
          order.push_back(4);
          order.push_back(5);
          order.push_back(2);
          gSwap = true;
        }
      } else {
        // W -> q g qbar q g qbar
        if (abs(ids[3]) == abs(ids[4])) {
          // Swap q g1 qbar' q' g2 qbar'' to q qbar' q' g1 g2 qbar''
          order.push_back(1);
          order.push_back(3);
          order.push_back(4);
          order.push_back(2);
          order.push_back(5);
          order.push_back(6);
        } else {
          // Swap q' g1 qbar'' q g2 qbar' to q qbar' q' g2 g1 qbar''
          order.push_back(4);
          order.push_back(6);
          order.push_back(1);
          order.push_back(5);
          order.push_back(2);
          order.push_back(3);
        }
      }
    }
    return order;
  }
  vector<int> MG4interface::orderWsixQuarks(vector<int> ids, map<int,int>&
    flavours, bool& flav2ordChange) {
    flavours.clear();
    flav2ordChange = false;
    for (int i=1; i<=5; ++i) flavours[i] = 0;
    for (int i=1; i<(int)ids.size(); ++i) flavours[abs(ids[i])]++;
    int nrDiffFlavs = 0;
    for (int i=1; i<=5; ++i) if (flavours[i] > 0) nrDiffFlavs++;
    if (nrDiffFlavs == 2) {
      // If we have 1 + 5 combi: q qbar q qbar q qbar'
      //                         q' qbar q qbar q qbar
      for (int i=1; i<=5; ++i) if (flavours[i] == 1) {
        int flav = i;
        int iQ1  = -1;
        int iQ2  = -1;
        for (int j=1; j<(int)ids.size(); ++j) if (abs(ids[j]) == flav) iQ1 = j;
        iQ2 = (iQ1 <= 3 ? iQ1+3 : iQ1-3);
        if (iQ2 % 2) { // iQ2 is q, but needs to be qbar
          int iQtmp = iQ1;
          iQ1 = iQ2;
          iQ2 = iQtmp;
        }
        return orderQQBqqbqqb(make_pair(iQ1, iQ2), true);
      }
      // If we have 3 + 3 combi
      int nrDiffFlavQP = 0;
      for (int i=1; i<6; i+=2) if (abs(ids[i]) != abs(ids[i+1])) nrDiffFlavQP++;
      if (nrDiffFlavQP == 1) { // q qbar q qbar' q' qbar'
        if (abs(ids[1]) != abs(ids[2]))
          return orderQQBqqbqqb((ids[1] == ids[3] ? make_pair(3, 6) : 
            make_pair(5, 4)), true);
        if (abs(ids[3]) != abs(ids[4]))
          return orderQQBqqbqqb((ids[1] == ids[3] ? make_pair(1, 6) : 
            make_pair(5, 2)), true);
        if (abs(ids[5]) != abs(ids[6]))
          return orderQQBqqbqqb((ids[1] == ids[5] ? make_pair(1, 4) : 
            make_pair(3, 2)), true);
      }
      vector<int> order;
      order.push_back(0);
      if (abs(ids[1]) == abs(ids[4]) && abs(ids[1]) == abs(ids[5])) {
        // q qbar' q' qbar q qbar'
        order.push_back(1);
        order.push_back(4);
        order.push_back(5);
        order.push_back(2);
        order.push_back(3);
        order.push_back(6);
      } else if (abs(ids[1]) == abs(ids[3]) && abs(ids[1]) == abs(ids[6])) {
        // q qbar' q qbar' q' qbar
        order.push_back(1);
        order.push_back(6);
        order.push_back(3);
        order.push_back(2);
        order.push_back(5);
        order.push_back(4);
      } else if (abs(ids[1]) == abs(ids[4]) && abs(ids[1]) == abs(ids[6])) {
        // q' qbar q qbar' q qbar'
        order.push_back(3);
        order.push_back(2);
        order.push_back(5);
        order.push_back(4);
        order.push_back(1);
        order.push_back(6);
      }
      flav2ordChange = true;
      if (flav2ordChange) return order;
    }
    if (nrDiffFlavs == 3) {
      bool standardProd = true;
      // If emitted quark pairs have the same flavour, do standard procedure
      for (int i=1; i<=5; ++i) if (flavours[i] == 3) standardProd = false;
      if (!standardProd) { // q qbar q qbar' q' qbar''
                           // q qbar' q' qbar q qbar''
                           // q'' qbar' q' qbar q qbar
                           // q'' qbar q qbar' q' qbar
        // Find quark flavours to look for
        int flav1 = 0;
        int flav2 = 0;
        for (int i=1; i<=5; ++i) {
          if (flavours[i] == 1) flav1 = i;
          else if ( (flavours[i] % 2) && (flavours[i] > 1) ) flav2 = i;
        }
        // Find first quark
        int iQ1 = -1;
        for (int i=1; i<(int)ids.size(); ++i) if (abs(ids[i]) == flav1) {
          iQ1 = i;
          break;
        }
        // Find second quark
        int iQ2 = -1;
        for (int i=1; i<(int)ids.size(); ++i) if (abs(ids[i]) == flav2) {
          // iQ1 is q
          if ((iQ1 % 2) && !(i % 2) && (i != iQ1+1)) iQ2 = i;
          // iQ1 is qbar
          else if (!(iQ1 % 2) && (i % 2) && (i != iQ1-1)) iQ2 = i;
          if (iQ2 > 0) break;
        }
        if (iQ2 % 2) { // iQ2 is q, but needs to be qbar
          int iQtmp = iQ1;
          iQ1 = iQ2;
          iQ2 = iQtmp;
        }
        return orderQQBqqbqqb(make_pair(iQ1, iQ2), true);
      }
    }
    // Standard procedure for: q'' qbar q qbar' q' qbar'''
    //                         q' qbar q qbar q qbar''
    vector<int> flavs;
    for (int i=1; i<=5; ++i) // 2 odd appearences
      if (flavours[i] % 2) flavs.push_back(i);
    int iQ    = -1;
    int iQbar = -1;
    for (int i=1; i<(int)ids.size(); ++i) for (int j=0; j<2; j++)
      if (abs(ids[i]) == flavs[j]) ( (ids[i] > 0) ? iQ = i : iQbar = i );
    return orderQQBqqbqqb(make_pair(iQ, iQbar), true);
  }

  //*********
    
  // Function to set the Yukawa couplings and masses
  void MG4interface::setYukawa(int id1, double m1) {
    coupl_mass_.bmass   = m1;
    double yukCoupNow1  = ( ((m1>0.0) && (abs(yukCoup[abs(id1)])>0.0)) ?
      (-m1/vev) : yukCoup[abs(id1)] );
    coupl_yuk_.ghbot[0] = complex( 0.0, yukCoupNow1 );
    coupl_yuk_.ghbot[1] = complex( 0.0, yukCoupNow1 );
  }
  void MG4interface::setYukawa(int id1, int id2, double m1, double m2) {
    coupl_mass_.bmass   = m1;
    coupl_mass_.cmass   = m2;
    double yukCoupNow1  = ( ((m1>0.0) && (abs(yukCoup[abs(id1)])>0.0)) ?
      (-m1/vev) : yukCoup[abs(id1)] );
    double yukCoupNow2  = ( ((m2>0.0) && (abs(yukCoup[abs(id2)])>0.0)) ?
      (-m2/vev) : yukCoup[abs(id2)] );
    coupl_yuk_.ghbot[0] = complex( 0.0, yukCoupNow1 );
    coupl_yuk_.ghbot[1] = complex( 0.0, yukCoupNow1 );
    coupl_yuk_.ghcha[0] = complex( 0.0, yukCoupNow2 );
    coupl_yuk_.ghcha[1] = complex( 0.0, yukCoupNow2 );
  }
  void MG4interface::setYukawa(int id1, int id2, int id3,
                               double m1, double m2, double m3) {
    coupl_mass_.bmass   = m1;
    coupl_mass_.cmass   = m2;
    coupl_mass_.smass   = m3;
    double yukCoupNow1  = ( ((m1>0.0) && (abs(yukCoup[abs(id1)])>0.0)) ?
      (-m1/vev) : yukCoup[abs(id1)] );
    double yukCoupNow2  = ( ((m2>0.0) && (abs(yukCoup[abs(id2)])>0.0)) ?
      (-m2/vev) : yukCoup[abs(id2)] );
    double yukCoupNow3  = ( ((m3>0.0) && (abs(yukCoup[abs(id3)])>0.0)) ?
      (-m3/vev) : yukCoup[abs(id3)] );
    coupl_yuk_.ghbot[0] = complex( 0.0, yukCoupNow1 );
    coupl_yuk_.ghbot[1] = complex( 0.0, yukCoupNow1 );
    coupl_yuk_.ghcha[0] = complex( 0.0, yukCoupNow2 );
    coupl_yuk_.ghcha[1] = complex( 0.0, yukCoupNow2 );
    coupl_yuk_.ghstr[0] = complex( 0.0, yukCoupNow3 );
    coupl_yuk_.ghstr[1] = complex( 0.0, yukCoupNow3 );
  }
    
  // Function to set either Higgs-g for Higgs-q coupling to zero
  void MG4interface::setHinterferenceOff(bool heftIn) {
    if (heftIn) {
      // Zero heft coupling
      coupl_heft_.gh[0] = complex( 0.0, 0.0 );
      coupl_heft_.gh[1] = complex( 0.0, 0.0 );
      // Recover Yukawa couplings
      yukCoup           = yukCoupSav;
    } else {
      // Recover heft coupling
      coupl_heft_.gh[0] = complex( 0.0, heftCoupSav );
      coupl_heft_.gh[1] = complex( 0.0, heftCoupSav );
      // Zero Yukawa couplings
      for (int i=0; i<(int)yukCoup.size(); i++) yukCoup[i] = 0.0;
    }
  }

  //*********

  // Initialize MadGraph
  // Function to initialize the MadGraph coupling common blocks, required
  // by the MG matrix elements. Adapted from couplings.f in the MG 
  // Models/sm/ directory.

  bool MG4interface::init() {
    
    // Set verbose level
    verbose = settingsPtr->mode("Vincia:verbose");

    // Verbose output
    if (verbose >= 2) printOut("MG4interface::init","...");

    // Check if pointers initialized
    if (! isInitPtr) {
      printOut("MG4interface::init","Cannot initialize, pointers not set.");
      return false;      
    }
    
    // Innocent until proven guilty
    isInit = true;

    // Verbose output
    if (verbose >= 3) printOut("MG4interface::init","   setting MG masses...");

    // Set masses and widths (initialized through Pythia)
    coupl_mass_.dmass   = particleDataPtr->m0(1);
    coupl_mass_.umass   = particleDataPtr->m0(2);
    coupl_mass_.smass   = particleDataPtr->m0(3);
    coupl_mass_.cmass   = particleDataPtr->m0(4);
    coupl_mass_.bmass   = particleDataPtr->m0(5);
    coupl_mass_.tmass   = particleDataPtr->m0(6);
    coupl_mass_.lmass   = particleDataPtr->m0(15);
    coupl_mass_.amass   = particleDataPtr->m0(22);
    coupl_mass_.zmass   = particleDataPtr->m0(23);
    coupl_mass_.wmass   = particleDataPtr->m0(24);
    coupl_mass_.hmass   = particleDataPtr->m0(25);
    coupl_width_.twidth = particleDataPtr->mWidth(6);
    coupl_width_.lwidth = particleDataPtr->mWidth(15);
    coupl_width_.awidth = particleDataPtr->mWidth(22);
    coupl_width_.zwidth = particleDataPtr->mWidth(23);
    coupl_width_.wwidth = particleDataPtr->mWidth(24);
    coupl_width_.hwidth = particleDataPtr->mWidth(25);

    // Useful shorthands
    const double Fourpi = 4.0 * M_PI;
    const double Zero   = 0.0;
    // Here using alpha(mZ) (as in SLHA) rather than alpha(0)
    // Check consistency when doing full matrix elements.
    const double alpha  = settingsPtr->parm("StandardModel:alphaEMmZ");
    // Use thetabar for VFF couplings? (as in Pythia)
    // Check consistency when doing full matrix elements.
    const double sin2w  = couplingsPtr->sin2thetaW();
    const double cw     = sqrt( 1.0 - sin2w );
    const double ee2    = alpha * Fourpi;
    const double sw     = sqrt( sin2w );
    const double ee     = sqrt( ee2 );
    const double ez     = ee/(sw*cw);
    const double ey     = ee*(sw/cw);
    const double sc2    = sin2w*( 1.0 - sin2w );
    const double v      = 2.0*coupl_mass_.wmass*sw/ee;    
    //const double lambda = pow2(coupl_mass_.hmass) / (2.0 * pow2(v));
    // CKM
    Vud = couplingsPtr->VCKMgen(1,1);
    Vus = couplingsPtr->VCKMgen(1,2);
    Vub = couplingsPtr->VCKMgen(1,3);
    Vcd = couplingsPtr->VCKMgen(2,1);
    Vcs = couplingsPtr->VCKMgen(2,2);
    Vcb = couplingsPtr->VCKMgen(2,3);
    Vtd = couplingsPtr->VCKMgen(3,1);
    Vts = couplingsPtr->VCKMgen(3,2);
    Vtb = couplingsPtr->VCKMgen(3,3);

    // Verbose output
    if (verbose >= 3) printOut("MG4interface::init","   setting MG couplings...");

    // Set QCD g = 1 to keep factored out (=> alphaS = 1/sqrt(4pi))
    coupl_qcd_.g      = 1.0;
    coupl_qcd_.gg[0]  = -coupl_qcd_.g;
    coupl_qcd_.gg[1]  = -coupl_qcd_.g;

    // Set Vector boson couplings
    coupl_gauge_.gw   = ee/sw;
    coupl_gauge_.gwwa = ee;
    coupl_gauge_.gwwz = ee*cw/sw;
    
    // Set gauge & higgs boson couplings
    coupl_scal_.gwwh  = complex( ee2/sin2w*0.5*v, Zero );
    coupl_scal_.gzzh  = complex( ee2/sc2*0.5*v  , Zero );
    coupl_scal_.ghhh  = complex( -pow2(coupl_mass_.hmass)/v*3.0, Zero );
    coupl_scal_.gwwhh = complex( ee2/sin2w*0.5  , Zero );
    coupl_scal_.gzzhh = complex( ee2/sc2*0.5    , Zero );
    coupl_scal_.ghhhh = coupl_scal_.ghhh/v;

    // Set fermion-fermion-vector couplings
    coupl_gauge_.gal[0] = complex(  ee          , Zero );
    coupl_gauge_.gal[1] = complex(  ee          , Zero );
    coupl_gauge_.gau[0] = complex( -ee*2./3.    , Zero );
    coupl_gauge_.gau[1] = complex( -ee*2./3.    , Zero );
    coupl_gauge_.gad[0] = complex(  ee/3.       , Zero );
    coupl_gauge_.gad[1] = complex(  ee/3.       , Zero );

    coupl_gauge_.gzn[0] = complex( -ez*0.5                  , Zero );
    coupl_gauge_.gzn[1] = complex(  Zero                    , Zero );
    coupl_gauge_.gzl[0] = complex( -ez*(-0.5 + sin2w)       , Zero );
    coupl_gauge_.gzl[1] = complex( -ey                      , Zero );
    coupl_gauge_.gzu[0] = complex( -ez*( 0.5 - sin2w*2./3.) , Zero );
    coupl_gauge_.gzu[1] = complex(  ey*2./3.                , Zero );
    coupl_gauge_.gzd[0] = complex( -ez*(-0.5 + sin2w/3.)    , Zero );
    coupl_gauge_.gzd[1] = complex( -ey/3.                   , Zero );
    
    gwf1 = -ee/sqrt(2.*sin2w);
    gwf2 = Zero;
    coupl_gauge_.gwf[0] = complex( gwf1 , Zero );
    coupl_gauge_.gwf[1] = complex( gwf2 , Zero );

    // CKM
    for (int iLR = 0; iLR <= 1; ++iLR) {
      coupl_ckm_.gwfud[iLR] = Vud * coupl_gauge_.gwf[iLR]; 
      coupl_ckm_.gwfus[iLR] = Vus * coupl_gauge_.gwf[iLR];
      coupl_ckm_.gwfub[iLR] = Vub * coupl_gauge_.gwf[iLR];
      coupl_ckm_.gwfcd[iLR] = Vcd * coupl_gauge_.gwf[iLR]; 
      coupl_ckm_.gwfcs[iLR] = Vcs * coupl_gauge_.gwf[iLR];
      coupl_ckm_.gwfcb[iLR] = Vcb * coupl_gauge_.gwf[iLR];
      coupl_ckm_.gwftd[iLR] = Vtd * coupl_gauge_.gwf[iLR]; 
      coupl_ckm_.gwfts[iLR] = Vts * coupl_gauge_.gwf[iLR];
      coupl_ckm_.gwftb[iLR] = Vtb * coupl_gauge_.gwf[iLR];
   }

    // Set the Yukawa couplings for the different flavours
    // This is used for H production (H decay uses actual quark masses if > 0)
    // Make sure the mass is not set to zero
    vev = v;
    yukCoup.clear();
    yukCoup.resize(0);
    yukCoup.push_back(0.0);
    if (particleDataPtr->m0(1) < TINY) yukCoup.push_back(-0.005/vev);
    else yukCoup.push_back(-particleDataPtr->m0(1)/vev);
    if (particleDataPtr->m0(1) < TINY) yukCoup.push_back(-0.003/vev);
    else yukCoup.push_back(-particleDataPtr->m0(2)/vev);
    if (particleDataPtr->m0(3) < TINY) yukCoup.push_back(-0.1/vev);
    else yukCoup.push_back(-particleDataPtr->m0(3)/vev);
    if (particleDataPtr->m0(4) < TINY) yukCoup.push_back(-1.5/vev);
    else yukCoup.push_back(-particleDataPtr->m0(4)/vev);
    if (particleDataPtr->m0(5) < TINY) yukCoup.push_back(-4.8/vev);
    else yukCoup.push_back(-particleDataPtr->m0(5)/vev);
    coupl_yuk_.ghbot[0] = complex( Zero, yukCoup[5] );
    coupl_yuk_.ghbot[1] = complex( Zero, yukCoup[5] );
    coupl_yuk_.ghcha[0] = complex( Zero, yukCoup[4] );
    coupl_yuk_.ghcha[1] = complex( Zero, yukCoup[4] );
    coupl_yuk_.ghstr[0] = complex( Zero, yukCoup[3] );
    coupl_yuk_.ghstr[1] = complex( Zero, yukCoup[3] );
    yukCoupSav = yukCoup;
    
    // TODO At the moment we don't include the running
    // Include coupling of Higgs to gluons
    // NOTE value taken from MadGraph ( = alphaS/3/pi/vev(1+corr) )
    heftCoupSav = 0.000054936089955251788;
    coupl_heft_.gh[0] = complex( Zero, heftCoupSav );
    coupl_heft_.gh[1] = complex( Zero, heftCoupSav );

    // Verbose output
    if (verbose >= 3) printOut("MG4interface::init","   setting Color Depth...");

    // Set color depth
    colorDepth = (settingsPtr->flag("Vincia:matchingFullColor")) ? 1 : 0;

    return true;
  }
  
  //*********

  // Wrapper to set values in the MG common blocks

  void MG4interface::setParm(string txt, complex val) {

    // Check if initialized
    if (! isInit) {
      printOut("MG4interface::setParm","MG4interface not initialized.");
    }
    // Convert to all lower-case
    for (unsigned int i=0;i<txt.length();++i) txt[i]=tolower(txt[i]);
    if (txt == "zmass" || txt == "mz") coupl_mass_.zmass   = val.real();
    else if (txt == "wmass" || txt == "mw") coupl_mass_.wmass   = val.real();
    else if (txt == "bmass" || txt == "mb") coupl_mass_.bmass   = val.real();
    else if (txt == "hmass" || txt == "mh") coupl_mass_.hmass   = val.real();
    // A (photon) couplings
    else if (txt == "gau0") coupl_gauge_.gau[0]=val;
    else if (txt == "gau1") coupl_gauge_.gau[1]=val;
    else if (txt == "gad0") coupl_gauge_.gad[0]=val;
    else if (txt == "gad1") coupl_gauge_.gad[1]=val;
    else if (txt == "gal0") coupl_gauge_.gal[0]=val;
    else if (txt == "gal1") coupl_gauge_.gal[1]=val;
    // Z couplings
    else if (txt == "gzu0") coupl_gauge_.gzu[0]=val;
    else if (txt == "gzu1") coupl_gauge_.gzu[1]=val;
    else if (txt == "gzd0") coupl_gauge_.gzd[0]=val;
    else if (txt == "gzd1") coupl_gauge_.gzd[1]=val;
    else if (txt == "gzn0") coupl_gauge_.gzn[0]=val;
    else if (txt == "gzn1") coupl_gauge_.gzn[1]=val;
    else if (txt == "gzl0") coupl_gauge_.gzl[0]=val;
    else if (txt == "gzl1") coupl_gauge_.gzl[1]=val;
    // W couplings
    else if (txt.find("gwf") != string::npos) {
      if (txt == "gwf0") coupl_gauge_.gwf[0]=val;
      else if (txt == "gwf1") coupl_gauge_.gwf[1]=val;
      // CKM
      else if (txt == "gwfud0") coupl_ckm_.gwfud[0]=val;
      else if (txt == "gwfud1") coupl_ckm_.gwfud[1]=val;
      else if (txt == "gwfcd0") coupl_ckm_.gwfcd[0]=val;
      else if (txt == "gwfcd1") coupl_ckm_.gwfcd[1]=val;
      else if (txt == "gwftd0") coupl_ckm_.gwftd[0]=val;
      else if (txt == "gwftd1") coupl_ckm_.gwftd[1]=val;
      else if (txt == "gwfus0") coupl_ckm_.gwfus[0]=val;
      else if (txt == "gwfus1") coupl_ckm_.gwfus[1]=val;
      else if (txt == "gwfcs0") coupl_ckm_.gwfcs[0]=val;
      else if (txt == "gwfcs1") coupl_ckm_.gwfcs[1]=val;
      else if (txt == "gwfts0") coupl_ckm_.gwfts[0]=val;
      else if (txt == "gwfts1") coupl_ckm_.gwfts[1]=val;
      else if (txt == "gwfub0") coupl_ckm_.gwfub[0]=val;
      else if (txt == "gwfub1") coupl_ckm_.gwfub[1]=val;
      else if (txt == "gwfcb0") coupl_ckm_.gwfcb[0]=val;
      else if (txt == "gwfcb1") coupl_ckm_.gwfcb[1]=val;
      else if (txt == "gwftb0") coupl_ckm_.gwftb[0]=val;
      else if (txt == "gwftb1") coupl_ckm_.gwftb[1]=val;
    }
    // H couplings
    else if (txt == "ghbot0") coupl_yuk_.ghbot[0] = val;
    else if (txt == "ghbot1") coupl_yuk_.ghbot[1] = val;
    else if (txt == "gh0") coupl_heft_.gh[0] = val;
    else if (txt == "gh1") coupl_heft_.gh[1] = val;
    // Else issue error message: unrecognized coupling parameter
    else {
      printOut("MG4interface::setParm","Unknown parameter "+txt);
    }

  }

  void MG4interface::setParm(string txt, double val) {

    complex valC(val,0.0);
    setParm(txt, valC);

  }

  //*********

  // Wrapper to return unpolarized Madgraph Matrix Elements (squared)
  double MG4interface::ME2(int nIn, vector<int> id, vector<Vec4> p) {

    vector<double> pol;
    for (unsigned int i=0; i<p.size(); ++i) pol.push_back(9.);
    return ME2(nIn, id, p, pol);

  }

  // Wrapper to return polarized MadGraph Matrix Elements (squared)
  // Input:
  //    nIn   = number of particles that are incoming (1 or 2)
  //    id    = vector of particle IDs. The incoming particle(s) MUST be the 
  //            first nIn entries of this vector. The outgoing particles
  //            MUST be arranged in color order thereafter. 
  //    p     = vector of particle momenta. The incoming particle(s) MUST be 
  //            the first nIn entries of this vector. The outgoing particles
  //            MUST be arranged in color order thereafter. 
  //    pol   = vector of particle polarizations.  pol = 9 is unpolarized.
  //            The incoming particle(s) MUST be the first nIn entries of 
  //            this vector. The outgoing particles MUST be arranged in 
  //                color order thereafter.

  // Return:   value of Matrix Element squared for this phase-space point

  double MG4interface::ME2(int nIn, vector<int> id, vector<Vec4> p, 
                           vector<int> pol) {

    // Check if initialized
    if (! isInit) {
      printOut("MG4interface::ME2","MG4interface not initialized.");
    }

    // Initialize ME weight with dummy value.
    const  int nMom  = p.size();
    const  int nOut  = nMom-nIn;
    double MGME2 = -1.;

    // Sanity checks
    if (nMom <= 2) return MGME2;
    if (int(id.size()) != nMom) return MGME2;

    // Take out polarization averages
    int nPolIn = 1;

    // Default color ordering is number 1
    int icol = 1;

    // HEL allows to specify up to N specific helicity amplitudes to sum over:
    // N= 4 for unpolarised decay or pure QCD, or
    // N=12 for single boson production (2 unpolarised incoming + boson)
    // (enumeration follows MADGRAPH IHELs)
    // (To get the full helicity sum, just leave first component = -1)
    // Initialise for full helicity sum (individual IHELs are always positive)
    int HEL[4]     = {-1, -1, -1, -1};
    int HELsbp[12] = {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1};

    //------------------------------------------------------------------------

    int id0 = abs(id[0]);
    int idN = abs(id[nMom-1]);
    // If not known, return.
    if (id0 == 0 || idN == 0) return -1;

    // Check if using massive matrix elements and what kind of ME is asked for 
    vector<double> m;    
    bool massless    = true;
    bool isQCD       = true;
    bool isPolarised = false;
    bool isEEtoJets  = false;
    int nQuarkOut    = 0;
    int nGluonOut    = 0;
    int nQuarkIn     = 0;
    int nGluonIn     = 0;
    int nLeptonIn    = 0;
    int nLeptonOut   = 0;
    int nPolarised   = 0;
    vector<int> iQuark;
    vector<int> iGluon;
    vector<int> iLepton;
    double sCM       = 0.0;
    // Decay processes
    if (nIn == 1) {
      isQCD = false;
      sCM = p[0].m2Calc();
      // Fill vector of desired masses (zero ones that are offshell only by
      // an amount consistent with numerical precision)
      m.push_back(sqrt(p[0].m2Calc()));
      // Also count how many particles have assigned helicities
      if (pol[0] != 9) ++nPolarised;
      for (unsigned int i=1;i<id.size(); ++i) {      
        if (pol[i] != 9) ++nPolarised;
        // Count number of quarks (+antiquarks) and gluons
        if (id[i] == 21) {
          nGluonOut++;
          iGluon.push_back(i);
        }        
        else if (abs(id[i]) <= 10) {
          nQuarkOut++;
          iQuark.push_back(i);
        }
        double m2 = p[i].m2Calc();      
        if (abs(m2/sCM) > TINY) {
          massless = false;
          m.push_back(sqrt(abs(m2)));
        } else {
          m.push_back(0.0);
        }
      }
    }
    // 2->n processes
    else {
      massless   = true; // ISR massless!
      sCM        = 4.0*p[pos1Sav].e()*p[pos2Sav].e();
      for (int i=0;i<(int)id.size(); ++i) {
        // Count number of particles have assigned helicities
        if (pol[i] != 9) ++nPolarised;
        // Count number of quarks (+antiquarks) and gluons
        if (id[i] == 21) {
          if (i==pos1Sav || i==pos2Sav) nGluonIn++;
          else nGluonOut++;
          iGluon.push_back(i);
        }        
        else if (abs(id[i]) <= 10) {
          if (i==pos1Sav || i==pos2Sav) nQuarkIn++;
          else nQuarkOut++;
          iQuark.push_back(i);
        }
        else if (abs(id[i]) < 19 && abs(id[i]) > 10) {
          if (i==pos1Sav || i==pos2Sav) nLeptonIn++;
	  else nLeptonOut++;
          iLepton.push_back(i);
        }
        // Store mass values
        m.push_back(sqrt(p[i].m2Calc()));
        // Check if this is a pure-QCD process
        if (abs(id[i]) > 10 && abs(id[i]) != 21) isQCD = false;
      }
      // e+ e- -> jets
      isEEtoJets = (nLeptonIn == 2);
    }
    int nGluonTot  = nGluonOut + nGluonIn;
    int nQuarkTot  = nQuarkOut + nQuarkIn;
    // int nLeptonTot = nLeptonIn + nLeptonOut;

    // Check if we should evaluate helicity sum or specified helicity(ies)
    int nUnpolMax = ( ((nIn == 1) || isQCD) ? 2 : 3);
    if (nPolarised >= 1 && nMom - nPolarised <= nUnpolMax) isPolarised = true;

    // Verbose output
    if (verbose >= 8) 
      printOut("MG4interface::ME2","massless approximation = "
               +bool2str(massless));

    // Check Lorentz invariance 
    // (change to (true) to always boost back to CM) 
    if (false) {
      Vec4 vCM;
      if (nIn == 1) vCM = p[0];
      else vCM = p[pos1Sav]+p[pos2Sav];
      double eCM = vCM.e();
      if (eCM <= 0.0) {
        printOut("MG4interface::ME2",
                 "Error: failed to evaluate ME, eCM <= 0.0 ");
        return -1;
      }
      if (abs(vCM.px()/eCM) + abs(vCM.py()/eCM) + abs(vCM.pz()/eCM) > TINY) {
        cout<<"We are in a boosted frame. Boosting back"<<endl;
        for (unsigned int i=0; i<p.size(); ++i) p[i].bstback(vCM);
      }
    }

    //------------------------------------------------------------------------
    // QCD 2->n    
    if (nIn == 2 && isQCD) {

      // Massless only, setting all quark masses to zero
      double bMassSav   = coupl_mass_.bmass;
      coupl_mass_.bmass = 0.0;
      double cMassSav   = coupl_mass_.cmass;
      coupl_mass_.cmass = 0.0;
      double sMassSav   = coupl_mass_.smass;
      coupl_mass_.smass = 0.0;
      double dMassSav   = coupl_mass_.dmass;
      coupl_mass_.dmass = 0.0;
      double uMassSav   = coupl_mass_.umass;
      coupl_mass_.umass = 0.0;

      // Initialise helicity-dependence
      int nHels = 1;
      if (isPolarised) HEL[0] = 1;

      // ALL-GLUON AMPLITUDES ------ START

      // gg -> gluons
      if (nQuarkTot == 0) {
        // Vincia is colour-ordered, MG starts with IS partons, then FS ones
        // Convention: pick icol so that FS partons are in ascending order
        map<int,int> mapVtoMG;
        // pos1Sav has to be 0
        if (pos1Sav != 0)
          cout << " MG4interface : unrecognised colour order for gg -> gluons "
               << "amplitude" << endl;
        // How many partons inbetween the two IS gluons
        int nSep = pos2Sav-1;
        // First store IS gluon positions
        mapVtoMG[0] = pos1Sav;
        mapVtoMG[1] = pos2Sav;
        // Then store FS gluons, in ascending order
        int iMGnow=1;
        // Leave out first gluon (is initial)
        for (int i=1; i<(int)iGluon.size(); ++i) 
          if (iGluon[i] != pos2Sav) mapVtoMG[++iMGnow] = iGluon[i];
        if (verbose >= 7) {
          cout << " gg -> gluons amplitude with II = " << pos1Sav << " "
               << pos2Sav << " nSep = " << nSep << " nOut = " << nOut << endl;
          for (int i=0; i<nMom; ++i)
            cout << " iMG = " << i << " <- iV = " << mapVtoMG[i] << endl;
        }
        // Set MG momenta and helicities, then evaluate MG matrix elements
        if (nOut == 2) {
          icol = 3; // F[1, 2, 3, 4]
          if (nSep == 1) icol = 4; // F[1, 3, 2, 4]
          else if (nSep == 2) icol = 6; // F[1, 3, 4, 2]
          double PNEW[4][4];
          for (int iMG=nMom-1; iMG>=0; --iMG) {
            Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
            if (HEL[0] > 0) HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
            nHels *= 2;
          }
          if (verbose >= 7) cout << " icol = " << icol << endl;
          if (verbose >= 5 && HEL[0] > 0)
            cout << " Using iHEL = " << HEL[0] << endl;
          sgg2gg_(&colorDepth,&icol,PNEW,HEL,&MGME2); 
        } else if (nOut == 3) {
          icol = 5; // F[1, 2, 3, 4, 5]
          if (nSep == 1) icol = 6; // F[1, 3, 2, 4, 5]
          else if (nSep == 2) icol = 18; // F[1, 3, 4, 2, 5]
          else if (nSep == 3) icol = 22; // F[1, 3, 4, 5, 2]
          double PNEW[5][4];
          for (int iMG=nMom-1; iMG>=0; --iMG) {
            Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
            if (HEL[0] > 0) HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
            nHels *= 2;
          }
          if (verbose >= 7) cout << " icol = " << icol << endl;
          if (verbose >= 5 && HEL[0] > 0)
            cout << " Using iHEL = " << HEL[0] << endl;
          sgg2ggg_(&colorDepth,&icol,PNEW,HEL,&MGME2); 
        } else if (nOut == 4) {
          icol = 9; // F[1, 2, 3, 4, 5, 6]
          if (nSep == 1) icol = 10; // F[1, 3, 2, 4, 5, 6]
          else if (nSep == 2) icol = 86; // F[1, 3, 4, 2, 5, 6]
          else if (nSep == 3) icol = 104; // F[1, 3, 4, 5, 2, 6]
          else if (nSep == 4) icol = 106; // F[1, 3, 4, 5, 6, 2]
          double PNEW[6][4];
          for (int iMG=nMom-1; iMG>=0; --iMG) {
            Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
            if (HEL[0] > 0) HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
            nHels *= 2;
          }
          if (verbose >= 7) cout << " icol = " << icol << endl;
          if (verbose >= 5 && HEL[0] > 0)
            cout << " Using iHEL = " << HEL[0] << endl;
          sgg2gggg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        }
      } // End gg -> gluons

      // ALL-GLUON AMPLITUDES ------ END

      // TWO-QUARK AMPLITUDES ------ START

      // qg -> q + gluons (& c.c.)
      else if (nQuarkIn == 1 && nQuarkOut == 1) {
        // Vincia is colour-ordered, MG starts with IS partons, then FS ones
        // Convention: pick icol so that FS partons are in ascending order
        map<int,int> mapVtoMG;
        // How many partons inbetween FS quark and IS gluon
        int direction = 1;
        int nSep      = pos1Sav-1;
        // Check if this is a qBar g -> qBar + gluons, to be charge-conjugated
        // If so, count separation from FS antiquark, now at the end of the chain
        if (pos2Sav == nMom -1) {
          // Standard colour order : qg -> q + gluons
          // First store quark positions
          mapVtoMG[0] = pos2Sav;
          mapVtoMG[2] = 0;
          // Store IS gluon
          mapVtoMG[1] = pos1Sav;
          // Then store FS gluons, in ascending order
          int iMG=2;
          for (int i=0; i<(int)iGluon.size(); ++i) 
            if (iGluon[i] != pos1Sav) mapVtoMG[++iMG] = iGluon[i];
        } else if (pos1Sav == 0) {
          // Reverse colour order : qbar g -> qbar + gluons
          direction = -1;
          nSep = (nMom - 1) - (pos2Sav + 1);
          // First store quark positions
          mapVtoMG[0] = pos1Sav;
          mapVtoMG[2] = nMom - 1;
          // Store IS gluon
          mapVtoMG[1] = pos2Sav;
          // Then store FS gluons, in ascending order (reverse direction)
          int iMG=2;
          for (int i=iGluon.size()-1; i>=0; --i) 
            if (iGluon[i] != pos2Sav) mapVtoMG[++iMG] = iGluon[i];
        } else {
          cout << " MG4interface : unrecognised colour order for qg -> q + "
               << "gluons amplitude" << endl;
        }
        if (verbose >= 7) {
          cout << " qg -> q + gluons amplitude with II = " << pos1Sav << " "
               << pos2Sav << " nSep = " << nSep << " direction = " << direction
               << " nOut = " << nOut << endl;
          for (int i=0; i<nMom; ++i)
            cout << " iMG = " << i << " <- iV = " << mapVtoMG[i] << endl;
        }
        // Set MG momenta and helicities, then evaluate MG matrix elements
        if (nOut == 2) {
          icol = ((nSep == 0) ? 1 : 2); // T[ 3, 1, 2, 4] : T[ 3, 1, 4, 2]
          double PNEW[4][4];
          for (int iMG=nMom-1; iMG>=0; --iMG) {
            Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
            if (HEL[0] > 0) HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
            nHels *= 2;
          }
          if (verbose >= 7) cout << " icol = " << icol << endl;
          if (verbose >= 5 && HEL[0] > 0)
            cout << " Using iHEL = " << HEL[0] << endl;
          sug2ug_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        } else if (nOut == 3) {
          icol = 1; // T[ 3, 1, 2, 4, 5]
          if (nSep == 1) icol = 4; // T[ 3, 1, 4, 2, 5]
          else if (nSep == 2) icol = 5; // T[ 3, 1, 4, 5, 2]
          double PNEW[5][4];
          for (int iMG=nMom-1; iMG>=0; --iMG) {
            Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
            if (HEL[0] > 0) HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
            nHels *= 2;
          }
          if (verbose >= 7) cout << " icol = " << icol << endl;
          if (verbose >= 5 && HEL[0] > 0)
            cout << " Using iHEL = " << HEL[0] << endl;
          sug2ugg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        } else if (nOut == 4) {
          icol = 1; // T[ 3, 1, 2, 4, 5, 6]
          if (nSep == 1) icol = 13; // T[ 3, 1, 4, 2, 5, 6]
          else if (nSep == 2) icol = 17; // T[ 3, 1, 4, 5, 2, 6]
          else if (nSep == 3) icol = 18; // T[ 3, 1, 4, 5, 6, 2]
          double PNEW[6][4];
          for (int iMG=nMom-1; iMG>=0; --iMG) {
            Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
            if (HEL[0] > 0) HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
            nHels *= 2;
          }
          if (verbose >= 7) cout << " icol = " << icol << endl;
          if (verbose >= 5 && HEL[0] > 0)
            cout << " Using iHEL = " << HEL[0] << endl;
          sug2uggg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        } else if (nOut == 5) {
          icol = 1; // T[ 3, 1, 2, 4, 5, 6, 7]
          if (nSep == 1) icol = 61; // T[ 3, 1, 4, 2, 5, 6, 7]
          else if (nSep == 2) icol = 81; // T[ 3, 1, 4, 5, 2, 6, 7]
          else if (nSep == 3) icol = 86; // T[ 3, 1, 4, 5, 6, 2, 7]
          else if (nSep == 4) icol = 87; // T[ 3, 1, 4, 5, 6, 7, 2]
          double PNEW[7][4];
          for (int iMG=nMom-1; iMG>=0; --iMG) {
            Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
            if (HEL[0] > 0) HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
            nHels *= 2;
          }
          if (verbose >= 7) cout << " icol = " << icol << endl;
          if (verbose >= 5 && HEL[0] > 0)
            cout << " Using iHEL = " << HEL[0] << endl;
          sug2ugggg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        }
      } // End qg -> q + gluons

      // qqbar -> gluons
      else if (nQuarkIn == 2 && nQuarkOut == 0) {
        map<int,int> mapVtoMG;
        // pos1Sav has to be 0 (incoming qbar)
        // pos2Sav has to be nMom-1 (incoming q)
        if (pos1Sav != 0 || pos2Sav != nMom-1)
          cout << " MG4interface : unrecognised colour order for qqbar -> "
               << "gluons amplitude" << endl;
        // First store IS quark positions
        mapVtoMG[0] = pos2Sav;
        mapVtoMG[1] = pos1Sav;
        // Then store FS gluons, in ascending order
        int iMGnow=1;
        for (int i=0; i<(int)iGluon.size(); ++i) 
          mapVtoMG[++iMGnow] = iGluon[i];
        if (verbose >= 7) {
          cout << " qqbar -> gluons amplitude with II = " << pos1Sav << " "
               << pos2Sav << " nOut = " << nOut << endl;
          for (int i=0; i<nMom; ++i)
            cout << " iMG = " << i << " <- iV = " << mapVtoMG[i] << endl;
        }
        // Set MG momenta and helicities, then evaluate MG matrix elements
        if (nOut == 2) {
          icol = 1; // T[ 2, 1, 3, 4]
          double PNEW[4][4];
          for (int iMG=nMom-1; iMG>=0; --iMG) {
            Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
            if (HEL[0] > 0) HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
            nHels *= 2;
          }
          if (verbose >= 7) cout << " icol = " << icol << endl;
          if (verbose >= 5 && HEL[0] > 0)
            cout << " Using iHEL = " << HEL[0] << endl;
          suux2gg_(&colorDepth,&icol,PNEW,HEL,&MGME2); 
        } else if (nOut == 3) {
          icol = 1; // T[ 2, 1, 3, 4, 5]
          double PNEW[5][4];
          for (int iMG=nMom-1; iMG>=0; --iMG) {
            Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
            if (HEL[0] > 0) HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
            nHels *= 2;
          }
          if (verbose >= 7) cout << " icol = " << icol << endl;
          if (verbose >= 5 && HEL[0] > 0)
            cout << " Using iHEL = " << HEL[0] << endl;
          suux2ggg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        } else if (nOut == 4) {
          icol = 1; //  T[ 2, 1, 3, 4, 5, 6]
          double PNEW[6][4];
          for (int iMG=nMom-1; iMG>=0; --iMG) {
            Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
            if (HEL[0] > 0) HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
            nHels *= 2;
          }
          if (verbose >= 7) cout << " icol = " << icol << endl;
          if (verbose >= 5 && HEL[0] > 0)
            cout << " Using iHEL = " << HEL[0] << endl;
          suux2gggg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        } else if (nOut == 5) {
          icol = 1; // T[ 2, 1, 3, 4, 5, 6, 7]
          double PNEW[7][4];
          for (int iMG=nMom-1; iMG>=0; --iMG) {
            Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
            if (HEL[0] > 0) HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
            nHels *= 2;
          }
          if (verbose >= 7) cout << " icol = " << icol << endl;
          if (verbose >= 5 && HEL[0] > 0)
            cout << " Using iHEL = " << HEL[0] << endl;
          suux2ggggg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        }
      } // End qqbar -> gluons

      // gg -> qqbar + gluons
      else if (nQuarkIn == 0 && nQuarkOut == 2) {
        map<int,int> mapVtoMG;
        // First store IS gluon positions
        mapVtoMG[0] = pos1Sav;
        mapVtoMG[1] = pos2Sav;
        // Then store FS quark positions
        mapVtoMG[2] = 0;
        mapVtoMG[3] = nMom-1;
        // Then store FS gluons, in ascending order
        int iMGnow=3;
        for (int i=0; i<(int)iGluon.size(); ++i) 
          if ((iGluon[i] != pos1Sav) && (iGluon[i] != pos2Sav)) 
            mapVtoMG[++iMGnow] = iGluon[i];
        if (verbose >= 7) {
          cout << " gg -> qqbar + gluons amplitude with II = " << pos1Sav << " "
               << pos2Sav << " nOut = " << nOut << endl;
          for (int i=0; i<nMom; ++i)
            cout << " iMG = " << i << " <- iV = " << mapVtoMG[i] << endl;
        }
        // Set MG momenta and helicities, then evaluate MG matrix elements
        if (nOut == 2) {
          if (pos1Sav == 1 && pos2Sav == 2)
            icol = 2; // T[ 3, 4, 1, 2] 
          else cout << " MG4interface : unrecognised colour order for qg -> "
                    << "qqbar + gluons amplitude" << endl;
          double PNEW[4][4];
          for (int iMG=nMom-1; iMG>=0; --iMG) {
            Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
            if (HEL[0] > 0) HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
            nHels *= 2;
          }
          if (verbose >= 7) cout << " icol = " << icol << endl;
          if (verbose >= 5 && HEL[0] > 0)
            cout << " Using iHEL = " << HEL[0] << endl;
          sgg2uux_(&colorDepth,&icol,PNEW,HEL,&MGME2); 
        } else if (nOut == 3) {
          if (pos1Sav == 1 && pos2Sav == 2)
            icol = 6; // T[ 3, 4, 1, 2, 5]
          else if (pos1Sav == 1 && pos2Sav == 3)
            icol = 4; // T[ 3, 4, 1, 5, 2]
          else if (pos1Sav == 2 && pos2Sav == 3)
            icol = 5; // T[ 3, 4, 5, 1, 2]
          else cout << " MG4interface : unrecognised colour order for qg -> "
                    << "qqbar + gluons amplitude" << endl;
          double PNEW[5][4];
          for (int iMG=nMom-1; iMG>=0; --iMG) {
            Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
            if (HEL[0] > 0) HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
            nHels *= 2;
          }
          if (verbose >= 7) cout << " icol = " << icol << endl;
          if (verbose >= 5 && HEL[0] > 0)
            cout << " Using iHEL = " << HEL[0] << endl;
          sgg2uuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        } else if (nOut == 4) {
          if (pos1Sav == 1 && pos2Sav == 2)
            icol = 24; // T[ 3, 4, 1, 2, 5, 6]
          else if (pos1Sav == 1 && pos2Sav == 3)
            icol = 17; // T[ 3, 4, 1, 5, 2, 6]
          else if (pos1Sav == 1 && pos2Sav == 4)
            icol = 19; // T[ 3, 4, 1, 5, 6, 2]
          else if (pos1Sav == 2 && pos2Sav == 3)
            icol = 18; // T[ 3, 4, 5, 1, 2, 6]
          else if (pos1Sav == 2 && pos2Sav == 4)
            icol = 13; // T[ 3, 4, 5, 1, 6, 2]
          else if (pos1Sav == 3 && pos2Sav == 4)
            icol = 15; // T[ 3, 4, 5, 6, 1, 2]
          else cout << " MG4interface : unrecognised colour order for qg -> "
                    << "qqbar + gluons amplitude" << endl;
          double PNEW[6][4];
          for (int iMG=nMom-1; iMG>=0; --iMG) {
            Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
            if (HEL[0] > 0) HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
            nHels *= 2;
          }
          if (verbose >= 7) cout << " icol = " << icol << endl;
          if (verbose >= 5 && HEL[0] > 0)
            cout << " Using iHEL = " << HEL[0] << endl;
          sgg2uuxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        } else if (nOut == 5) {
          if (pos1Sav == 1 && pos2Sav == 2)
            icol = 119; // T[ 3, 4, 1, 2, 5, 6, 7]
          else if (pos1Sav == 1 && pos2Sav == 3)
            icol =  93; // T[ 3, 4, 1, 5, 2, 6, 7]
          else if (pos1Sav == 1 && pos2Sav == 4)
            icol =  90; // T[ 3, 4, 1, 5, 6, 2, 7]
          else if (pos1Sav == 1 && pos2Sav == 5)
            icol =  91; // T[ 3, 4, 1, 5, 6, 7, 2]
          else if (pos1Sav == 2 && pos2Sav == 3)
            icol =  94; // T[ 3, 4, 5, 1, 2, 6, 7]
          else if (pos1Sav == 2 && pos2Sav == 4)
            icol =  69; // T[ 3, 4, 5, 1, 6, 2, 7]
          else if (pos1Sav == 2 && pos2Sav == 5)
            icol =  70; // T[ 3, 4, 5, 1, 6, 7, 2]
          else if (pos1Sav == 3 && pos2Sav == 4)
            icol =  73; // T[ 3, 4, 5, 6, 1, 2, 7]
          else if (pos1Sav == 3 && pos2Sav == 5)
            icol =  65; // T[ 3, 4, 5, 6, 1, 7, 2]
          else if (pos1Sav == 4 && pos2Sav == 5)
            icol =  66; // T[ 3, 4, 5, 6, 7, 1, 2]
          else cout << " MG4interface : unrecognised colour order for qg -> "
                    << "qqbar + gluons amplitude" << endl;
          double PNEW[7][4];
          for (int iMG=nMom-1; iMG>=0; --iMG) {
            Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
            if (HEL[0] > 0) HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
            nHels *= 2;
          }
          if (verbose >= 7) cout << " icol = " << icol << endl;
          if (verbose >= 5 && HEL[0] > 0)
            cout << " Using iHEL = " << HEL[0] << endl;
          sgg2uuxggg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        }
      } // End gg -> qqbar + gluons

      // TWO-QUARK AMPLITUDES ------ END

      // FOUR-QUARK AMPLITUDES ------ START

      else if (nQuarkTot == 4) {

        // Check if all quarks have same flavour
        // Compare to first quark
        int idComp = abs(id[0]);
        int count  = 0;
        for (int i=1; i<nMom; i++) if (abs(id[i]) == idComp) count++;
        bool sameFlav = (count > 1);

        // 4 quarks
        if (nOut == 2) {
          double PNEW[4][4];
          int id1 = id[pos1Sav];
          int id2 = id[pos2Sav];
          if (verbose >= 7) {
            cout << " 4-quark amplitude with II = " << pos1Sav << " "
                 << pos2Sav << " idII = " << id1 << id2 << endl << " ";
            for (int i=0; i<nMom; ++i) cout << id[i] << " "; cout << endl;
          }
          // q q -> q q and qbar qbar -> qbar qbar
          if (id1 == id2) {
            if (!sameFlav) cout << " MG4interface : unrecognised colour order "
                                << "for 4-quark amplitude" << endl;
            if (pos1Sav == 0 && pos2Sav == 2) {
              // (-q, qbar, -q, qbar)
              Vec4toMGQCD(p, 0, 2, 1, 3, PNEW);
              icol = 1; // T[ 4, 2]T[ 3, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[4] = {0, 2, 1, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uu_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 3) {
              // (q, -qbar, q, -qbar)
              Vec4toMGQCD(p, 3, 1, 0, 2, PNEW);
              icol = 2; // T[ 4, 1]T[ 3, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[4] = {3, 1, 0, 2};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uu_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark amplitude" << endl;
          }
          // q qbar -> q qbar and q qbar -> q' qbar'
          else if (id1 == -id2) {
            if (pos1Sav == 0 && pos2Sav == 1) {
              // (-q, -qbar, q, qbar)
              Vec4toMGQCD(p, 1, 0, 2, 3, PNEW);
              if (HEL[0] > 0) {
                int mapVtoMG[4] = {1, 0, 2, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else cout << " MG4interface : unrecognised colour order "
                        << "for 4-quark amplitude" << endl;
            }
            else if (pos1Sav == 2 && pos2Sav == 3) {
              // (q, qbar, -q, -qbar)
              Vec4toMGQCD(p, 3, 2, 0, 1, PNEW);
              icol = 1; // T[ 2, 1]T[ 3, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[4] = {3, 2, 0, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else cout << " MG4interface : unrecognised colour order "
                        << "for 4-quark amplitude" << endl;
            }
            else if (pos1Sav == 0 && pos2Sav == 3) {
              // (-q, qbar, q, -qbar)
              Vec4toMGQCD(p, 3, 0, 2, 1, PNEW);
              icol = (sameFlav ? 2 : 1); // T[ 3, 1]T[ 2, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[4] = {3, 0, 2, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          suux2ddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 2) {
              // (q, -qbar, -q, qbar)
              Vec4toMGQCD(p, 1, 2, 0, 3, PNEW);
              icol = (sameFlav ? 2 : 1); // T[ 3, 1]T[ 2, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[4] = {1, 2, 0, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          suux2ddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark amplitude" << endl;
          }
          // q q' -> q q' and qbar qbar' -> qbar qbar'
          else if (id1*id2 > 0) {
            if (sameFlav) cout << " MG4interface : unrecognised colour order "
                               << "for 4-quark amplitude" << endl;
            if (pos1Sav == 0 && pos2Sav == 2) {
              // (-q, qbar, -q, qbar)
              Vec4toMGQCD(p, 0, 2, 3, 1, PNEW);
              icol = 1; // T[ 4, 1]T[ 3, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[4] = {0, 2, 3, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2ud_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 3) {
              // (q, -qbar, q, -qbar)
              Vec4toMGQCD(p, 1, 3, 2, 0, PNEW);
              icol = 1; // T[ 4, 1]T[ 3, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[4] = {1, 3, 2, 0};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2ud_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark amplitude" << endl;
          }
          // qqbar' -> qqbar' and qbarq' -> qbarq'
          else if (id1*id2 < 0) {
            if (sameFlav) cout << " MG4interface : unrecognised colour order "
                               << "for 4-quark amplitude" << endl;
            if (pos1Sav == 0 && pos2Sav == 1) {
              // (-q, -qbar, q, qbar)
              Vec4toMGQCD(p, 1, 0, 2, 3, PNEW);
              icol = 1; // T[ 2, 1]T[ 3, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[4] = {1, 0, 2, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 3) {
              // (q, qbar, -q, -qbar)
              Vec4toMGQCD(p, 3, 2, 0, 1, PNEW);
              icol = 1; // T[ 2, 1]T[ 3, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[4] = {3, 2, 0, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark amplitude" << endl;
          }
          else cout << " MG4interface : unrecognised colour order "
                    << "for 4-quark amplitude" << endl;
        } // End 4 quarks

        // 4 quarks + gluon
        else if (nOut == 3) {
          double PNEW[5][4];
          int id1 = id[pos1Sav];
          int id2 = id[pos2Sav];
          if (verbose >= 7) {
            cout << " 4-quark + g amplitude with II = " << pos1Sav << " "
                 << pos2Sav << " idII = " << id1 << id2 << endl << " ";
            for (int i=0; i<nMom; ++i) cout << id[i] << " "; cout << endl;
          }
          // q q -> q q g and qbar qbar -> qbar qbar g
          if (id1 == id2) {
            if (!sameFlav) cout << " MG4interface : unrecognised colour order "
                                << "for 4-quark + g amplitude" << endl;
            if (pos1Sav == 0 && pos2Sav == 2) {
              // (-q, qbar, -q, g, qbar)
              Vec4toMGQCD(p, 0, 2, 1, 4, 3, PNEW);
              icol = 3; // T[ 4, 2, 5]T[ 3, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[5] = {0, 2, 1, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uug_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 4) {
              // (q, -qbar, q, g, -qbar)
              Vec4toMGQCD(p, 4, 1, 0, 2, 3, PNEW);
              icol = 2; // T[ 4, 1, 5]T[ 3, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[5] = {4, 1, 0, 2, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uug_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + g amplitude" << endl;
          }
          // q qbar -> q qbar g and q qbar -> q' qbar' g
          else if (id1 == -id2) {
            if (pos1Sav == 0 && pos2Sav == 1) {
              // (-q, -qbar, q, g, qbar)
              Vec4toMGQCD(p, 1, 0, 2, 4, 3, PNEW);
              icol = 2; // T[ 2, 1]T[ 3, 4, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[5] = {1, 0, 2, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else cout << " MG4interface : unrecognised colour order "
                        << "for 4-quark + g amplitude" << endl;
            }
            else if (pos1Sav == 2 && pos2Sav == 4) {
              // (q, qbar, -q, g, -qbar)
              Vec4toMGQCD(p, 4, 2, 0, 1, 3, PNEW);
              icol = 1; // T[ 2, 1, 5]T[ 3, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[5] = {4, 2, 0, 1, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else cout << " MG4interface : unrecognised colour order "
                        << "for 4-quark + g amplitude" << endl;
            }
            else if (pos1Sav == 0 && pos2Sav == 4) {
              // (-q, qbar, q, g, -qbar)
              Vec4toMGQCD(p, 4, 0, 2, 1, 3, PNEW);
              icol = (sameFlav ? 3 : 1); // T[ 3, 1, 5]T[ 2, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[5] = {4, 0, 2, 1, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          suux2ddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 2) {
              // (q, -qbar, -q, g, qbar)
              Vec4toMGQCD(p, 1, 2, 0, 4, 3, PNEW);
              icol = (sameFlav ? 4 : 2); // T[ 3, 1]T[ 2, 4, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[5] = {1, 2, 0, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          suux2ddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + g amplitude" << endl;
          }
          // q q' -> q q' g and qbar qbar' -> qbar qbar' g
          else if (id1*id2 > 0 && id1 != 21 && id2 != 21) {
            if (sameFlav) cout << " MG4interface : unrecognised colour order "
                               << "for 4-quark + g amplitude" << endl;
            if (pos1Sav == 0 && pos2Sav == 2) {
              // (-q, qbar, -q, g, qbar)
              Vec4toMGQCD(p, 0, 2, 4, 1, 3, PNEW);
              icol = 2; // T[ 4, 1]T[ 3, 2, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[5] = {0, 2, 4, 1, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 4) {
              // (q, -qbar, q, g, -qbar)
              Vec4toMGQCD(p, 1, 4, 2, 0, 3, PNEW);
              icol = 2; // T[ 4, 1]T[ 3, 2, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[5] = {1, 4, 2, 0, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + g amplitude" << endl;
          }
          // q qbar' -> q qbar' g and qbar q' -> qbar q' g
          else if (id1*id2 < 0 && id1 != 21 && id2 != 21) {
            if (sameFlav) cout << " MG4interface : unrecognised colour order "
                               << "for 4-quark + g amplitude" << endl;
            if (pos1Sav == 0 && pos2Sav == 1) {
              // (-q, -qbar, q, g, qbar)
              Vec4toMGQCD(p, 1, 0, 2, 4, 3, PNEW);
              icol = 2; // T[ 2, 1]T[ 3, 4, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[5] = {1, 0, 2, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 4) {
              // (q, qbar, -q, g, -qbar)
              Vec4toMGQCD(p, 4, 2, 0, 1, 3, PNEW);
              icol = 1; // T[ 2, 1, 5]T[ 3, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[5] = {4, 2, 0, 1, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + g amplitude" << endl;
          }
          // q g -> q q qbar and q g -> q q' qbar'
          // qbar g -> qbar qbar q and qbar g -> qbar qbar' q'
          else if (id1 == 21 || id2 == 21) {
            if (pos1Sav == 0 && pos2Sav == 3) {
              // (-q, qbar, q, -g, qbar)
              Vec4toMGQCD(p, 0, 3, 4, 1, 2, PNEW);
              icol = (sameFlav ? 3 : 1); // T[ 4, 1]T[ 3, 5, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[5] = {0, 3, 4, 1, 2};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 3) {
              // (q, -qbar, q, -g, qbar)
              Vec4toMGQCD(p, 1, 3, 2, 0, 4, PNEW);
              icol = (sameFlav ? 3 : 1); // T[ 4, 1]T[ 3, 5, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[5] = {1, 3, 2, 0, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 3) {
              // (q, qbar, -q, -g, qbar)
              Vec4toMGQCD(p, 2, 3, 1, 4, 0, PNEW);
              icol = (sameFlav ? 4 : 2); // T[ 4, 1, 2]T[ 3, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[5] = {2, 3, 1, 4, 0};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 3 && pos2Sav == 4) {
              // (q, qbar, q, -g, -qbar)
              Vec4toMGQCD(p, 4, 3, 0, 2, 1, PNEW);
              icol = (sameFlav ? 4 : 2); // T[ 4, 1, 2]T[ 3, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[5] = {4, 3, 0, 2, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + g amplitude" << endl;
          }
          else cout << " MG4interface : unrecognised colour order "
                    << "for 4-quark + g amplitude" << endl;
        } // End 4 quarks + gluon

        // 4 quarks + 2 connected gluons
        else if (nOut == 4 && id[3] == 21 && id[4] == 21) {
          double PNEW[6][4];
          int id1 = id[pos1Sav];
          int id2 = id[pos2Sav];
          if (verbose >= 7) {
            cout << " 4-quark + gg amplitude with II = " << pos1Sav << " "
                 << pos2Sav << " idII = " << id1 << id2 << endl << " ";
            for (int i=0; i<nMom; ++i) cout << id[i] << " "; cout << endl;
          }
          // q q -> q q g g and qbar qbar -> qbar qbar g g
          if (id1 == id2 && id1 != 21) {
            if (!sameFlav) cout << " MG4interface : unrecognised colour order "
                                << "for 4-quark + gg amplitude" << endl;
            if (pos1Sav == 0 && pos2Sav == 2) {
              // (-q, qbar, -q, g, g, qbar)
              Vec4toMGQCD(p, 0, 2, 1, 5, 3, 4, PNEW);
              icol = 5; // T[ 4, 2, 5, 6]T[ 3, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {0, 2, 1, 5, 3, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uugg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 5) {
              // (q, -qbar, q, g, g, -qbar)
              Vec4toMGQCD(p, 5, 1, 0, 2, 3, 4, PNEW);
              icol = 7; // T[ 4, 1, 5, 6]T[ 3, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 1, 0, 2, 3, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uugg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                    << "for 4-quark + gg amplitude" << endl;
          }
          // q qbar -> q qbar g g and q qbar -> q' qbar' g g
          else if (id1 == -id2) {
            if (pos1Sav == 0 && pos2Sav == 1) {
              // (-q, -qbar, q, g, g, qbar)
              Vec4toMGQCD(p, 1, 0, 2, 5, 3, 4, PNEW);
              icol = 6; // T[ 2, 1]T[ 3, 4, 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 0, 2, 5, 3, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uuxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else cout << " MG4interface : unrecognised colour order "
                        << "for 4-quark + gg amplitude" << endl;
            }
            else if (pos1Sav == 0 && pos2Sav == 5) {
              // (-q, qbar, q, g, g, -qbar)
              Vec4toMGQCD(p, 5, 0, 2, 1, 3, 4, PNEW);
              icol = (sameFlav ? 7 : 1); // T[ 3, 1, 5, 6]T[ 2, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 0, 2, 1, 3, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uuxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          suux2ddxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 2) {
              // (q, -qbar, -q, g, g, qbar)
              Vec4toMGQCD(p, 1, 2, 0, 5, 3, 4, PNEW);
              icol = (sameFlav ? 12 : 6); // T[ 3, 1]T[ 2, 4, 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 2, 0, 5, 3, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uuxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          suux2ddxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 5) {
              // (q, qbar, -q, g, g, -qbar)
              Vec4toMGQCD(p, 5, 2, 0, 1, 3, 4, PNEW);
              icol = 1; // T[ 2, 1, 5, 6]T[ 3, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 2, 0, 1, 3, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uuxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else cout << " MG4interface : unrecognised colour order "
                        << "for 4-quark + gg amplitude" << endl;
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + gg amplitude" << endl;
          }
          // q q' -> q q' g g and qbar qbar' -> qbar qbar' g g
          else if (id1*id2 > 0 && id1 != 21 && id2 != 21) {
            if (sameFlav) cout << " MG4interface : unrecognised colour order "
                               << "for 4-quark + gg amplitude" << endl;
            if (pos1Sav == 0 && pos2Sav == 2) {
              // (-q, qbar, -q, g, g, qbar)
              Vec4toMGQCD(p, 0, 2, 5, 1, 3, 4, PNEW);
              icol = 5; // T[ 4, 1]T[ 3, 2, 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {0, 2, 5, 1, 3, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 5) {
              // (q, -qbar, q, g, g, -qbar)
              Vec4toMGQCD(p, 1, 5, 2, 0, 3, 4, PNEW);
              icol = 5; // T[ 4, 1]T[ 3, 2, 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 5, 2, 0, 3, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + gg amplitude" << endl;
          }
          // q qbar' -> q qbar' g g and qbar q' -> qbar q' g g
          else if (id1*id2 < 0 && id1 != 21 && id2 != 21) {
            if (sameFlav) cout << " MG4interface : unrecognised colour order "
                               << "for 4-quark + gg amplitude" << endl;
            if (pos1Sav == 0 && pos2Sav == 1) {
              // (-q, -qbar, q, g, g, qbar)
              Vec4toMGQCD(p, 1, 0, 2, 5, 3, 4, PNEW);
              icol = 6; // T[ 2, 1]T[ 3, 4, 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 0, 2, 5, 3, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 5) {
              // (q, qbar, -q, g, g, -qbar)
              Vec4toMGQCD(p, 5, 2, 0, 1, 3, 4, PNEW);
              icol = 1; // T[ 2, 1, 5, 6]T[ 3, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 2, 0, 1, 3, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + gg amplitude" << endl;
          }
          // q g -> q q qbar g and q g -> q q' qbar' g
          // qbar g -> qbar qbar q g and qbar g -> qbar qbar' q' g
          else if ( (id1 == 21 && id2 != 21 ) || (id2 == 21 && id1 != 21) ) {
            if (pos1Sav == 0 && pos2Sav == 3) {
              // (-q, qbar, q, -g, g, qbar)
              Vec4toMGQCD(p, 0, 3, 5, 1, 2, 4, PNEW);
              icol = (sameFlav ? 9 : 3); // T[ 4, 1]T[ 3, 5, 6, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {0, 3, 5, 1, 2, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 0 && pos2Sav == 4) {
              // (-q, qbar, q, g, -g, qbar)
              Vec4toMGQCD(p, 0, 4, 5, 1, 2, 3, PNEW);
              icol = (sameFlav ? 8 : 2); // T[ 4, 1]T[ 3, 5, 2, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {0, 4, 5, 1, 2, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 3) {
              // (q, -qbar, q, -g, g, qbar)
              Vec4toMGQCD(p, 1, 3, 2, 0, 5, 4, PNEW);
              icol = (sameFlav ? 8 : 2); // T[ 4, 1]T[ 3, 5, 2, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 3, 2, 0, 5, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 4) {
              // (q, -qbar, q, g, -g, qbar)
              Vec4toMGQCD(p, 1, 4, 2, 0, 5, 3, PNEW);
              icol = (sameFlav ? 9 : 3); // T[ 4, 1]T[ 3, 5, 6, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 4, 2, 0, 5, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 3) {
              // (q, qbar, -q, -g, g, qbar)
              Vec4toMGQCD(p, 2, 3, 1, 5, 0, 4, PNEW);
              icol = (sameFlav ? 11 : 5); // T[ 4, 1, 6, 2]T[ 3, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {2, 3, 1, 5, 0, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 4) {
              // (q, qbar, -q, g, -g, qbar)
              Vec4toMGQCD(p, 2, 4, 1, 5, 0, 3, PNEW);
              icol = (sameFlav ? 10 : 4); // T[ 4, 1, 2, 6]T[ 3, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {2, 4, 1, 5, 0, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 3 && pos2Sav == 5) {
              // (q, qbar, q, -g, g, -qbar)
              Vec4toMGQCD(p, 5, 3, 0, 2, 1, 4, PNEW);
              icol = (sameFlav ? 10 : 4); // T[ 4, 1, 2, 6]T[ 3, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 3, 0, 2, 1, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 4 && pos2Sav == 5) {
              // (q, qbar, q, g, -g, -qbar)
              Vec4toMGQCD(p, 5, 4, 0, 2, 1, 3, PNEW);
              icol = (sameFlav ? 11 : 5); // T[ 4, 1, 6, 2]T[ 3, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 4, 0, 2, 1, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + gg amplitude" << endl;
          }
          
          // gg -> qqbarqqbar and gg -> qqbarq'qbar'
          else if (id1 == 21 && id2 == 21) {
            if (pos1Sav == 3 && pos2Sav == 4) {
              // (q, qbar, q, -g, -g, qbar)
              Vec4toMGQCD(p, 3, 4, 0, 5, 2, 1, PNEW);
              icol = (sameFlav ? 5 : 5); // T[ 5, 4, 1, 2]T[ 3, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {3, 4, 0, 5, 2, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sgg2uuxuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sgg2uuxddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + gg amplitude" << endl;
          }
          else cout << " MG4interface : unrecognised colour order "
                    << "for 4-quark + gg amplitude" << endl;
        } // End 4 quarks + 2 connected gluons

        // 4 quarks + 2 unconnected gluons
        else if (nOut == 4 && id[1] == 21 && id[4] == 21) {
          double PNEW[6][4];
          int id1 = id[pos1Sav];
          int id2 = id[pos2Sav];
          if (verbose >= 7) {
            cout << " 4-quark + gg amplitude with II = " << pos1Sav << " "
                 << pos2Sav << " idII = " << id1 << id2 << endl << " ";
            for (int i=0; i<nMom; ++i) cout << id[i] << " "; cout << endl;
          }
          // q q -> q q g g and qbar qbar -> qbar qbar g g
          if (id1 == id2 && id1 != 21) {
            if (!sameFlav) cout << " MG4interface : unrecognised colour order "
                                << "for 4-quark + gg amplitude" << endl;
            if (pos1Sav == 0 && pos2Sav == 3) {
              // (-q, g, qbar, -q, g, qbar)
              Vec4toMGQCD(p, 0, 3, 2, 5, 1, 4, PNEW);
              icol = 2; // T[ 4, 2, 6]T[ 3, 1, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {0, 3, 2, 5, 1, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uugg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 5) {
              // (q, g, -qbar, q, g, -qbar)
              Vec4toMGQCD(p, 5, 2, 0, 3, 1, 4, PNEW);
              icol = 10; // T[ 4, 1, 6]T[ 3, 2, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 2, 0, 3, 1, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uugg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + gg amplitude" << endl;
          }
          // q qbar -> q qbar g g and q qbar -> q' qbar' g g
          else if (id1 == -id2) {
            if (pos1Sav == 0 && pos2Sav == 2) {
              // (-q, g, -qbar, q, g, qbar)
              Vec4toMGQCD(p, 2, 0, 3, 5, 4, 1, PNEW);
              icol = 4; // T[ 2, 1, 6]T[ 3, 4, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {2, 0, 3, 5, 4, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uuxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else cout << " MG4interface : unrecognised colour order "
                        << "for 4-quark + gg amplitude" << endl;
            }
            else if (pos1Sav == 3 && pos2Sav == 5) {
              // (q, g, qbar, -q, g, -qbar)
              Vec4toMGQCD(p, 5, 3, 0, 2, 4, 1, PNEW);
              icol = 3; // T[ 2, 1, 5]T[ 3, 4, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 3, 0, 2, 4, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uuxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else cout << " MG4interface : unrecognised colour order "
                        << "for 4-quark + gg amplitude" << endl;
            }
            else if (pos1Sav == 0 && pos2Sav == 5) {
              // (-q, g, qbar, q, g, -qbar)
              Vec4toMGQCD(p, 5, 0, 3, 2, 4, 1, PNEW);
              icol = (sameFlav ? 8 : 2); // T[ 3, 1, 5]T[ 2, 4, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 0, 3, 2, 4, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uuxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          suux2ddxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 3) {
              // (q, g, -qbar, -q, g, qbar)
              Vec4toMGQCD(p, 2, 3, 0, 5, 4, 1, PNEW);
              icol = (sameFlav ? 10 : 4); // T[ 3, 1, 6]T[ 2, 4, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {2, 3, 0, 5, 4, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) suux2uuxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          suux2ddxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + gg amplitude" << endl;
          }
          // q q' -> q q' g g and qbar qbar' -> qbar qbar' g g
          else if (id1*id2 > 0 && id1 != 21 && id2 != 21) {
            if (sameFlav) cout << " MG4interface : unrecognised colour order "
                               << "for 4-quark + gg amplitude" << endl;
            if (pos1Sav == 0 && pos2Sav == 3) {
              // (-q, g, qbar, -q, g, qbar)
              Vec4toMGQCD(p, 0, 3, 5, 2, 1, 4, PNEW);
              icol = 2; // T[ 4, 1, 5]T[ 3, 2, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {0, 3, 5, 2, 1, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 5) {
              // (q, g, -qbar, q, g, -qbar)
              Vec4toMGQCD(p, 2, 5, 3, 0, 1, 4, PNEW);
              icol = 2; // T[ 4, 1, 5]T[ 3, 2, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {2, 5, 3, 0, 1, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + gg amplitude" << endl;
          }
          // q qbar' -> q qbar' g g and qbar q' -> qbar q' g g
          else if (id1*id2 < 0 && id1 != 21 && id2 != 21) {
            if (sameFlav) cout << " MG4interface : unrecognised colour order "
                               << "for 4-quark + gg amplitude" << endl;
            if (pos1Sav == 0 && pos2Sav == 2) {
              // (-q, g, -qbar, q, g, qbar)
              Vec4toMGQCD(p, 2, 0, 3, 5, 1, 4, PNEW);
              icol = 3; // T[ 2, 1, 5]T[ 3, 4, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {2, 0, 3, 5, 1, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 3 && pos2Sav == 5) {
              // (q, g, qbar, -q, g, -qbar)
              Vec4toMGQCD(p, 5, 3, 0, 2, 1, 4, PNEW);
              icol = 4; // T[ 2, 1, 6]T[ 3, 4, 5]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 3, 0, 2, 1, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxgg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + gg amplitude" << endl;
          }
          // q g -> q q qbar g and q g -> q q' qbar' g
          // qbar g -> qbar qbar q g and qbar g -> qbar qbar' q' g
          else if ( (id1 == 21 && id2 != 21 ) || (id2 == 21 && id1 != 21) ) {
            if (pos1Sav == 0 && pos2Sav == 1) {
              // (-q, -g, qbar, q, g, qbar)
              Vec4toMGQCD(p, 0, 1, 5, 2, 3, 4, PNEW);
              icol = (sameFlav ? 12 : 6); // T[ 4, 1, 2]T[ 3, 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {0, 1, 5, 2, 3, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 0 && pos2Sav == 4) {
              // (-q, g, qbar, q, -g, qbar)
              Vec4toMGQCD(p, 0, 4, 5, 2, 3, 1, PNEW);
              icol = (sameFlav ? 7 : 1); // T[ 4, 1, 6]T[ 3, 5, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {0, 4, 5, 2, 3, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 2) {
              // (q, -g, -qbar, q, g, qbar)
              Vec4toMGQCD(p, 2, 1, 3, 0, 5, 4, PNEW);
              icol = (sameFlav ? 12 : 6); // T[ 4, 1, 2]T[ 3, 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {2, 1, 3, 0, 5, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 3) {
              // (q, -g, qbar, -q, g, qbar)
              Vec4toMGQCD(p, 3, 1, 2, 5, 0, 4, PNEW);
              icol = (sameFlav ? 7 : 1); // T[ 4, 1, 6]T[ 3, 5, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {3, 1, 2, 5, 0, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 5) {
              // (q, -g, qbar, q, g, -qbar)
              Vec4toMGQCD(p, 5, 1, 0, 3, 2, 4, PNEW);
              icol = (sameFlav ? 7 : 1); // T[ 4, 1, 6]T[ 3, 5, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 1, 0, 3, 2, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 4) {
              // (q, g, -qbar, q, -g, qbar)
              Vec4toMGQCD(p, 2, 4, 3, 0, 5, 1, PNEW);
              icol = (sameFlav ? 7 : 1); // T[ 4, 1, 6]T[ 3, 5, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {2, 4, 3, 0, 5, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 3 && pos2Sav == 4) {
              // (q, g, qbar, -q, -g, qbar)
              Vec4toMGQCD(p, 3, 4, 2, 5, 0, 1, PNEW);
              icol = (sameFlav ? 12 : 6); // T[ 4, 1, 2]T[ 3, 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {3, 4, 2, 5, 0, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 4 && pos2Sav == 5) {
              // (q, g, qbar, q, -g, -qbar)
              Vec4toMGQCD(p, 5, 4, 0, 3, 2, 1, PNEW);
              icol = (sameFlav ? 12 : 6); // T[ 4, 1, 2]T[ 3, 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 4, 0, 3, 2, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sug2uuuxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sug2uddxg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + gg amplitude" << endl;
          }
          // g g -> q qbar q qbar and g g -> q qbar q' qbar'
          else if (id1 == 21 && id2 == 21) {
            if (pos1Sav == 1 && pos2Sav == 4) {
              // (q, -g, qbar, q, -g, qbar)
              Vec4toMGQCD(p, 1, 4, 0, 5, 3, 2, PNEW);
              icol = (sameFlav ? 4 : 4); // T[ 5, 4, 2]T[ 3, 6, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 4, 0, 5, 3, 2};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              if (sameFlav) sgg2uuxuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
              else          sgg2uuxddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 4-quark + gg amplitude" << endl;
          }
          else cout << " MG4interface : unrecognised colour order "
                    << "for 4-quark + gg amplitude" << endl;
        } // End 4 quarks + 2 unconnected gluons

        else cout << " MG4interface : unrecognised colour order "
                  << "for 4-quark amplitude" << endl;

      }

      // FOUR-QUARK AMPLITUDES ------ END

      // SIX-QUARK AMPLITUDES ------ START

      else if (nQuarkTot == 6 && nOut == 4) {

        // Check how many different flavours we have
        map<int,int> flavours;
        for (int i=1; i<=5; ++i) flavours[i] = 0;
        for (int i=0; i<nMom; ++i) flavours[abs(id[i])]++;
        // 0 if 0 or 1 quark pair, 1 if 2 quark pairs, 2 if 3 quark pais
        int countSameFlavour = 0;
        for (int i=1; i<=5; ++i)
          countSameFlavour = max( max((flavours[i]-2)/2,0) , countSameFlavour);

        double PNEW[6][4];
        int id1 = id[pos1Sav];
        int id2 = id[pos2Sav];

        // One quark flavour
        if (countSameFlavour == 2) {
          if (verbose >= 7) {
            cout << " 6-quark amplitude (one quark flavour) with II = " 
                 << pos1Sav << " " << pos2Sav << " idII = " << id1 << id2
                 << endl << " ";
            for (int i=0; i<nMom; ++i) cout << id[i] << " "; cout << endl;
          }
          // q q -> q q q qbar and qbar qbar -> qbar qbar q qbar
          if (id1 == id2) {
            if (pos1Sav == 0 && pos2Sav == 2) {
              // (-q, qbar, -q, qbar, q, qbar)
              Vec4toMGQCD(p, 0, 2, 1, 3, 5, 4, PNEW);
              icol = 4; // T[ 3, 1]T[ 4, 2]T[ 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {0, 2, 1, 3, 5, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uuuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 0 && pos2Sav == 4) {
              // (-q, qbar, q, qbar, -q, qbar)
              Vec4toMGQCD(p, 0, 4, 1, 3, 5, 2, PNEW);
              icol = 2; // T[ 3, 1]T[ 4, 6]T[ 5, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {0, 4, 1, 3, 5, 2};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uuuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 4) {
              // (q, qbar, -q, qbar, -q, qbar)
              Vec4toMGQCD(p, 2, 4, 1, 3, 5, 0, PNEW);
              icol = 3; // T[ 3, 6]T[ 4, 1]T[ 5, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {2, 4, 1, 3, 5, 0};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uuuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 3) {
              // (q, -qbar, q, -qbar, q, qbar)
              Vec4toMGQCD(p, 1, 3, 0, 2, 4, 5, PNEW);
              icol = 4; // T[ 3, 1]T[ 4, 2]T[ 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 3, 0, 2, 4, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uuuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 5) {
              // (q, -qbar, q, qbar, q, -qbar)
              Vec4toMGQCD(p, 1, 5, 0, 2, 4, 3, PNEW);
              icol = 2; // T[ 3, 1]T[ 4, 6]T[ 5, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 5, 0, 2, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uuuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 3 && pos2Sav == 5) {
              // (q, qbar, q, -qbar, q, -qbar)
              Vec4toMGQCD(p, 3, 5, 0, 2, 4, 1, PNEW);
              icol = 3; // T[ 3, 6]T[ 4, 1]T[ 5, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {3, 5, 0, 2, 4, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uuuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 6-quark amplitude" << endl;
          }
          // q qbar -> q qbar q qbar
          else if (id1 == -id2) {
            if (pos1Sav == 0 && pos2Sav == 1) {
              // (-q, -qbar, q, -qbar, q, qbar)
              Vec4toMGQCD(p, 1, 0, 2, 3, 4, 5, PNEW);
              icol = 4; // T[ 2, 1]T[ 3, 4]T[ 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 0, 2, 3, 4, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2uuxuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 0 && pos2Sav == 3) {
              // (-q, qbar, q, -qbar, q, qbar)
              Vec4toMGQCD(p, 3, 0, 2, 1, 4, 5, PNEW);
              icol = 2; // T[ 2, 4]T[ 3, 1]T[ 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {3, 0, 2, 1, 4, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2uuxuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 0 && pos2Sav == 5) {
              // (-q, qbar, q, qbar, q, -qbar)
              Vec4toMGQCD(p, 5, 0, 2, 1, 4, 3, PNEW);
              icol = 5; // T[ 2, 4]T[ 3, 6]T[ 5, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 0, 2, 1, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2uuxuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 2) {
              // (q, -qbar, -q, qbar, q, qbar)
              Vec4toMGQCD(p, 1, 2, 0, 3, 4, 5, PNEW);
              icol = 2; // T[ 2, 4]T[ 3, 1]T[ 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 2, 0, 3, 4, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2uuxuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 4) {
              // (q, -qbar, q, qbar, -q, qbar)
              Vec4toMGQCD(p, 1, 4, 0, 3, 2, 5, PNEW);
              icol = 3; // T[ 5, 4]T[ 3, 1]T[ 2, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 4, 0, 3, 2, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2uuxuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 3) {
              // (q, qbar, -q, -qbar, q, qbar)
              Vec4toMGQCD(p, 3, 2, 0, 1, 4, 5, PNEW);
              icol = 4; // T[ 2, 1]T[ 3, 4]T[ 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {3, 2, 0, 1, 4, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2uuxuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 5) {
              // (q, qbar, -q, qbar, q, -qbar)
              Vec4toMGQCD(p, 5, 2, 0, 1, 4, 3, PNEW);
              icol = 6; // T[ 5, 1]T[ 3, 4]T[ 2, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 2, 0, 1, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2uuxuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 3 && pos2Sav == 4) {
              // (q, qbar, q, -qbar, -q, qbar)
              Vec4toMGQCD(p, 3, 4, 0, 1, 2, 5, PNEW);
              icol = 6; // T[ 5, 1]T[ 3, 4]T[ 2, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {3, 4, 0, 1, 2, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2uuxuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 4 && pos2Sav == 5) {
              // (q, qbar, q, qbar, -q, -qbar)
              Vec4toMGQCD(p, 5, 4, 0, 1, 2, 3, PNEW);
              icol = 4; // T[ 2, 1]T[ 3, 4]T[ 5, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 4, 0, 1, 2, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2uuxuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 6-quark amplitude" << endl;
          }
          else cout << " MG4interface : unrecognised colour order "
                    << "for 6-quark amplitude" << endl;
        } // End one quark flavour

        // Two quark flavours
        else if (countSameFlavour == 1) {
          // Order as (q', qbar, q, qbar, q, qbar')
          int absFlav = 0;
          for (int i=1; i<=5; ++i) if (flavours[i] == 2) absFlav = i;
          pair<int,int> flavourPos; // quark first, antiquark second
          for (int i=0; i<nMom; ++i) { 
            if (abs(id[i]) == absFlav) {
              if (i % 2 == 0) // quarks sitting on even numbers
                flavourPos.first = i;
              else 
                flavourPos.second = i;
            }
          }
          vector<int> order = orderQQBqqbqqbQCD(flavourPos);
          vector<Vec4> pTmp   = p;
          vector<int>  polTmp = pol;
          for (int i=0; i<nMom; ++i) {
            p[i]   = pTmp[order[i]];
            pol[i] = polTmp[order[i]];
          }
          // Update position of incoming quarks
          for (int i=0; i<nMom; ++i) if (order[i] == pos1Sav) {
            pos1Sav = i;
            break;
          }
          for (int i=0; i<nMom; ++i) if (order[i] == pos2Sav) {
            pos2Sav = i;
            break;
          }
          if (pos2Sav < pos1Sav) {
            int posTmp = pos1Sav;
            pos1Sav = pos2Sav;
            pos2Sav = posTmp;
          }
          if (verbose >= 7) {
            cout << " 6-quark amplitude (two quark quark flavour) with II = " 
                 << pos1Sav << " " << pos2Sav << " idII = " << id1 << id2
                 << endl << " ";
            for (int i=0; i<nMom; ++i) cout << id[i] << " "; cout << endl;
          }
          // q q -> q q q' qbar' and qbar qbar -> qbar qbar q' qbar'
          if (id1 == id2) {
            if (pos1Sav == 1 && pos2Sav == 3) {
              // (q', -qbar, q, -qbar, q, qbar')
              Vec4toMGQCD(p, 1, 3, 2, 4, 0, 5, PNEW);
              icol = 3; // T[ 5, 11T[ 3, 2]T[ 4, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 3, 2, 4, 0, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uuddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 4) {
              // (q', qbar, -q, qbar, -q, qbar')
              Vec4toMGQCD(p, 2, 4, 1, 3, 5, 0, PNEW);
              icol = 4; // T[ 3, 6]T[ 4, 1]T[ 5, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {2, 4, 1, 3, 5, 0};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suu2uuddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 6-quark amplitude" << endl;
          }
          // q qbar -> q qbar q' qbar' and q' qbar' -> q qbar q qbar
          else if (id1 == -id2) {
            if (pos1Sav == 0 && pos2Sav == 5) {
              // (-q', qbar, q, qbar, q, -qbar')
              Vec4toMGQCD(p, 5, 0, 2, 1, 4, 3, PNEW);
              icol = 3; // T[ 2, 4]T[ 3, 6]T[ 5, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 0, 2, 1, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2ddxddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 2) {
              // (q', -qbar, -q, qbar, q, qbar')
              Vec4toMGQCD(p, 1, 2, 4, 3, 0, 5, PNEW);
              icol = 3; // T[ 2, 4]T[ 3, 6]T[ 5, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 2, 4, 3, 0, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2uuxddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 4) {
              // (q', -qbar, q, qbar, -q, qbar')
              Vec4toMGQCD(p, 1, 4, 2, 3, 0, 5, PNEW);
              icol = 1; // T[ 3, 4]T[ 2, 6]T[ 5, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 4, 2, 3, 0, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2uuxddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 3) {
              // (q', qbar, -q, -qbar, q, qbar')
              Vec4toMGQCD(p, 3, 2, 4, 1, 0, 5, PNEW);
              icol = 2; // T[ 2, 1]T[ 3, 6]T[ 5, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {3, 2, 4, 1, 0, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2uuxddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 3 && pos2Sav == 4) {
              // (q', qbar, q, -qbar, -q, qbar')
              Vec4toMGQCD(p, 3, 4, 2, 1, 0, 5, PNEW);
              icol = 4; // T[ 3, 1]T[ 2, 6]T[ 5, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {3, 4, 2, 1, 0, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2uuxddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 6-quark amplitude" << endl;
          }
          // q' q -> q' q q qbar and qbar' qbar -> qbar' qbar q qbar
          else if (id1*id2 > 0) {
            if (pos1Sav == 1 && pos2Sav == 5) {
              // (q', -qbar, q, qbar, q, -qbar')
              Vec4toMGQCD(p, 5, 1, 0, 2, 4, 3, PNEW);
              icol = 3; // T[ 4, 6]T[ 3, 2]T[ 5, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 1, 0, 2, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 3 && pos2Sav == 5) {
              // (q', qbar, q, -qbar, q, -qbar')
              Vec4toMGQCD(p, 5, 3, 0, 2, 4, 1, PNEW);
              icol = 2; // T[ 4, 2]T[ 3, 6]T[ 5, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 3, 0, 2, 4, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 0 && pos2Sav == 2) {
              // (-q', qbar, -q, qbar, q, qbar')
              Vec4toMGQCD(p, 0, 2, 5, 1, 3, 4, PNEW);
              icol = 4; // T[ 4, 1]T[ 5, 2]T[ 3, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {0, 2, 5, 1, 3, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 0 && pos2Sav == 4) {
              // (-q', qbar, q, qbar, -q, qbar')
              Vec4toMGQCD(p, 0, 4, 5, 1, 3, 2, PNEW);
              icol = 1; // T[ 4, 1]T[ 5, 6]T[ 3, 2]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {0, 4, 5, 1, 3, 2};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 6-quark amplitude" << endl;
          }
          // q qbar' -> q qbar' q qbar and q' qbar -> q' qbar q qbar
          else if (id1*id2 < 0) {
            if (pos1Sav == 0 && pos2Sav == 1) {
              // (-q', -qbar, q, qbar, q, qbar')
              Vec4toMGQCD(p, 1, 0, 2, 5, 4, 3, PNEW);
              icol = 4; // T[ 2, 1]T[ 3, 6]T[ 5, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 0, 2, 5, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 0 && pos2Sav == 3) {
              // (-q', qbar, q, -qbar, q, qbar')
              Vec4toMGQCD(p, 3, 0, 2, 5, 4, 1, PNEW);
              icol = 1; // T[ 2, 6]T[ 3, 1]T[ 5, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {3, 0, 2, 5, 4, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxuux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 5) {
              // (q', qbar, -q, qbar, q, -qbar')
              Vec4toMGQCD(p, 5, 2, 0, 1, 4, 3, PNEW);
              icol = 3; // T[ 2, 6]T[ 5, 1]T[ 3, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 2, 0, 1, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 4 && pos2Sav == 5) {
              // (q', qbar, q, qbar, -q, -qbar')
              Vec4toMGQCD(p, 5, 4, 0, 1, 2, 3, PNEW);
              icol = 2; // T[ 5, 6]T[ 2, 1]T[ 3, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 4, 0, 1, 2, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 6-quark amplitude" << endl;
          }
          else cout << " MG4interface : unrecognised colour order "
                    << "for 6-quark amplitude" << endl;
        } // End two quark flavours

        // Three quark flavours
        else if (countSameFlavour == 0) {
          // Order as (q, qbar', q', qbar'', q'', qbar)
          int absFlav = abs(id1); // pick random flavour
          pair<int,int> flavourPos; // quark first, antiquark second
          for (int i=0; i<nMom; ++i) { 
            if (abs(id[i]) == absFlav) {
              if (i % 2 == 0) // quarks sitting on even numbers
                flavourPos.first = i;
              else 
                flavourPos.second = i;
            }
          }
          vector<int> order = orderQQBqqbqqbQCD(flavourPos);
          vector<Vec4> pTmp   = p;
          vector<int>  polTmp = pol;
          for (int i=0; i<nMom; ++i) {
            p[i]   = pTmp[order[i]];
            pol[i] = polTmp[order[i]];
          }
          // Update position of incoming quarks
          for (int i=0; i<nMom; ++i) if (order[i] == pos1Sav) {
            pos1Sav = i;
            break;
          }
          for (int i=0; i<nMom; ++i) if (order[i] == pos2Sav) {
            pos2Sav = i;
            break;
          }
          if (pos2Sav < pos1Sav) {
            int posTmp = pos1Sav;
            pos1Sav = pos2Sav;
            pos2Sav = posTmp;
          }
          if (verbose >= 7) {
            cout << " 6-quark amplitude (three quark quark flavour) with II = " 
                 << pos1Sav << " " << pos2Sav << " idII = " << id1 << id2
                 << endl << " ";
            for (int i=0; i<nMom; ++i) cout << id[i] << " "; cout << endl;
          }
          // q qbar -> q' qbar' q'' qbar''
          if (id1 == -id2) {
            if (pos1Sav == 0 && pos2Sav == 5) {
              // (-q, qbar', q', qbar'', q'', -qbar)
              Vec4toMGQCD(p, 5, 0, 2, 1, 4, 3, PNEW);
              icol = 1; // T[ 5, 1]T[ 2, 4]T[ 3, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 0, 2, 1, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2ddxssx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 2) {
              // (q, -qbar', -q', qbar'', q'', qbar)
              Vec4toMGQCD(p, 1, 2, 0, 5, 4, 3, PNEW);
              icol = 2; // T[ 3, 1]T[ 5, 4]T[ 2, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 2, 0, 5, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2ddxssx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 3 && pos2Sav == 4) {
              // (q, qbar', q', -qbar'', -q'', qbar)
              Vec4toMGQCD(p, 3, 4, 2, 1, 0, 5, PNEW);
              icol = 2; // T[ 3, 1]T[ 5, 4]T[ 2, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {3, 4, 2, 1, 0, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              suux2ddxssx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 6-quark amplitude" << endl;
          }
          // q q' -> q q' q'' qbar'' and q qbar' -> q qbar' q'' qbar''
          else if (id1*id2 > 0) {
            if (pos1Sav == 1 && pos2Sav == 3) {
              // (q, -qbar', q', -qbar'', q'', qbar)
              Vec4toMGQCD(p, 1, 3, 2, 4, 0, 5, PNEW);
              icol = 1; // T[ 4, 6]T[ 3, 2]T[ 5, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 3, 2, 4, 0, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udssx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 5) {
              // (q, -qbar', q', qbar'', q'', -qbar)
              Vec4toMGQCD(p, 1, 5, 2, 0, 4, 3, PNEW);
              icol = 2; // T[ 5, 2]T[ 3, 6]T[ 4, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 5, 2, 0, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udssx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 3 && pos2Sav == 5) {
              // (q, qbar', q', -qbar'', q'', -qbar)
              Vec4toMGQCD(p, 3, 5, 4, 0, 2, 1, PNEW);
              icol = 1; // T[ 4, 6]T[ 3, 2]T[ 5, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {3, 5, 4, 0, 2, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udssx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 0 && pos2Sav == 2) {
              // (-q, qbar', -q', qbar'', q'', qbar)
              Vec4toMGQCD(p, 0, 2, 5, 1, 3, 4, PNEW);
              icol = 2; // T[ 5, 2]T[ 3, 6]T[ 4, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {0, 2, 5, 1, 3, 4};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udssx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 0 && pos2Sav == 4) {
              // (-q, qbar', q', qbar'', -q'', qbar)
              Vec4toMGQCD(p, 0, 4, 5, 3, 1, 2, PNEW);
              icol = 1; // T[ 4, 6]T[ 3, 2]T[ 5, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {0, 4, 5, 3, 1, 2};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udssx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 4) {
              // (q, qbar', -q', qbar'', -q'', qbar)
              Vec4toMGQCD(p, 2, 4, 1, 3, 5, 0, PNEW);
              icol = 2; // T[ 4, 1]T[ 5, 2]T[ 3, 6]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {2, 4, 1, 3, 5, 0};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sud2udssx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 6-quark amplitude" << endl;
          }
          // q qbar' -> q qbar' q'' qbar''
          else if (id1*id2 < 0) {
            if (pos1Sav == 0 && pos2Sav == 1) {
              // (-q, -qbar', q', qbar'', q'', qbar)
              Vec4toMGQCD(p, 1, 0, 2, 5, 4, 3, PNEW);
              icol = 2; // T[ 2, 1]T[ 3, 6]T[ 5, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 0, 2, 5, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxssx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 0 && pos2Sav == 3) {
              // (-q, qbar', q', -qbar'', q'', qbar)
              Vec4toMGQCD(p, 3, 0, 4, 5, 2, 1, PNEW);
              icol = 1; // T[ 3, 4]T[ 2, 6]T[ 5, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {3, 0, 4, 5, 2, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxssx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 1 && pos2Sav == 4) {
              // (q, -qbar', q', qbar'', -q'', qbar)
              Vec4toMGQCD(p, 1, 4, 2, 3, 0, 5, PNEW);
              icol = 1; // T[ 3, 4]T[ 2, 6]T[ 5, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {1, 4, 2, 3, 0, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxssx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 3) {
              // (q, qbar', -q', -qbar'', q'', qbar)
              Vec4toMGQCD(p, 3, 2, 4, 1, 0, 5, PNEW);
              icol = 2; // T[ 2, 1]T[ 3, 6]T[ 5, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {3, 2, 4, 1, 0, 5};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxssx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 2 && pos2Sav == 5) {
              // (q, qbar', -q', qbar'', q'', -qbar)
              Vec4toMGQCD(p, 5, 2, 0, 1, 4, 3, PNEW);
              icol = 1; // T[ 3, 4]T[ 2, 6]T[ 5, 1]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 2, 0, 1, 4, 3};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxssx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else if (pos1Sav == 4 && pos2Sav == 5) {
              // (q, qbar', q', qbar'', -q'', -qbar)
              Vec4toMGQCD(p, 5, 4, 0, 3, 2, 1, PNEW);
              icol = 2; // T[ 2, 1]T[ 3, 6]T[ 5, 4]
              if (HEL[0] > 0) {
                int mapVtoMG[6] = {5, 4, 0, 3, 2, 1};
                for (int iMG=nMom-1; iMG>=0; --iMG) {
                  HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
                  nHels *= 2;
                }
              }
              if (verbose >= 7) cout << " icol = " << icol << endl;
              if (verbose >= 5 && HEL[0] > 0)
                cout << " Using iHEL = " << HEL[0] << endl;
              sudx2udxssx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            }
            else cout << " MG4interface : unrecognised colour order "
                      << "for 6-quark amplitude" << endl;
          }
          else cout << " MG4interface : unrecognised colour order "
                    << "for 6-quark amplitude" << endl;
        } // End three quark flavours

        else cout << " MG4interface : unrecognised colour order "
                  << "for 6-quark amplitude" << endl;

      }

      // SIX-QUARK AMPLITUDES ------ END

      // Set masses back to original values
      coupl_mass_.bmass = bMassSav;
      coupl_mass_.cmass = cMassSav;
      coupl_mass_.smass = sMassSav;
      coupl_mass_.umass = uMassSav;
      coupl_mass_.dmass = dMassSav;

      // Add average over colours in the initial state
      if (id[pos1Sav]==21) MGME2*=1.0/8.0;
      else MGME2*=1.0/3.0;
      if (id[pos2Sav]==21) MGME2*=1.0/8.0;
      else MGME2*=1.0/3.0;

    }
    // end QCD 2->n
    //------------------------------------------------------------------------

    //------------------------------------------------------------------------
    // Drell-Yan + gluons. Only Polarised, only inputs are e+e- -> qqbar+gluons
    // Used for MHV checks
    else if (nIn == 2 && isEEtoJets) {
      // Initialise helicity-dependence
      int nHels = 1;
      if (isPolarised) HEL[0] = 1;
      else { 
        if (verbose>=1) printOut("MG4interface::ME2",
		      "EEtoJets must be polarised. ");
        return -1;
      }
      
      // Sanity check/here until we add e.g. DrellYan
      if (nLeptonIn != 2) {
	if (verbose>=1) printOut("MG4interface::ME2",
		 "2->n EW processes must be e+e- -> jets");
	return -1;
      }
      
      // create map to MG ordering (e+e- -> q gluons qBar)
      map<int,int> mapVtoMG;
      mapVtoMG[0] = ( (id[pos1Sav]>0) ? pos2Sav:pos1Sav );
      mapVtoMG[1] = ( (id[pos1Sav]<0) ? pos2Sav:pos1Sav );
      mapVtoMG[2] = ( (id[iQuark[0]]>0) ? iQuark[0]:iQuark[1] );
      int imG=2;
      for (int i = 0; i<(int)iGluon.size();i++) mapVtoMG[++imG] = iGluon[i];
      mapVtoMG[nMom-1] = ( (id[iQuark[1]]>0) ? iQuark[0]:iQuark[1] );
      
      //e+e- -> qqbar
      if (nOut == 2) {
        icol = 1; //Only 1 colour connection
        double PNEW[4][4];
        for (int iMG=nMom-1; iMG>=0; --iMG) {
          Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
          if (HEL[0] > 0) {
            HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
          }
          nHels *= 2;
        }
        //cout << "  Going for helicity amp " << HEL[0] << endl;
        if (abs(id[2] + 1) % 2 == 0) {swapAud();swapZud();}
        sexe2uux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        if (abs(id[2] + 1) % 2 == 0) {swapAud();swapZud();}
      }
      
      //e+e- -> q g qbar
      else if (nOut == 3) {
        icol = 1;
        double PNEW[4][4];
        for (int iMG=nMom-1; iMG>=0; --iMG) {
          Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
          if (HEL[0] > 0) {
            HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
          }
          nHels *= 2;
        }
//         cout << "  Going for helicity amp " << HEL[0] << endl;
        if (abs(id[2] + 1) % 2 == 0) {swapAud();swapZud();}
        sexe2ugux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        if (abs(id[2] + 1) % 2 == 0) {swapAud();swapZud();}
      }
      
      //e+e- -> q gg qbar
      else if (nOut == 4) {
        icol = 2;
        double PNEW[4][4];
        for (int iMG=nMom-1; iMG>=0; --iMG) {
          Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
          if (HEL[0] > 0) {
            HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
          }
          nHels *= 2;
        }
//         cout << "  Going for helicity amp " << HEL[0] << endl;
        if (abs(id[2] + 1) % 2 == 0) {swapAud();swapZud();}
        sexe2uggux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        if (abs(id[2] + 1) % 2 == 0) {swapAud();swapZud();}
      }
      
      //e+e- -> q ggg qbar
      else if (nOut == 5) {
        icol = 6;
        double PNEW[4][4];
        for (int iMG=nMom-1; iMG>=0; --iMG) {
          Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
          if (HEL[0] > 0) {
            HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
          }
          nHels *= 2;
        }
//         cout << "  Going for helicity amp " << HEL[0] << endl;
        if (abs(id[2] + 1) % 2 == 0) {swapAud();swapZud();}
        sexe2ugggux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        if (abs(id[2] + 1) % 2 == 0) {swapAud();swapZud();}
      }
      
      //e+e- -> q gggg qbar
      else if (nOut == 6) {
        icol = 24;
        double PNEW[8][4];
        for (int iMG=nMom-1; iMG>=0; --iMG) {
          Vec4toMG(p[mapVtoMG[iMG]],PNEW[iMG]);
          if (HEL[0] > 0) {
            HEL[0] += nHels*(int)(pol[mapVtoMG[iMG]]+1)/2;
          }
          nHels *= 2;
        }
//         cout << "  Going for helicity amp " << HEL[0] << endl;
        if (abs(id[2] + 1) % 2 == 0) {swapAud();swapZud();}
        sexe2uggggux_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        if (abs(id[2] + 1) % 2 == 0) {swapAud();swapZud();}
      }
    }
    // end Drell-Yan + gluons
    //------------------------------------------------------------------------

    // MEs for IF helicitiy ants checks
    else if ( (id[nMom-1]==11 && id[1]==11) || (id[nMom-1]==15 && id[1]==15) ) {

      // Massless only, setting all quark masses to zero
      double bMassSav   = coupl_mass_.bmass;
      coupl_mass_.bmass = 0.0;
      double cMassSav   = coupl_mass_.cmass;
      coupl_mass_.cmass = 0.0;
      double sMassSav   = coupl_mass_.smass;
      coupl_mass_.smass = 0.0;
      double dMassSav   = coupl_mass_.dmass;
      coupl_mass_.dmass = 0.0;
      double uMassSav   = coupl_mass_.umass;
      coupl_mass_.umass = 0.0;

      // Safety
      for (int ihel=0; ihel<4; ihel++) HEL[ihel] = -1;
      // Sum/Average over e-/ta- polarizations
      if (nOut == 2 && pol[0]!=9 && pol[2]!=9) {
        for (int iHel=0; iHel<4; iHel++) {
          HEL[iHel] = 1;
          HEL[iHel] += ( (pol[2] == -1) ? 0 : 2); // p-f
          HEL[iHel] += ( (pol[0] == -1) ? 0 : 8); // p-i
        }     // e-f // e-i
        HEL[0] += 0 + 0;
        HEL[1] += 0 + 4;
        HEL[2] += 1 + 0;
        HEL[3] += 1 + 4;
      } else if (nOut == 3 && pol[0]!=9  && pol[2]!=9 && pol[3]!=9) {
        for (int iHel=0; iHel<4; iHel++) {
          HEL[iHel] = 1;
          HEL[iHel] += ( (pol[3] == -1) ? 0 :  2); // p-f-2
          HEL[iHel] += ( (pol[2] == -1) ? 0 :  4); // p-f-1
          HEL[iHel] += ( (pol[0] == -1) ? 0 : 16); // p-i
        }     // e-f // e-i
        HEL[0] += 0 + 0;
        HEL[1] += 0 + 8;
        HEL[2] += 1 + 0;
        HEL[3] += 1 + 8;
      }
      //cout << " Using ihels = " << HEL[0] << " " << HEL[1] << " "
      //                          << HEL[2] << " " << HEL[3] << endl;

      // Xe -> Xe
      if (id[nMom-1]==11 && id[1]==11) {
        if (nOut == 2) {
          // Convert momenta to MadGraph format
          double PNEW[4][4];
          for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
          if (id[0] > 0) sde2de_(&colorDepth,&icol,PNEW,HEL,&MGME2);
          else           sdxe2dxe_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        } else if (nOut == 3) {
          // Convert momenta to MadGraph format
          double PNEW[5][4];
          for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
          if (id[0]==21) {
            if (id[2] > 0) sge2ddxe_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            else           sge2dxde_(&colorDepth,&icol,PNEW,HEL,&MGME2);
          } else {
            if (id[0] > 0) sde2dge_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            else           sdxe2dxge_(&colorDepth,&icol,PNEW,HEL,&MGME2);
          }
        }
      }

      // Xta -> Xta with Hq coupling
      else if (id[nMom-1]==15 && id[1]==15 && abs(id[2])==5) {
        // Zero Z quark and A quark couplings -> H only
        complex gzu0 = coupl_gauge_.gzu[0];
        complex gzu1 = coupl_gauge_.gzu[1];
        complex gzd0 = coupl_gauge_.gzd[0];
        complex gzd1 = coupl_gauge_.gzd[1];
        coupl_gauge_.gzu[0] = complex( 0.0, 0.0 );
        coupl_gauge_.gzu[1] = complex( 0.0, 0.0 );
        coupl_gauge_.gzd[0] = complex( 0.0, 0.0 );
        coupl_gauge_.gzd[1] = complex( 0.0, 0.0 );
        complex gau0 = coupl_gauge_.gau[0];
        complex gau1 = coupl_gauge_.gau[1];
        complex gad0 = coupl_gauge_.gad[0];
        complex gad1 = coupl_gauge_.gad[1];
        coupl_gauge_.gau[0] = complex( 0.0, 0.0 );
        coupl_gauge_.gau[1] = complex( 0.0, 0.0 );
        coupl_gauge_.gad[0] = complex( 0.0, 0.0 );
        coupl_gauge_.gad[1] = complex( 0.0, 0.0 );
        // Set H tau coupling to unity
        coupl_yuk_.ghtau[0] = complex( 1.0, 0.0 );
        coupl_yuk_.ghtau[1] = complex( 1.0, 0.0 );
        if (nOut == 2) {
          // Convert momenta to MadGraph format
          double PNEW[4][4];
          for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
          if (id[0] > 0) sbta2bta_(&colorDepth,&icol,PNEW,HEL,&MGME2);
          else           sbxta2bxta_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        } else if (nOut == 3) {
          // Convert momenta to MadGraph format
          double PNEW[5][4];
          for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
          if (id[0]==21) {
            if (id[2] > 0) sgta2bbxta_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            else           sgta2bxbta_(&colorDepth,&icol,PNEW,HEL,&MGME2);
          } else {
            if (id[0] > 0) sbta2bgta_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            else           sbxta2bxgta_(&colorDepth,&icol,PNEW,HEL,&MGME2);
          }
        }
        // Put couplings back
        coupl_gauge_.gzu[0] = gzu0;
        coupl_gauge_.gzu[1] = gzu1;
        coupl_gauge_.gzd[0] = gzd0;
        coupl_gauge_.gzd[1] = gzd1;
        coupl_gauge_.gau[0] = gau0;
        coupl_gauge_.gau[1] = gau1;
        coupl_gauge_.gad[0] = gad0;
        coupl_gauge_.gad[1] = gad1;
      }

      // Xta -> Xta with Hg coupling
      else if (id[nMom-1]==15 && id[1]==15) {
        // Set H tau coupling to unity
        coupl_yuk_.ghtau[0] = complex( 1.0, 0.0 );
        coupl_yuk_.ghtau[1] = complex( 1.0, 0.0 );
        if (nOut == 2) {
          // Convert momenta to MadGraph format
          double PNEW[4][4];
          for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
          sgta2gta_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        } else if (nOut == 3) {
          // Convert momenta to MadGraph format
          double PNEW[5][4];
          for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
          if (id[0]==21 && id[2]==21) {
            if (checkIS) icol = -1;
            sgta2ggta_(&colorDepth,&icol,PNEW,HEL,&MGME2);
          } else if (id[0]==21) {
            if (id[2] > 0) sgta2ddxta_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            else           sgta2dxdta_(&colorDepth,&icol,PNEW,HEL,&MGME2);
          } else {
            if (id[0] > 0) sdta2gdta_(&colorDepth,&icol,PNEW,HEL,&MGME2);
            else           sdxta2gdxta_(&colorDepth,&icol,PNEW,HEL,&MGME2);
          }
        }
      }

      // Set masses back to original values
      coupl_mass_.bmass = bMassSav;
      coupl_mass_.cmass = cMassSav;
      coupl_mass_.smass = sMassSav;
      coupl_mass_.umass = uMassSav;
      coupl_mass_.dmass = dMassSav;

      // Add average over colours in the initial state
      if (id[0]==21) MGME2*=1.0/8.0;
      else MGME2*=1.0/3.0;

    }

    //------------------------------------------------------------------------
    // Z production
    else if (nIn == 2 && idN == 23) {

      // Massless only, setting all quark masses to zero
      double bMassSav   = coupl_mass_.bmass;
      coupl_mass_.bmass = 0.0;
      double cMassSav   = coupl_mass_.cmass;
      coupl_mass_.cmass = 0.0;
      double sMassSav   = coupl_mass_.smass;
      coupl_mass_.smass = 0.0;
      double dMassSav   = coupl_mass_.dmass;
      coupl_mass_.dmass = 0.0;
      double uMassSav   = coupl_mass_.umass;
      coupl_mass_.umass = 0.0;

      // Safety
      for (int iHel=0; iHel<12; iHel++) HELsbp[iHel] = -1;
      // Sum over Z polarizations
      bool sumZpol  = (pol[nMom-1] == 9);
      // Sum and average over incoming polarizations
      bool avIncome = (pol[pos1Sav] == 9 && pol[pos2Sav] == 9);
      // Average factor for incoming polarization (+- for partons)
      nPolIn = (avIncome ? 4 : 1);
      // NOTE not checking for sum over incoming, but explicit Z polarization!

      // Ensure ME evaluated on shell
      const double mSys = p[nMom-1].mCalc();
      coupl_mass_.zmass = mSys;

      // pp->Z
      if (nOut == 1 && id[0]!=21 && id[1]!=21 && abs(id[0])==abs(id[1])) {
        // 0) Default assumption: fully unpolarized
        // 1) Sum over Z polarization, but explicit polarizations overwise
        if (sumZpol && !avIncome) {
          int iHelNow = 1 + ((pol[0] == -1) ? 0 : 3) + ((pol[1] == -1) ? 0 : 6);
          for (int iHel=0; iHel<3; iHel++) HELsbp[iHel] = iHelNow+iHel;
          if (verbose >= 5) printHels(HELsbp);
        }
        // 2) Sum over incoming and Z polarizations
        else if (avIncome && isPolarised) {
          // Nothing to do here as fully unpolarized
        }
        // Boost into rest frame of the Z
        p[2].e(mSys);     p[2].px(0.0); p[2].py(0.0); p[2].pz(0.0);
        p[0].e(mSys/2.0); p[0].px(0.0); p[0].py(0.0); p[0].pz(mSys/2.0);
        p[1].e(mSys/2.0); p[1].px(0.0); p[1].py(0.0); p[1].pz(-mSys/2.0);
        // Convert momenta to MadGraph format
        double PNEW[3][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        // Swap if necessary
        if (abs(id[0]) % 2 == 0) swapZud();
        sddx2z_(&colorDepth,&icol,PNEW,HELsbp,&MGME2);
        // Swap back if necessary
        if (abs(id[0]) % 2 == 0) swapZud();
      }

      // pp->jZ
      else if (nOut == 2) {
        // Find order
        int order[4] = { -1, -1, -1, -1 };
        if (pos1Sav == 0 && pos2Sav == 1) {
          // (-q, -g, qbar, Z)
          // MG order 1, 0, 2
          order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 0;
        } else if (pos1Sav == 1 && pos2Sav == 2) {
          // (q, -g, -qbar, Z)
          // MG order 1, 2, 0
          order[0] = 1; order[1] = 2; order[2] = 0; order[3] = 1;
        } else if (pos1Sav == 0 && pos2Sav == 2) {
          // (-q, g, -qbar, Z)
          // MG order 2, 0, 1
          order[0] = 2; order[1] = 0; order[2] = 1; order[3] = 2;
        }
        // Convert momenta to MadGraph format
        double PNEW[4][4];
        Vec4toMG(p, order[0], order[1], order[2], PNEW);
        if (verbose >= 5) cout << " Using order = " << order[3] << endl;
        // Swap if necessary
        if (abs(id[0]) % 2 == 0) swapZud();
        // 0) Default assumption: fully unpolarized
        // 1) Sum over Z polarization, but explicit polarizations overwise
        if (sumZpol && !avIncome) {
          int polMap[3] = { 12, 6, 3 };
          int iHelNow   = 1;
          for (int iHel=0; iHel<3; iHel++)
            iHelNow += ((pol[order[iHel]] == -1) ? 0 : polMap[iHel]);
          for (int iHel=0; iHel<3; iHel++) HELsbp[iHel] = iHelNow+(iHel%3);
          if (verbose >= 5) printHels(HELsbp);
        }
        // 2) Sum over incoming and Z polarizations
        else if (avIncome && isPolarised) {
          int polInt2 = ((pol[order[2]] == -1) ? 0 : 3);
          for (int iHel=0; iHel<12; iHel++) {
            HELsbp[iHel] = 1 + iHel + polInt2;
            if ((iHel%3) == 2) polInt2 += 3;
          }
          if (verbose >= 5) printHels(HELsbp);
        }
        if (order[3] == 0)
          sgdx2dxz_(&colorDepth,&icol,PNEW,HELsbp,&MGME2); // T[ 1, 2, 0]
        else if (order[3] == 1)
          sgd2dz_(&colorDepth,&icol,PNEW,HELsbp,&MGME2); // T[ 2, 1, 0]
        else if (order[3] == 2)
          sddx2gz_(&colorDepth,&icol,PNEW,HELsbp,&MGME2); // T[ 1, 0, 2]
        // Swap back if necessary
        if (abs(id[0]) % 2 == 0) swapZud();
      }

      // pp->jjZ
      else if (nOut == 3) {
        // Find order
        int  order[5] = { -1, -1, -1, -1, -1 };
        bool needSwap = false, sameFlav = false, sameType = false;
        if (nGluonTot == 2) {
          // Momenta and ids are now given in colour order q g g qbar
          if (abs(id[0]) % 2 == 0) needSwap = true;
          if (pos1Sav == 0 && pos2Sav == 1) {
            // (-q, -g, g, qbar, Z)
            // MG order 1, 0, 2, 3
            order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 3; order[4] = 0;
          } else if (pos1Sav == 0 && pos2Sav == 2) {
            // (-q, g, -g, qbar, Z)
            // MG order 2, 0, 1, 3
            order[0] = 2; order[1] = 0; order[2] = 1; order[3] = 3; order[4] = 1;
          } else if (pos1Sav == 0 && pos2Sav == 3) {
            // (-q, g, g, -qbar, Z)
            // MG order 3, 0, 1, 2
            order[0] = 3; order[1] = 0; order[2] = 1; order[3] = 2; order[4] = 2;
          } else if (pos1Sav == 1 && pos2Sav == 2) {
            // (q, -g, -g, qbar, Z)
            // MG order 1, 2, 0, 3
            order[0] = 1; order[1] = 2; order[2] = 0; order[3] = 3; order[4] = 3;
          } else if (pos1Sav== 1 && pos2Sav == 3) {
            // (q, -g, g, -qbar, Z)
            // MG order 1, 3, 2, 0
            order[0] = 1; order[1] = 3; order[2] = 2; order[3] = 0; order[4] = 4;
          } else if (pos1Sav == 2 && pos2Sav == 3) {
            // (q, g, -g, -qbar, Z)
            // MG order 2, 3, 1, 0
            order[0] = 2; order[1] = 3; order[2] = 1; order[3] = 0; order[4] = 5;
          }
        } else if (nGluonTot == 0) {
          // Momenta and ids are now given in order q qbar q qbar
          // Check if all quarks have the same flavour / same type (u/d)
          if ( (abs(id[0])==abs(id[1])) && (abs(id[0])==abs(id[2])) && 
               (abs(id[0])==abs(id[3])) ) sameFlav = true;
          if ( (abs(id[0])%2==1) && (abs(id[1])%2==1) && (abs(id[2])%2==1) && 
               (abs(id[3])%2==1) ) sameType = true;
          if ( (abs(id[0])%2==0) && (abs(id[1])%2==0) && (abs(id[2])%2==0) && 
               (abs(id[3])%2==0) ) sameType = true;
          if (pos1Sav == 1 && pos2Sav == 3) {
            if (abs(id[pos1Sav]) % 2 == 0) needSwap = true;
            // (q, -qbar, q, -qbar, Z)
            // MG order 1, 3, 2, 0
            order[0] = 1; order[1] = 3; order[2] = 2; order[3] = 0; order[4] = 6;
          }
          else if (pos1Sav == 0 && pos2Sav == 2) {
            if (abs(id[pos1Sav]) % 2 == 0) needSwap = true;
            // (-q, qbar, -q, qbar, Z)
            // MG order 0, 2, 3, 1
            order[0] = 0; order[1] = 2; order[2] = 3; order[3] = 1; order[4] = 7;
          }
          else if (pos1Sav == 0 && pos2Sav == 1) {
            if (abs(id[pos2Sav]) % 2 == 0) needSwap = true;
            // (-q, -qbar, q, qbar, Z)
            // MG order 1, 0, 2, 3
            order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 3; order[4] = 8;
          }
          else if (pos1Sav == 2 && pos2Sav == 3) {
            if (abs(id[pos2Sav]) % 2 == 0) needSwap = true;
            // (q, qbar, -q, -qbar, Z)
            // MG order 3, 2, 0, 1
            order[0] = 3; order[1] = 2; order[2] = 0; order[3] = 1; order[4] = 9;
          }
          else if (pos1Sav == 0 && pos2Sav == 3) {
            if (abs(id[pos2Sav]) % 2 == 0) needSwap = true;
            // (-q, qbar, q, -qbar, Z)
            // MG order 3, 0, 2, 1
            order[0] = 3; order[1] = 0; order[2] = 2; order[3] = 1; order[4] = 10;
          }
          else if (pos1Sav == 1 && pos2Sav == 2) {
            if (abs(id[pos1Sav]) % 2 == 0) needSwap = true;
            // (q, -qbar, -q, qbar, Z)
            // MG order 1, 2, 0, 3
            order[0] = 1; order[1] = 2; order[2] = 0; order[3] = 3; order[4] = 11;
          }
        }
        // Convert momenta to MadGraph format
        double PNEW[5][4];
        Vec4toMG(p, order[0], order[1], order[2], order[3], PNEW);
        if (verbose >= 5) cout << " Using order = " << order[4] << endl;
        // Swap if necessary
        if (needSwap) swapZud();
        // Use -pol, eg for gqbar -> gqbarZ (we only have gq -> gqZ)
        if (order[4]==0 || order[4]==1 || order[4]==7)
          for (int i=0; i<nMom; i++) if (abs(pol[i]) == 1) pol[i] = -pol[i];
        // 0) Default assumption: fully unpolarized
        // 1) Sum over Z polarization, but explicit polarizations overwise
        if (sumZpol && !avIncome) {
          int polMap[4] = { 24, 12, 6, 3 };
          int iHelNow   = 1;
          for (int iHel=0; iHel<4; iHel++)
            iHelNow += ((pol[order[iHel]] == -1) ? 0 : polMap[iHel]);
          for (int iHel=0; iHel<3; iHel++) HELsbp[iHel] = iHelNow+(iHel%3);
          if (verbose >= 5) printHels(HELsbp);
        }
        // 2) Sum over incoming and Z polarizations
        else if (avIncome && isPolarised) {
          int polInt2 = ((pol[order[2]] == -1) ? 0 : 6);
          int polInt3 = ((pol[order[3]] == -1) ? 0 : 3);
          for (int iHel=0; iHel<12; iHel++) {
            HELsbp[iHel] = 1 + iHel + polInt2 + polInt3;
            if ((iHel%3) == 2) polInt2 += 9;
          }
          if (verbose >= 5) printHels(HELsbp);
        }
        // Different color flows
        int icol1 = 1;
        int icol2 = 2;
        if (order[4] == 0)
          sgd2gdz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 1, 2, 0]
        else if (order[4] == 1)
          sgd2gdz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 1, 0, 2]
        else if (order[4] == 2)
          sddx2ggz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 0, 2, 3]
        else if (order[4] == 3)
          sgg2ddxz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 3, 0, 1]
        else if (order[4] == 4)
          sgd2gdz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 1, 0, 2]
        else if (order[4] == 5)
          sgd2gdz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 1, 2, 0]
        else if (order[4] == 6) {
               if (!sameFlav &&  sameType) sds2dsz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 1]
          else if (!sameFlav && !sameType) sdu2duz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 1]
          else if ( sameFlav &&  sameType) sdd2ddz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 1]
        } else if (order[4] == 7) {
               if (!sameFlav &&  sameType) sds2dsz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 1]
          else if (!sameFlav && !sameType) sdu2duz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 1]
          else if ( sameFlav &&  sameType) sdd2ddz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 1]
        } else if (order[4] == 8) {
               if (!sameFlav &&  sameType) sdsx2dsxz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0]
          else if (!sameFlav && !sameType) sdux2duxz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0]
          else if ( sameFlav &&  sameType) sddx2ddxz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0]
        } else if (order[4] == 9) {
               if (!sameFlav &&  sameType) sdsx2dsxz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0]
          else if (!sameFlav && !sameType) sdux2duxz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0]
          else if ( sameFlav &&  sameType) sddx2ddxz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0]
        } else if (order[4] == 10) {
               if (!sameFlav &&  sameType) sddx2ssxz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3]
          else if (!sameFlav && !sameType) sddx2uuxz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3]
          else if ( sameFlav &&  sameType) sddx2ddxz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3]
        } else if (order[4] == 11) {
               if (!sameFlav &&  sameType) sddx2ssxz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3]
          else if (!sameFlav && !sameType) sddx2uuxz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3]
          else if ( sameFlav &&  sameType) sddx2ddxz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3]
        }
        // Swap back if necessary
        if (needSwap) swapZud();
      }

      // pp->jjjZ
      else if (nOut == 4) {
        // Find order
        int  order[6] = { -1, -1, -1, -1, -1, -1 };
        bool needSwap = false, sameFlav = false, sameType = false;
        if (nGluonTot == 3) {
          // Momenta and ids are now given in colour order q g g g qbar
          if (abs(id[0]) % 2 == 0) needSwap = true;
          if (pos1Sav == 0 && pos2Sav == 4) {
            // (-q, g, g, g, -qbar, Z)
            // MG order 4, 0, 1, 2, 3
            order[0] = 4; order[1] = 0; order[2] = 1; order[3] = 2; order[4] = 3; order[5] = 0;
          }
          else if (pos1Sav == 1 && pos2Sav == 2) {
            // (q, -g, -g, g, qbar, Z)
            // MG order 1, 2, 0, 4, 3
            order[0] = 1; order[1] = 2; order[2] = 0; order[3] = 4; order[4] = 3; order[5] = 1;
          }
          else if (pos1Sav == 1 && pos2Sav == 3) {
            // (q, -g, g, -g, qbar, Z)
            // MG order 1, 3, 0, 4, 2
            order[0] = 1; order[1] = 3; order[2] = 0; order[3] = 4; order[4] = 2; order[5] = 2;
          }
          else if (pos1Sav == 2 && pos2Sav == 3) {
            // (q, g, -g, -g, qbar, Z)
            // MG order 2, 3, 0, 4, 1
            order[0] = 2; order[1] = 3; order[2] = 0; order[3] = 4; order[4] = 1; order[5] = 3;
          }
          else if (pos1Sav == 3 && pos2Sav == 4) {
            // (q, g, g, -g, -qbar, Z)
            // MG order 4, 3, 0, 1, 2
            order[0] = 4; order[1] = 3; order[2] = 0; order[3] = 1; order[4] = 2; order[5] = 4;
          }
          else if (pos1Sav == 2 && pos2Sav == 4) {
            // (q, g, -g, g, -qbar, Z)
            // MG order 4, 2, 0, 3, 1
            order[0] = 4; order[1] = 2; order[2] = 0; order[3] = 3; order[4] = 1; order[5] = 5;
          }
          else if (pos1Sav == 1 && pos2Sav == 4) {
            // (q, -g, g, g, -qbar, Z)
            // MG order 4, 1, 0, 2, 3
            order[0] = 4; order[1] = 1; order[2] = 0; order[3] = 2; order[4] = 3; order[5] = 6;
          }
          else if (pos1Sav == 0 && pos2Sav == 1) {
            // (-q, -g, g, g, qbar, Z)
            // MG order 0, 1, 4, 3, 2
            order[0] = 0; order[1] = 1; order[2] = 4; order[3] = 3; order[4] = 2; order[5] = 7;
          }
          else if (pos1Sav == 0 && pos2Sav == 2) {
            // (-q, g, -g, g, qbar, Z)
            // MG order 0, 2, 4, 1, 3
            order[0] = 0; order[1] = 2; order[2] = 4; order[3] = 1; order[4] = 3; order[5] = 8;
          }
          else if (pos1Sav == 0 && pos2Sav == 3) {
            // (-q, g, g, -g, qbar, Z)
            // MG order 0, 3, 4, 2, 1
            order[0] = 0; order[1] = 3; order[2] = 4; order[3] = 2; order[4] = 1; order[5] = 9;
          }
        } else if (nGluonTot == 1) {
          // Momenta and ids are now given in colour order q qbar q g qbar
          // Check if all quarks have the same flavour / same type (u/d)
          vector<int> idsPosCheck; idsPosCheck.clear(); idsPosCheck.resize(0);
          for (int i=0; i<(int)id.size(); i++) if (id[i]<20) idsPosCheck.push_back(i);
          if ( (abs(idsPosCheck[0])==abs(idsPosCheck[1])) && 
               (abs(idsPosCheck[0])==abs(idsPosCheck[2])) && 
               (abs(idsPosCheck[0])==abs(idsPosCheck[3])) ) sameFlav = true;
          if ( (abs(idsPosCheck[0])%2==1) && (abs(idsPosCheck[1])%2==1) && 
               (abs(idsPosCheck[2])%2==1) && (abs(idsPosCheck[2])%2==1) ) sameType = true;
          if ( (abs(idsPosCheck[0])%2==0) && (abs(idsPosCheck[1])%2==0) && 
               (abs(idsPosCheck[2])%2==0) && (abs(idsPosCheck[2])%2==0) ) sameType = true;
          if (pos1Sav == 1 && pos2Sav == 3) {
            if (abs(id[pos1Sav]) % 2 == 0) needSwap = true;
            // (q, -qbar, q, -g, qbar, Z)
            // MG order 1, 3, 2, 0, 4
            order[0] = 1; order[1] = 3; order[2] = 2; order[3] = 0; order[4] = 4; order[5] = 10;
          }
          else if (pos1Sav == 3 && pos2Sav == 4) {
            if (abs(id[pos2Sav]) % 2 == 0) needSwap = true;
            // (q, qbar, q, -g, -qbar, Z)
            // MG order 4, 3, 0, 2, 1
            order[0] = 4; order[1] = 3; order[2] = 0; order[3] = 2; order[4] = 1; order[5] = 11;
          }
          else if (pos1Sav == 0 && pos2Sav == 3) {
            if (abs(id[pos1Sav]) % 2 == 0) needSwap = true;
            // (-q, qbar, q, -g, qbar, Z)
            // MG order 0, 3, 4, 2, 1
            order[0] = 0; order[1] = 3; order[2] = 4; order[3] = 2; order[4] = 1; order[5] = 12;
          }
          else if (pos1Sav == 2 && pos2Sav == 3) {
            if (abs(id[pos1Sav]) % 2 == 0) needSwap = true;
            // (q, qbar, -q, -g, qbar, Z)
            // MG order 2, 3, 1, 0, 4
            order[0] = 2; order[1] = 3; order[2] = 1; order[3] = 0; order[4] = 4; order[5] = 13;
          }
          else if (pos1Sav == 0 && pos2Sav == 4) {
            if (abs(id[pos2Sav]) % 2 == 0) needSwap = true;
            // (-q, qbar, q, g, -qbar, Z)
            // MG order 4, 0, 2, 1, 3
            order[0] = 4; order[1] = 0; order[2] = 2; order[3] = 1; order[4] = 3; order[5] = 14;
          }
          else if (pos1Sav == 1 && pos2Sav == 2) {
            if (abs(id[pos1Sav]) % 2 == 0) needSwap = true;
            // (q, -qbar, -q, g, qbar, Z)
            // MG order 1, 2, 0, 4, 3
            order[0] = 1; order[1] = 2; order[2] = 0; order[3] = 4; order[4] = 3; order[5] = 15;
          }
          else if (pos1Sav == 2 && pos2Sav == 4) {
            if (abs(id[pos2Sav]) % 2 == 0) needSwap = true;
            // (q, qbar, -q, g, -qbar, Z)
            // MG order 4, 2, 0, 1, 3
            order[0] = 4; order[1] = 2; order[2] = 0; order[3] = 1; order[4] = 3; order[5] = 16;
          }
          else if (pos1Sav == 0 && pos2Sav == 1) {
            if (abs(id[pos2Sav]) % 2 == 0) needSwap = true;
            // (-q, -qbar, q, g, qbar, Z)
            // MG order 1, 0, 2, 4, 3
            order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 4; order[4] = 3; order[5] = 17;
          }
          else if (pos1Sav == 1 && pos2Sav == 4) {
            if (abs(id[pos2Sav]) % 2 == 0) needSwap = true;
            // (q, -qbar, q, g, -qbar, Z)
            // MG order 4, 1, 0, 2, 3
            order[0] = 4; order[1] = 1; order[2] = 0; order[3] = 2; order[4] = 3; order[5] = 18;
          }
          else if (pos1Sav == 0 && pos2Sav == 2) {
            if (abs(id[pos2Sav]) % 2 == 0) needSwap = true;
            // (-q, qbar, -q, g, qbar, Z)
            // MG order 2, 0, 1, 4, 3
            order[0] = 2; order[1] = 0; order[2] = 1; order[3] = 4; order[4] = 3; order[5] = 19;
          }
        }
        // Convert momenta to MadGraph format
        double PNEW[6][4];
        Vec4toMG(p, order[0], order[1], order[2], order[3], order[4], PNEW);
        if (verbose >= 5) cout << " Using order = " << order[5] << endl;
        // Swap if necessary
        if (needSwap) swapZud();
        // Use -pol, eg for gqbar -> ggqbarZ (we only have gq -> ggqZ)
        if (order[5]==7 || order[5]==8 || order[5]==9 || order[5]==19)
          for (int i=0; i<nMom; i++) if (abs(pol[i]) == 1) pol[i] = -pol[i];
        // 0) Default assumption: fully unpolarized
        // 1) Sum over Z polarization, but explicit polarizations overwise
        if (sumZpol && !avIncome) {
          int polMap[5] = { 48, 24, 12, 6, 3 };
          int iHelNow   = 1;
          for (int iHel=0; iHel<5; iHel++)
            iHelNow += ((pol[order[iHel]] == -1) ? 0 : polMap[iHel]);
          for (int iHel=0; iHel<3; iHel++) HELsbp[iHel] = iHelNow+(iHel%3);
          if (verbose >= 5) printHels(HELsbp);
        }
        // 2) Sum over incoming and Z polarizations
        else if (avIncome && isPolarised) {
          int polInt2 = ((pol[order[2]] == -1) ? 0 : 12);
          int polInt3 = ((pol[order[3]] == -1) ? 0 :  6);
          int polInt4 = ((pol[order[4]] == -1) ? 0 :  3);
          for (int iHel=0; iHel<12; iHel++) {
            HELsbp[iHel] = 1 + iHel + polInt2 + polInt3 + polInt4;
            if ((iHel%3) == 2) polInt2 += 21;
          }
          if (verbose >= 5) printHels(HELsbp);
        }
        // Different color flows
        int icol1 = 1;
        int icol2 = 2;
        int icol3 = 3;
        int icol4 = 4;
        int icol5 = 5;
        int icol6 = 6;
        if (order[5] == 0)
          sddx2gggz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 0, 2, 3, 4]
        else if (order[5] == 1)
          sgg2ddxgz_(&colorDepth,&icol6,PNEW,HELsbp,&MGME2); // T[ 2, 3, 0, 1, 4]
        else if (order[5] == 2)
          sgg2ddxgz_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 2, 3, 0, 4, 1]
        else if (order[5] == 3)
          sgg2ddxgz_(&colorDepth,&icol5,PNEW,HELsbp,&MGME2); // T[ 2, 3, 4, 0, 1]
        else if (order[5] == 4)
          sdg2dggz_(&colorDepth,&icol5,PNEW,HELsbp,&MGME2); // T[ 2, 0, 3, 4, 1]
        else if (order[5] == 5)
          sdg2dggz_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 2, 0, 4, 1, 3]
        else if (order[5] == 6)
          sdg2dggz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 0, 1, 3, 4]
        else if (order[5] == 7)
          sdg2dggz_(&colorDepth,&icol5,PNEW,HELsbp,&MGME2); // T[ 2, 0, 3, 4, 1]
        else if (order[5] == 8)
          sdg2dggz_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 2, 0, 4, 1, 3]
        else if (order[5] == 9)
          sdg2dggz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 0, 1, 3, 4]
        else if (order[5] == 10) {
               if (!sameFlav &&  sameType) sdg2dssxz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 4, 1]
          else if (!sameFlav && !sameType) sdg2duuxz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 4, 1]
          else if ( sameFlav &&  sameType) sdg2dddxz_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 4, 1]
        } else if (order[5] == 11) {
               if (!sameFlav &&  sameType) sdg2dssxz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 4]T[ 3, 0, 1]
          else if (!sameFlav && !sameType) sdg2duuxz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 4]T[ 3, 0, 1]
          else if ( sameFlav &&  sameType) sdg2dddxz_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 2, 4]T[ 3, 0, 1]
        } else if (order[5] == 12) {
               if (!sameFlav &&  sameType) sdxg2dxssxz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 0, 4]T[ 3, 2, 1]
          else if (!sameFlav && !sameType) sdxg2dxuuxz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 0, 4]T[ 3, 2, 1]
          else if ( sameFlav &&  sameType) sdxg2dxddxz_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 0, 4]T[ 3, 2, 1]
        } else if (order[5] == 13) {
               if (!sameFlav &&  sameType) sdxg2dxssxz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 2]T[ 0, 4, 1]
          else if (!sameFlav && !sameType) sdxg2dxuuxz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 2]T[ 0, 4, 1]
          else if ( sameFlav &&  sameType) sdxg2dxddxz_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 3, 2]T[ 0, 4, 1]
        } else if (order[5] == 14) {
               if (!sameFlav &&  sameType) sddx2ssxgz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 3]T[ 2, 0, 4]
          else if (!sameFlav && !sameType) sddx2uuxgz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 3]T[ 2, 0, 4]
          else if ( sameFlav &&  sameType) sddx2ddxgz_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 1, 3]T[ 2, 0, 4]
        } else if (order[5] == 15) {
               if (!sameFlav &&  sameType) sddx2ssxgz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3, 4]
          else if (!sameFlav && !sameType) sddx2uuxgz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3, 4]
          else if ( sameFlav &&  sameType) sddx2ddxgz_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3, 4]
        } else if (order[5] == 16) {
               if (!sameFlav &&  sameType) sdsx2dsxgz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0, 4]
          else if (!sameFlav && !sameType) sdux2duxgz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0, 4]
          else if ( sameFlav &&  sameType) sddx2ddxgz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0, 4]
        } else if (order[5] == 17) {
               if (!sameFlav &&  sameType) sdsx2dsxgz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 0]T[ 2, 3, 4]
          else if (!sameFlav && !sameType) sdux2duxgz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 0]T[ 2, 3, 4]
          else if ( sameFlav &&  sameType) sddx2ddxgz_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 0]T[ 2, 3, 4]
        } else if (order[5] == 18) {
               if (!sameFlav &&  sameType) sds2dsgz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 1]T[ 3, 0, 4]
          else if (!sameFlav && !sameType) sdu2dugz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 1]T[ 3, 0, 4]
          else if ( sameFlav &&  sameType) sdd2ddgz_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 2, 1]T[ 3, 0, 4]
        } else if (order[5] == 19) {
               if (!sameFlav &&  sameType) sds2dsgz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 1]T[ 3, 0, 4]
          else if (!sameFlav && !sameType) sdu2dugz_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 1]T[ 3, 0, 4]
          else if ( sameFlav &&  sameType) sdd2ddgz_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 2, 1]T[ 3, 0, 4]
        }
        // Swap back if necessary
        if (needSwap) swapZud();
      }

      // Set masses back to original values
      coupl_mass_.bmass = bMassSav;
      coupl_mass_.cmass = cMassSav;
      coupl_mass_.smass = sMassSav;
      coupl_mass_.umass = uMassSav;
      coupl_mass_.dmass = dMassSav;

      // Add average over colours in the initial state
      if (id[pos1Sav]==21) MGME2*=1.0/8.0;
      else MGME2*=1.0/3.0;
      if (id[pos2Sav]==21) MGME2*=1.0/8.0;
      else MGME2*=1.0/3.0;

    }
    // Z production
    //------------------------------------------------------------------------

    //------------------------------------------------------------------------
    // W production
    else if (nIn == 2 && idN == 24) {

      // Massless only, setting all quark masses to zero
      double bMassSav   = coupl_mass_.bmass;
      coupl_mass_.bmass = 0.0;
      double cMassSav   = coupl_mass_.cmass;
      coupl_mass_.cmass = 0.0;
      double sMassSav   = coupl_mass_.smass;
      coupl_mass_.smass = 0.0;
      double dMassSav   = coupl_mass_.dmass;
      coupl_mass_.dmass = 0.0;
      double uMassSav   = coupl_mass_.umass;
      coupl_mass_.umass = 0.0;

      // Safety
      for (int iHel=0; iHel<12; iHel++) HELsbp[iHel] = -1;
      // Sum over W polarizations
      bool sumWpol  = (pol[nMom-1] == 9);
      // Sum and average over incoming polarizations
      bool avIncome = (pol[pos1Sav] == 9 && pol[pos2Sav] == 9);
      // Average factor for incoming polarization (+- for partons)
      nPolIn = (avIncome ? 4 : 1);
      // NOTE not checking for sum over incoming, but explicit W polarization!

      // Ensure ME evaluated on shell
      const double mSys = p[nMom-1].mCalc();
      coupl_mass_.wmass = mSys;

      // In case of W- reverse order of partons, use -id and -pol
      if (id[nMom-1] < 0) {
        map<int,int> swapMap;
        for (int i=0; i<(nMom-1); ++i) swapMap[i] = nMom-2-i;
        swapMap[nMom-1] = nMom-1;
        if (nOut == 4 && nGluonTot == 1) {
          // Special case for q qbar q g qbar W
          swapMap[0] = 1;
          swapMap[1] = 0;
          swapMap[2] = 4;
          swapMap[3] = 3;
          swapMap[4] = 2;
        }
        vector<Vec4> tmpMom = p;
        vector<int>  tmpId  = id, tmpPol = pol;
        for (int i=0; i<nMom; i++) {
          p[i] = tmpMom[swapMap[i]];
          if (tmpId[swapMap[i]] == 21) id[i] = tmpId[swapMap[i]];
          else id[i] = -tmpId[swapMap[i]];
          if (abs(tmpPol[swapMap[i]]) == 1) pol[i] = -tmpPol[swapMap[i]];
          else pol[i] = tmpPol[swapMap[i]];
        }
        pos1Sav = swapMap[pos1Sav];
        pos2Sav = swapMap[pos2Sav];
        setIIpos(pos1Sav, pos2Sav);
      }

      // Get absolute values of ids
      vector<int> idAbs = id;
      for (int i=0; i<(int)idAbs.size(); i++) idAbs[i] = abs(idAbs[i]);

      // pp->W
      if (nOut == 1 && id[0]!=21 && id[1]!=21) {
        // 0) Default assumption: fully unpolarized
        // 1) Sum over W polarization, but explicit polarizations otherwise
        if (sumWpol && !avIncome) {
          int iHelNow = 1 + ((pol[0] == -1) ? 0 : 3) + ((pol[1] == -1) ? 0 : 6);
          for (int iHel=0; iHel<3; iHel++) HELsbp[iHel] = iHelNow+iHel;
          if (verbose >= 5) printHels(HELsbp);
        }
        // 2) Sum over incoming and W polarizations
        else if (avIncome && isPolarised) {
          // Nothing to do here as fully unpolarized
        }
        // Boost into rest frame of the W
        p[2].e(mSys);     p[2].px(0.0); p[2].py(0.0); p[2].pz(0.0);
        p[0].e(mSys/2.0); p[0].px(0.0); p[0].py(0.0); p[0].pz(mSys/2.0);
        p[1].e(mSys/2.0); p[1].px(0.0); p[1].py(0.0); p[1].pz(-mSys/2.0);
        // Convert momenta to MadGraph format
        double PNEW[3][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        setWcouplings(idAbs[0], idAbs[1]);
        sudx2wx_(&colorDepth,&icol,PNEW,HELsbp,&MGME2);
      }

      // pp->jW
      else if (nOut == 2) {
        // Find order
        int order[4] = { -1, -1, -1, -1 };
        if (pos1Sav == 0 && pos2Sav == 1) {
          // (-q, -g, qbar, W)
          // MG order 1, 0, 2
          order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 0;
        } else if (pos1Sav == 1 && pos2Sav == 2) {
          // (q, -g, -qbar, W)
          // MG order 2, 1, 0
          order[0] = 2; order[1] = 1; order[2] = 0; order[3] = 1;
        } else if (pos1Sav == 0 && pos2Sav == 2) {
          // (-q, g, -qbar, W)
          // MG order 2, 0, 1
          order[0] = 2; order[1] = 0; order[2] = 1; order[3] = 2;
        }
        // Convert momenta to MadGraph format
        double PNEW[4][4];
        Vec4toMG(p, order[0], order[1], order[2], PNEW);
        if (verbose >= 5) cout << " Using order = " << order[3] << endl;
        // Set W couplings
        setWcouplings(idAbs[0], idAbs[2]);
        // 0) Default assumption: fully unpolarized
        // 1) Sum over W polarization, but explicit polarizations overwise
        if (sumWpol && !avIncome) {
          int polMap[3] = { 12, 6, 3 };
          int iHelNow   = 1;
          for (int iHel=0; iHel<3; iHel++)
            iHelNow += ((pol[order[iHel]] == -1) ? 0 : polMap[iHel]);
          for (int iHel=0; iHel<3; iHel++) HELsbp[iHel] = iHelNow+(iHel%3);
          if (verbose >= 5) printHels(HELsbp);
        }
        // 2) Sum over incoming and W polarizations
        else if (avIncome && isPolarised) {
          int polInt2 = ((pol[order[2]] == -1) ? 0 : 3);
          for (int iHel=0; iHel<12; iHel++) {
            HELsbp[iHel] = 1 + iHel + polInt2;
            if ((iHel%3) == 2) polInt2 += 3;
          }
          if (verbose >= 5) printHels(HELsbp);
        }
        if (order[3] == 0)
          sgdx2uxwx_(&colorDepth,&icol,PNEW,HELsbp,&MGME2); // T[ 1, 2, 0]
        else if (order[3] == 1)
          sug2dwx_(&colorDepth,&icol,PNEW,HELsbp,&MGME2); // T[ 2, 1, 0]
        else if (order[3] == 2)
          sudx2gwx_(&colorDepth,&icol,PNEW,HELsbp,&MGME2); // T[ 1, 0, 2]
      }

      // pp->jjW
      else if (nOut == 3) {
        // Find order
        int order[5] = { -1, -1, -1, -1, -1 };
        int nFlavs   = 0;
        map<int,int> flavours;
        for (int i=1; i<=5; ++i) flavours[i] = 0;
        bool singleFirst = false, swapPairs = false;
        if (nGluonTot == 2) {
          // Momenta and ids are now given in colour order q g g qbar
          if (pos1Sav == 0 && pos2Sav == 1) {
            // (-q, -g, g, qbar, W)
            // MG order 1, 0, 2, 3
            order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 3; order[4] = 0;
          } else if (pos1Sav == 0 && pos2Sav == 2) {
            // (-q, g, -g, qbar, W)
            // MG order 2, 0, 1, 3
            order[0] = 2; order[1] = 0; order[2] = 1; order[3] = 3; order[4] = 1;
          } else if (pos1Sav == 0 && pos2Sav == 3) {
            // (-q, g, g, -qbar, W)
            // MG order 3, 0, 1, 2
            order[0] = 3; order[1] = 0; order[2] = 1; order[3] = 2; order[4] = 2;
          } else if (pos1Sav == 1 && pos2Sav == 2) {
            // (q, -g, -g, qbar, W)
            // MG order 1, 2, 3, 0
            order[0] = 1; order[1] = 2; order[2] = 3; order[3] = 0; order[4] = 3;
          } else if (pos1Sav== 1 && pos2Sav == 3) {
            // (q, -g, g, -qbar, W)
            // MG order 3, 1, 2, 0
            order[0] = 3; order[1] = 1; order[2] = 2; order[3] = 0; order[4] = 4;
          } else if (pos1Sav == 2 && pos2Sav == 3) {
            // (q, g, -g, -qbar, W)
            // MG order 3, 2, 1, 0
            order[0] = 3; order[1] = 2; order[2] = 1; order[3] = 0; order[4] = 5;
          }
        } else if (nGluonTot == 0) {
          // Momenta and ids are now given in order q qbar q qbar
          // Count number of flavours
          for (int i=0; i<nMom; ++i) flavours[idAbs[i]]++;
          for (int i=1; i<=5; ++i) if (flavours[i] > 0) nFlavs++;
          // Check if quark pairs should be swapped
          // a) 3 different quark flavours, identify shower quark pair
          if (nFlavs == 3) {
            // Swap q qbar'' q' qbar -> q' qbar q qbar''
            if (idAbs[0] == idAbs[3]) swapPairs = true;
          }
          // b) 2 different quark flavours, identify single flavour
          else {
            // Swap q qbar' q qbar -> q qbar q qbar'
            if ( (idAbs[0] == idAbs[2]) && (idAbs[0] == idAbs[3]) )
              swapPairs = true;
            else if ( (idAbs[1] == idAbs[2]) && (idAbs[1] == idAbs[3]) )
              singleFirst = true;
            // Swap q qbar q' qbar -> q' qbar q qbar
            else if ( (idAbs[0] == idAbs[1]) && (idAbs[0] == idAbs[3]) ) {
              singleFirst = true;
              swapPairs   = true;
            }
          }
          if (swapPairs) {
            int swapMap[4] = { 2, 3, 0, 1 };
            vector<Vec4> tmpMom = p;
            vector<int>  tmpId  = id, tmpIdAbs = idAbs, tmpPol = pol;
            for (int i=0; i<4; i++) {
              p[i]     = tmpMom[swapMap[i]];
              id[i]    = tmpId[swapMap[i]];
              idAbs[i] = tmpIdAbs[swapMap[i]];
              pol[i]   = tmpPol[swapMap[i]];
            }
            pos1Sav = swapMap[pos1Sav];
            pos2Sav = swapMap[pos2Sav];
            setIIpos(pos1Sav, pos2Sav);
          }
          // a) 3 different quark flavours
          if (nFlavs == 3) {
            if (pos1Sav == 0 && pos2Sav == 1) {
              // (-q', -qbar, q, qbar'', W)
              // MG order 1, 0, 2, 3
              order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 3; order[4] = 6;
            }
            else if (pos1Sav == 0 && pos2Sav == 2) {
              // (-q', qbar, -q, qbar'', W)
              // MG order 2, 0, 1, 3
              order[0] = 2; order[1] = 0; order[2] = 1; order[3] = 3; order[4] = 7;
            }
            else if (pos1Sav == 0 && pos2Sav == 3) {
              // (-q', qbar, q, -qbar'', W)
              // MG order 3, 0, 2, 1
              order[0] = 3; order[1] = 0; order[2] = 2; order[3] = 1; order[4] = 8;
            }
            else if (pos1Sav == 1 && pos2Sav == 2) {
              // Not possible in shower
              // (q', -qbar, -q, qbar'', W)
              // MG order 1, 2, 3, 0
              order[0] = 1; order[1] = 2; order[2] = 3; order[3] = 0; order[4] = 9;
            }
            else if (pos1Sav == 1 && pos2Sav == 3) {
              // (q', -qbar, q, -qbar'', W)
              // MG order 1, 3, 2, 0
              order[0] = 1; order[1] = 3; order[2] = 2; order[3] = 0; order[4] = 10;
            }
            else if (pos1Sav == 2 && pos2Sav == 3) {
              // (q', qbar, -q, -qbar'', W)
              // MG order 3, 2, 0, 1
              order[0] = 3; order[1] = 2; order[2] = 0; order[3] = 1; order[4] = 11;
            }
          }
          // b) 2 different quark flavours
          else {
            if (!singleFirst) { // q qbar q qbar'
              if (pos1Sav == 0 && pos2Sav == 1) {
                // (-q, -qbar, q, qbar', W)
                // MG order 1, 0, 2, 3
                order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 3; order[4] = 12;
              }
              else if (pos1Sav == 0 && pos2Sav == 2) {
                // (-q, qbar, -q, qbar', W)
                // MG order 2, 0, 1, 3
                order[0] = 2; order[1] = 0; order[2] = 1; order[3] = 3; order[4] = 13;
              }
              else if (pos1Sav == 0 && pos2Sav == 3) {
                // (-q, qbar, q, -qbar', W)
                // MG order 0, 3, 2, 1
                order[0] = 0; order[1] = 3; order[2] = 2; order[3] = 1; order[4] = 14;
              }
              else if (pos1Sav == 1 && pos2Sav == 2) {
                // (q, -qbar, -q, qbar', W)
                // MG order 1, 2, 0, 3
                order[0] = 1; order[1] = 2; order[2] = 0; order[3] = 3; order[4] = 15;
              }
              else if (pos1Sav == 1 && pos2Sav == 3) {
                // (q, -qbar, q, -qbar', W)
                // MG order 1, 3, 2, 0
                order[0] = 1; order[1] = 3; order[2] = 2; order[3] = 0; order[4] = 16;
              }
              else if (pos1Sav == 2 && pos2Sav == 3) {
                // (q, qbar, -q, -qbar', W)
                // MG order 2, 3, 0, 1
                order[0] = 2; order[1] = 3; order[2] = 0; order[3] = 1; order[4] = 17;
              }
            } else { // q' qbar q qbar
              if (pos1Sav == 0 && pos2Sav == 1) {
                // (-q', -qbar, q, qbar, W)
                // MG order 1, 0, 2, 3
                order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 3; order[4] = 18;
              }
              else if (pos1Sav == 0 && pos2Sav == 2) {
                // (-q', qbar, -q, qbar, W)
                // MG order 0, 2, 1, 3
                order[0] = 0; order[1] = 2; order[2] = 1; order[3] = 3; order[4] = 19;
              }
              else if (pos1Sav == 0 && pos2Sav == 3) {
                // (-q', qbar, q, -qbar, W)
                // MG order 3, 0, 2, 1
                order[0] = 3; order[1] = 0; order[2] = 2; order[3] = 1; order[4] = 20;
              }
              else if (pos1Sav == 1 && pos2Sav == 2) {
                // (q', -qbar, -q, qbar, W)
                // MG order 3, 0, 2, 1
                order[0] = 3; order[1] = 0; order[2] = 2; order[3] = 1; order[4] = 21;
              }
              else if (pos1Sav == 1 && pos2Sav == 3) {
                // (q', -qbar, q, -qbar, W)
                // MG order 1, 3, 2, 0
                order[0] = 1; order[1] = 3; order[2] = 2; order[3] = 0; order[4] = 22;
              }
              else if (pos1Sav == 2 && pos2Sav == 3) {
                // (q', qbar, -q, -qbar, W)
                // MG order 3, 2, 0, 1
                order[0] = 3; order[1] = 2; order[2] = 0; order[3] = 1; order[4] = 23;
              }
            }
          }
        }
        // Convert momenta to MadGraph format
        double PNEW[5][4];
        Vec4toMG(p, order[0], order[1], order[2], order[3], PNEW);
        if (verbose >= 5) cout << " Using order = " << order[4] << endl;
        // Set W couplings
        setWcouplings(idAbs[0], idAbs[3]);
        // 0) Default assumption: fully unpolarized
        // 1) Sum over W polarization, but explicit polarizations overwise
        if (sumWpol && !avIncome) {
          int polMap[4] = { 24, 12, 6, 3 };
          int iHelNow   = 1;
          for (int iHel=0; iHel<4; iHel++)
            iHelNow += ((pol[order[iHel]] == -1) ? 0 : polMap[iHel]);
          for (int iHel=0; iHel<3; iHel++) HELsbp[iHel] = iHelNow+(iHel%3);
          if (verbose >= 5) printHels(HELsbp);
        }
        // 2) Sum over incoming and W polarizations
        else if (avIncome && isPolarised) {
          int polInt2 = ((pol[order[2]] == -1) ? 0 : 6);
          int polInt3 = ((pol[order[3]] == -1) ? 0 : 3);
          for (int iHel=0; iHel<12; iHel++) {
            HELsbp[iHel] = 1 + iHel + polInt2 + polInt3;
            if ((iHel%3) == 2) polInt2 += 9;
          }
          if (verbose >= 5) printHels(HELsbp);
        }
        // Different color flows
        int icol1 = 1;
        int icol2 = 2;
        if (order[4] == 0)
          sgdx2guxwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 3, 0, 3]
        else if (order[4] == 1)
          sgdx2guxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 3, 2, 0]
        else if (order[4] == 2)
          sudx2ggwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 0, 2, 3]
        else if (order[4] == 3)
          sgg2uxdwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 2, 0, 1]
        else if (order[4] == 4)
          sug2gdwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 0, 1, 2]
        else if (order[4] == 5)
          sug2gdwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 0, 2, 1]
        else if (order[4] == 6)
          scdx2cuxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 0]T[ 2, 3]
        else if (order[4] == 7)
          scxdx2cxuxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 2]T[ 0, 3]
        else if (order[4] == 8)
          sudx2ccxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3]
        else if (order[4] == 9)
          sccx2uxdwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 1, 2]
        else if (order[4] == 10)
          scu2cdwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 1]
        else if (order[4] == 11)
          sucx2dcxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 0]T[ 2, 3]
        else if (order[4] == 12)
          sddx2duxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 0]T[ 2, 3]
        else if (order[4] == 13)
          sdxdx2dxuxwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 2]T[ 0, 3]
        else if (order[4] == 14)
          sdxu2ddxwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 1]T[ 0, 3]
        else if (order[4] == 15)
          sddx2duxwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3]
        else if (order[4] == 16)
          sdu2ddwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 1]
        else if (order[4] == 17)
          sdxu2ddxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 0, 1]T[ 2, 3]
        else if (order[4] == 18)
          sudx2uuxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 0]T[ 2, 3]
        else if (order[4] == 19)
          sdxux2uxuxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 0, 2]T[ 1, 3]
        else if (order[4] == 20)
          sudx2uuxwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3]
        else if (order[4] == 21)
          suux2duxwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3]
        else if (order[4] == 22)
          suu2udwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 1]
        else if (order[4] == 23)
          suux2duxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 0, 1]T[ 2, 3]
      }
      
      // pp->jjjW
      else if (nOut == 4) {
        // Find order
        int  order[6] = { -1, -1, -1, -1, -1, -1 };
        int nFlavs   = 0;
        map<int,int> flavours;
        for (int i=1; i<=5; ++i) flavours[i] = 0;
        bool singleFirst = false, swapPairs = false;
        if (nGluonTot == 3) {
          // Momenta and ids are now given in colour order q g g g qbar
          if (pos1Sav == 0 && pos2Sav == 4) {
            // (-q, g, g, g, -qbar, W)
            // MG order 4, 0, 1, 2, 3
            order[0] = 4; order[1] = 0; order[2] = 1; order[3] = 2; order[4] = 3; order[5] = 0;
          }
          else if (pos1Sav == 1 && pos2Sav == 2) {
            // (q, -g, -g, g, qbar, W)
            // MG order 1, 2, 3, 4, 0
            order[0] = 1; order[1] = 2; order[2] = 3; order[3] = 4; order[4] = 0; order[5] = 1;
          }
          else if (pos1Sav == 1 && pos2Sav == 3) {
            // (q, -g, g, -g, qbar, W)
            // MG order 1, 3, 2, 4, 0
            order[0] = 1; order[1] = 3; order[2] = 2; order[3] = 4; order[4] = 0; order[5] = 2;
          }
          else if (pos1Sav == 2 && pos2Sav == 3) {
            // (q, g, -g, -g, qbar, W)
            // MG order 2, 3, 1, 4, 0
            order[0] = 2; order[1] = 3; order[2] = 1; order[3] = 4; order[4] = 0; order[5] = 3;
          }
          else if (pos1Sav == 3 && pos2Sav == 4) {
            // (q, g, g, -g, -qbar, W)
            // MG order 4, 3, 1, 2, 0
            order[0] = 4; order[1] = 3; order[2] = 1; order[3] = 2; order[4] = 0; order[5] = 4;
          }
          else if (pos1Sav == 2 && pos2Sav == 4) {
            // (q, g, -g, g, -qbar, W)
            // MG order 4, 2, 3, 1, 0
            order[0] = 4; order[1] = 2; order[2] = 3; order[3] = 1; order[4] = 0; order[5] = 5;
          }
          else if (pos1Sav == 1 && pos2Sav == 4) {
            // (q, -g, g, g, -qbar, W)
            // MG order 4, 1, 2, 3, 0
            order[0] = 4; order[1] = 1; order[2] = 2; order[3] = 3; order[4] = 0; order[5] = 6;
          }
          else if (pos1Sav == 0 && pos2Sav == 1) {
            // (-q, -g, g, g, qbar, W)
            // MG order 1, 0, 2, 3, 4
            order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 3; order[4] = 4; order[5] = 7;
          }
          else if (pos1Sav == 0 && pos2Sav == 2) {
            // (-q, g, -g, g, qbar, W)
            // MG order 2, 0, 1, 3, 4
            order[0] = 2; order[1] = 0; order[2] = 1; order[3] = 3; order[4] = 4; order[5] = 8;
          }
          else if (pos1Sav == 0 && pos2Sav == 3) {
            // (-q, g, g, -g, qbar, W)
            // MG order 3, 0, 1, 2, 4
            order[0] = 3; order[1] = 0; order[2] = 1; order[3] = 2; order[4] = 4; order[5] = 9;
          }
        } else if (nGluonTot == 1) {
          // Momenta and ids are now given in order q qbar q qbar
          // Count number of flavours
          for (int i=0; i<nMom; ++i) flavours[idAbs[i]]++;
          for (int i=1; i<=5; ++i) if (flavours[i] > 0) nFlavs++;
          // Check if quark pairs should be swapped
          // a) 3 different quark flavours, identify shower quark pair
          if (nFlavs == 3) {
            // Swap q qbar'' q' g qbar -> q' qbar q g qbar''
            if (idAbs[0] == idAbs[4]) swapPairs = true;
          }
          // b) 2 different quark flavours, identify single flavour
          else {
            // Swap q qbar' q g qbar -> q qbar q g qbar'
            if ( (idAbs[0] == idAbs[2]) && (idAbs[0] == idAbs[4]) )
              swapPairs = true;
            if ( (idAbs[1] == idAbs[2]) && (idAbs[1] == idAbs[4]) )
              singleFirst = true;
            // Swap q qbar q' g qbar -> q' qbar q g qbar
            if ( (idAbs[0] == idAbs[1]) && (idAbs[0] == idAbs[4]) ) {
              singleFirst = true;
              swapPairs   = true;
            }
          }
          if (swapPairs) {
            int swapMap[5] = { 2, 4, 0, 3, 1 };
            vector<Vec4> tmpMom = p;
            vector<int>  tmpId  = id, tmpIdAbs = idAbs, tmpPol = pol;
            for (int i=0; i<5; i++) {
              p[i]     = tmpMom[swapMap[i]];
              id[i]    = tmpId[swapMap[i]];
              idAbs[i] = tmpIdAbs[swapMap[i]];
              pol[i]   = tmpPol[swapMap[i]];
            }
            pos1Sav = swapMap[pos1Sav];
            pos2Sav = swapMap[pos2Sav];
            setIIpos(pos1Sav, pos2Sav);
          }
          // a) 3 different quark flavours
          if (nFlavs == 3) {
            if (pos1Sav == 0 && pos2Sav == 1) {
              // (-q', -qbar, q, g, qbar'', W)
              // MG order 1, 0, 2, 4, 3
              order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 4; order[4] = 3; order[5] = 10;
            }
            else if (pos1Sav == 0 && pos2Sav == 2) {
              // (-q', qbar, -q, g, qbar'', W)
              // MG order 2, 0, 1, 4, 3
              order[0] = 2; order[1] = 0; order[2] = 1; order[3] = 4; order[4] = 3; order[5] = 11;
            }
            else if (pos1Sav == 0 && pos2Sav == 3) {
              // (-q', qbar, q, -g, qbar'', W)
              // MG order 0, 3, 2, 1, 4
              order[0] = 0; order[1] = 3; order[2] = 2; order[3] = 1; order[4] = 4; order[5] = 12;
            }
            else if (pos1Sav == 0 && pos2Sav == 4) {
              // (-q', qbar, q, g, -qbar'', W)
              // MG order 0, 4, 2, 1, 3
              order[0] = 0; order[1] = 4; order[2] = 2; order[3] = 1; order[4] = 3; order[5] = 13;
            }
            else if (pos1Sav == 1 && pos2Sav == 2) {
              // Not possible in shower
              // (q', -qbar, -q, g, qbar'', W)
              // MG order 1, 2, 4, 0, 3
              order[0] = 1; order[1] = 2; order[2] = 4; order[3] = 0; order[4] = 3; order[5] = 14;
            }
            else if (pos1Sav == 1 && pos2Sav == 3) {
              // (q', -qbar, q, -g, qbar'', W)
              // MG order 1, 3, 2, 0, 4
              order[0] = 1; order[1] = 3; order[2] = 2; order[3] = 0; order[4] = 4; order[5] = 15;
            }
            else if (pos1Sav == 1 && pos2Sav == 4) {
              // (q', -qbar, q, g, -qbar'', W)
              // MG order 1, 4, 2, 0, 3
              order[0] = 1; order[1] = 4; order[2] = 2; order[3] = 0; order[4] = 3; order[5] = 16;
            }
            else if (pos1Sav == 2 && pos2Sav == 3) {
              // (q', qbar, -q, -g, qbar'', W)
              // MG order 2, 3, 1, 0, 4
              order[0] = 2; order[1] = 3; order[2] = 1; order[3] = 0; order[4] = 4; order[5] = 17;
            }
            else if (pos1Sav == 2 && pos2Sav == 4) {
              // (q', qbar, -q, g, -qbar'', W)
              // MG order 2, 4, 1, 0, 3
              order[0] = 2; order[1] = 4; order[2] = 1; order[3] = 0; order[4] = 3; order[5] = 18;
            }
            else if (pos1Sav == 3 && pos2Sav == 4) {
              // (q', qbar, q, -g, -qbar'', W)
              // MG order 4, 3, 2, 1, 0
              order[0] = 4; order[1] = 3; order[2] = 2; order[3] = 1; order[4] = 0; order[5] = 19;
            }
          }
          // b) 2 different quark flavours
          else {
            if (!singleFirst) { // q qbar q g qbar'
              if (pos1Sav == 0 && pos2Sav == 1) {
                // (-q, -qbar, q, g, qbar', W)
                // MG order 1, 0, 2, 4, 3
                order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 4; order[4] = 3; order[5] = 20;
              }
              else if (pos1Sav == 0 && pos2Sav == 2) {
                // (-q, qbar, -q, g, qbar', W)
                // MG order 2, 0, 1, 4, 3
                order[0] = 2; order[1] = 0; order[2] = 1; order[3] = 4; order[4] = 3; order[5] = 21;
              }
              else if (pos1Sav == 0 && pos2Sav == 3) {
                // (-q, qbar, q, -g, qbar', W)
                // MG order 0, 3, 2, 1, 4
                order[0] = 0; order[1] = 3; order[2] = 2; order[3] = 1; order[4] = 4; order[5] = 22;
              }
              else if (pos1Sav == 0 && pos2Sav == 4) {
                // (-q, qbar, q, g, -qbar', W)
                // MG order 4, 0, 2, 1, 3
                order[0] = 4; order[1] = 0; order[2] = 2; order[3] = 1; order[4] = 3; order[5] = 23;
              }
              else if (pos1Sav == 1 && pos2Sav == 2) {
                // (q, -qbar, -q, g, qbar', W)
                // MG order 1, 2, 0, 4, 3
                order[0] = 1; order[1] = 2; order[2] = 0; order[3] = 4; order[4] = 3; order[5] = 24;
              }
              else if (pos1Sav == 1 && pos2Sav == 3) {
                // (q, -qbar, q, -g, qbar', W)
                // MG order 1, 3, 0, 2, 4
                order[0] = 1; order[1] = 3; order[2] = 0; order[3] = 2; order[4] = 4; order[5] = 25;
              }
              else if (pos1Sav == 1 && pos2Sav == 4) {
                // (q, -qbar, q, g, -qbar', W)
                // MG order 1, 4, 0, 2, 3
                order[0] = 1; order[1] = 4; order[2] = 0; order[3] = 2; order[4] = 3; order[5] = 26;
              }
              else if (pos1Sav == 2 && pos2Sav == 3) {
                // (q, qbar, -q, -g, qbar', W)
                // MG order 2, 3, 0, 1, 4
                order[0] = 2; order[1] = 3; order[2] = 0; order[3] = 1; order[4] = 4; order[5] = 27;
              }
              else if (pos1Sav == 2 && pos2Sav == 4) {
                // (q, qbar, -q, g, -qbar', W)
                // MG order 4, 2, 0, 1, 3
                order[0] = 4; order[1] = 2; order[2] = 0; order[3] = 1; order[4] = 3; order[5] = 28;
              }
              else if (pos1Sav == 3 && pos2Sav == 4) {
                // (q, qbar, q, -g, -qbar', W)
                // MG order 4, 3, 0, 2, 1
                order[0] = 4; order[1] = 3; order[2] = 0; order[3] = 2; order[4] = 1; order[5] = 29;
              }
            } else { // q' qbar q g qbar
              if (pos1Sav == 0 && pos2Sav == 1) {
                // (-q', -qbar, q, g, qbar, W)
                // MG order 1, 0, 2, 4, 3
                order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 4; order[4] = 3; order[5] = 30;
              }
              else if (pos1Sav == 0 && pos2Sav == 2) {
                // (-q', qbar, -q, g, qbar, W)
                // MG order 0, 2, 1, 4, 3
                order[0] = 0; order[1] = 2; order[2] = 1; order[3] = 4; order[4] = 3; order[5] = 31;
              }
              else if (pos1Sav == 0 && pos2Sav == 3) {
                // (-q', qbar, q, -g, qbar, W)
                // MG order 3, 0, 2, 1, 4
                order[0] = 0; order[1] = 3; order[2] = 2; order[3] = 1; order[4] = 4; order[5] = 32;
              }
              else if (pos1Sav == 0 && pos2Sav == 4) {
                // (-q', qbar, q, g, -qbar, W)
                // MG order 4, 0, 2, 1, 3
                order[0] = 4; order[1] = 0; order[2] = 2; order[3] = 1; order[4] = 3; order[5] = 33;
              }
              else if (pos1Sav == 1 && pos2Sav == 2) {
                // (q', -qbar, -q, g, qbar, W)
                // MG order 1, 2, 0, 4, 3
                order[0] = 1; order[1] = 2; order[2] = 0; order[3] = 4; order[4] = 3; order[5] = 34;
              }
              else if (pos1Sav == 1 && pos2Sav == 3) {
                // (q', -qbar, q, -g, qbar, W)
                // MG order 1, 3, 0, 2, 4
                order[0] = 1; order[1] = 3; order[2] = 0; order[3] = 2; order[4] = 4; order[5] = 35;
              }
              else if (pos1Sav == 1 && pos2Sav == 4) {
                // (q', -qbar, q, g, -qbar, W)
                // MG order 1, 4, 0, 2, 3
                order[0] = 1; order[1] = 4; order[2] = 0; order[3] = 2; order[4] = 3; order[5] = 36;
              }
              else if (pos1Sav == 2 && pos2Sav == 3) {
                // (q', qbar, -q, -g, qbar, W)
                // MG order 3, 2, 1, 4, 0
                order[0] = 3; order[1] = 2; order[2] = 1; order[3] = 4; order[4] = 0; order[5] = 37;
              }
              else if (pos1Sav == 2 && pos2Sav == 4) {
                // (q', qbar, -q, g, -qbar, W)
                // MG order 4, 2, 0, 1, 3
                order[0] = 4; order[1] = 2; order[2] = 0; order[3] = 1; order[4] = 3; order[5] = 38;
              }
              else if (pos1Sav == 3 && pos2Sav == 4) {
                // (q', qbar, q, -g, -qbar, W)
                // MG order 4, 3, 0, 2, 1
                order[0] = 4; order[1] = 3; order[2] = 0; order[3] = 2; order[4] = 1; order[5] = 39;
              }
            }
          }
        }
        // Convert momenta to MadGraph format
        double PNEW[6][4];
        Vec4toMG(p, order[0], order[1], order[2], order[3], order[4], PNEW);
        if (verbose >= 5) cout << " Using order = " << order[5] << endl;
        // Set W couplings
        setWcouplings(idAbs[0], idAbs[4]);
        // 0) Default assumption: fully unpolarized
        // 1) Sum over W polarization, but explicit polarizations overwise
        if (sumWpol && !avIncome) {
          int polMap[5] = { 48, 24, 12, 6, 3 };
          int iHelNow   = 1;
          for (int iHel=0; iHel<5; iHel++)
            iHelNow += ((pol[order[iHel]] == -1) ? 0 : polMap[iHel]);
          for (int iHel=0; iHel<3; iHel++) HELsbp[iHel] = iHelNow+(iHel%3);
          if (verbose >= 5) printHels(HELsbp);
        }
        // 2) Sum over incoming and W polarizations
        else if (avIncome && isPolarised) {
          int polInt2 = ((pol[order[2]] == -1) ? 0 : 12);
          int polInt3 = ((pol[order[3]] == -1) ? 0 :  6);
          int polInt4 = ((pol[order[4]] == -1) ? 0 :  3);
          for (int iHel=0; iHel<12; iHel++) {
            HELsbp[iHel] = 1 + iHel + polInt2 + polInt3 + polInt4;
            if ((iHel%3) == 2) polInt2 += 21;
          }
          if (verbose >= 5) printHels(HELsbp);
        }
        // Different color flows
        int icol1 = 1;
        int icol2 = 2;
        int icol3 = 3;
        int icol4 = 4;
        int icol6 = 6;
        if (order[5] == 0)
          sudx2gggwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 0, 2, 3, 4]
        else if (order[5] == 1)
          sgg2guxdwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 4, 3, 0, 1, 2]
        else if (order[5] == 2)
          sgg2guxdwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 4, 3, 0, 2, 1]
        else if (order[5] == 3)
          sgg2guxdwx_(&colorDepth,&icol6,PNEW,HELsbp,&MGME2); // T[ 4, 3, 2, 0, 1]
        else if (order[5] == 4)
          sug2ggdwx_(&colorDepth,&icol6,PNEW,HELsbp,&MGME2); // T[ 4, 0, 2, 3, 1]
        else if (order[5] == 5)
          sug2ggdwx_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 4, 0, 3, 1, 2]
        else if (order[5] == 6)
          sug2ggdwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 4, 0, 1, 2, 3]
        else if (order[5] == 7)
          sgdx2gguxwx_(&colorDepth,&icol6,PNEW,HELsbp,&MGME2); // T[ 1, 4, 0, 2, 3]
        else if (order[5] == 8)
          sgdx2gguxwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 4, 2, 0, 3]
        else if (order[5] == 9)
          sgdx2gguxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 4, 2, 3, 0]
        else if (order[5] == 10) {
          if (swapPairs) scdx2cuxgwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0, 4]
          else           scdx2cuxgwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 0]T[ 2, 3, 4]
        } else if (order[5] == 11) {
          if (swapPairs) scxdx2cxuxgwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 0, 3]T[ 1, 2, 4]
          else           scxdx2cxuxgwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 2]T[ 0, 3, 4]
        } else if (order[5] == 12) {
          if (swapPairs) sdxg2ccxuxwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 4]T[ 0, 3, 1]
          else           sdxg2ccxuxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 0, 3]T[ 2, 4, 1]
        } else if (order[5] == 13) {
          if (swapPairs) sdxu2ccxgwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 1]T[ 0, 3, 4]
          else           sdxu2ccxgwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 0, 3]T[ 2, 1, 4]
        } else if (order[5] == 14) {
          if (swapPairs) sccx2uxdgwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 0, 4]T[ 1, 2]
          else           sccx2uxdgwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 1, 2, 4]
        } else if (order[5] == 15) {
          if (swapPairs) scg2cduxwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 4]T[ 3, 0, 1]
          else           scg2cduxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 4, 1]
        } else if (order[5] == 16) {
          if (swapPairs) scu2cdgwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 1]T[ 3, 0, 4]
          else           scu2cdgwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 1, 4]
        } else if (order[5] == 17) {
          if (swapPairs) scxg2cxduxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 0, 4]T[ 3, 2, 1]
          else           scxg2cxduxwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 2]T[ 0, 4, 1]
        } else if (order[5] == 18) {
          if (swapPairs) scxu2cxdgwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 0, 1]T[ 3, 2, 4]
          else           scxu2cxdgwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 2]T[ 0, 1, 4]
        } else if (order[5] == 19) {
          if (swapPairs) sug2ccxdwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 4, 3, 1]
          else           sug2ccxdwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 4, 3]T[ 2, 0, 1]
        } else if (order[5] == 20) {
          if (swapPairs) sddx2duxgwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0, 4]
          else           sddx2duxgwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 0]T[ 2, 3, 4]
        } else if (order[5] == 21) {
          if (swapPairs) sdxdx2dxuxgwx_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 0, 3]T[ 1, 2, 4]
          else           sdxdx2dxuxgwx_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 1, 2]T[ 0, 3, 4]
        } else if (order[5] == 22) {
          if (swapPairs) sdxg2ddxuxwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 4]T[ 0, 3, 1]
          else           sdxg2ddxuxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 0, 3]T[ 2, 4, 1]
        } else if (order[5] == 23) {
          if (swapPairs) sudx2ddxgwx_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3, 4]
          else           sudx2ddxgwx_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 1, 3]T[ 2, 0, 4]
        } else if (order[5] == 24) {
          if (swapPairs) sddx2duxgwx_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 1, 3]T[ 2, 0, 4]
          else           sddx2duxgwx_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3, 4]
        } else if (order[5] == 25) {
          if (swapPairs) sdg2dduxwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 4]T[ 2, 0, 1]
          else           sdg2dduxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 3, 4, 1]
        } else if (order[5] == 26) {
          if (swapPairs) sdu2ddgwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 1]T[ 2, 0, 4]
          else           sdu2ddgwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 3, 1, 4]
        } else if (order[5] == 27) {
          if (swapPairs) sdxg2ddxuxwx_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 0, 4]T[ 2, 3, 1]
          else           sdxg2ddxuxwx_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 0, 4, 1]
        } else if (order[5] == 28) {
          if (swapPairs) sudx2ddxgwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 4]T[ 2, 3, 4]
          else           sudx2ddxgwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0, 4]
        } else if (order[5] == 29) {
          if (swapPairs) sug2dddxwx_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 4, 1]
          else           sug2dddxwx_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 2, 4]T[ 3, 0, 1]
        } else if (order[5] == 30) {
          if (swapPairs) sudx2uuxgwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0, 4]
          else           sudx2uuxgwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 0]T[ 2, 3, 4]
        } else if (order[5] == 31) {
          if (swapPairs) sdxux2uxuxgwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 3]T[ 0, 2, 4]
          else           sdxux2uxuxgwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 0, 2]T[ 1, 3, 4]
        } else if (order[5] == 32) {
          if (swapPairs) sgdx2uuxuxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 4]T[ 1, 3, 0]
          else           sgdx2uuxuxwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 3]T[ 2, 4, 0]
        } else if (order[5] == 33) {
          if (swapPairs) sudx2uuxgwx_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3, 4]
          else           sudx2uuxgwx_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 1, 3]T[ 2, 0, 4]
        } else if (order[5] == 34) {
          if (swapPairs) suux2duxgwx_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 1, 3]T[ 2, 0, 4]
          else           suux2duxgwx_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3, 4]
        } else if (order[5] == 35) {
          if (swapPairs) sug2duuxwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 4]T[ 2, 0, 1]
          else           sug2duuxwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 3, 4, 1]
        } else if (order[5] == 36) {
          if (swapPairs) suu2dugwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 1]T[ 2, 0, 4]
          else           suu2dugwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 3, 1, 4]
        } else if (order[5] == 37) {
          if (swapPairs) sgux2uxuxdwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 3]T[ 4, 2, 0]
          else           sgux2uxuxdwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 4, 2]T[ 1, 3, 0]
        } else if (order[5] == 38) {
          if (swapPairs) suux2duxgwx_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 0]T[ 2, 3, 4]
          else           suux2duxgwx_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0, 4]
        } else if (order[5] == 39) {
          if (swapPairs) sug2duuxwx_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 4, 1]
          else           sug2duuxwx_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 2, 4]T[ 3, 0, 1]
        }
      }

      // Set masses back to original values
      coupl_mass_.bmass = bMassSav;
      coupl_mass_.cmass = cMassSav;
      coupl_mass_.smass = sMassSav;
      coupl_mass_.umass = uMassSav;
      coupl_mass_.dmass = dMassSav;

      // Add average over colours in the initial state
      if (id[pos1Sav]==21) MGME2*=1.0/8.0;
      else MGME2*=1.0/3.0;
      if (id[pos2Sav]==21) MGME2*=1.0/8.0;
      else MGME2*=1.0/3.0;

    }
    // W production
    //------------------------------------------------------------------------

    //------------------------------------------------------------------------
    // H production
    // NOTE If the quarks are massless, the pre-calculated Yukawa couplings
    // are used. If they are massive, the coupling gets re-calculated
    else if (nIn == 2 && idN == 25) {

      // Massless only, setting all quark masses to zero
      double bMassSav   = coupl_mass_.bmass;
      coupl_mass_.bmass = 0.0;
      double cMassSav   = coupl_mass_.cmass;
      coupl_mass_.cmass = 0.0;
      double sMassSav   = coupl_mass_.smass;
      coupl_mass_.smass = 0.0;
      double dMassSav   = coupl_mass_.dmass;
      coupl_mass_.dmass = 0.0;
      double uMassSav   = coupl_mass_.umass;
      coupl_mass_.umass = 0.0;

      // Safety
      for (int iHel=0; iHel<12; iHel++) HELsbp[iHel] = -1;
      // Sum over H polarizations
      bool sumHpol  = (pol[nMom-1] == 9);
      // Sum and average over incoming polarizations
      bool avIncome = (pol[pos1Sav] == 9 && pol[pos2Sav] == 9);
      // Average factor for incoming polarization (+- for partons)
      nPolIn = (avIncome ? 4 : 1);
      // NOTE not checking for sum over incoming, but explicit H polarization!

      // Ensure ME evaluated on shell
      const double mSys = p[nMom-1].mCalc();
      coupl_mass_.hmass = mSys;

      // pp->H
      if (nOut == 1) {
        // 0) Default assumption: fully unpolarized
        // 1) Sum over H polarization, but explicit polarizations otherwise
        if (sumHpol && !avIncome) {
          HELsbp[0] = 1 + ((pol[0] == -1) ? 0 : 1) + ((pol[1] == -1) ? 0 : 2);
          if (verbose >= 5) printHels(HELsbp);
        }
        // 2) Sum over incoming and H polarizations
        else if (avIncome && isPolarised) {
          // Nothing to do here as fully unpolarized
        }
        // Boost into rest frame of the H
        p[2].e(mSys);     p[2].px(0.0); p[2].py(0.0); p[2].pz(0.0);
        p[0].e(mSys/2.0); p[0].px(0.0); p[0].py(0.0); p[0].pz(mSys/2.0);
        p[1].e(mSys/2.0); p[1].px(0.0); p[1].py(0.0); p[1].pz(-mSys/2.0);
        // Convert momenta to MadGraph format
        double PNEW[3][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        if (nGluonTot == 2) sgg2h_(&colorDepth,&icol,PNEW,HELsbp,&MGME2);
        else {
          // Set couplings
          coupl_yuk_.ghbot[0] = complex( 0.0, yukCoup[abs(id[0])] );
          coupl_yuk_.ghbot[1] = complex( 0.0, yukCoup[abs(id[0])] );
          sbbx2h_(&colorDepth,&icol,PNEW,HELsbp,&MGME2);
        }
      }

      // pp->jH
      else if (nOut == 2) {
        // Find order
        int order[4] = { -1, -1, -1, -1 };
        if (nGluonTot == 3) {
          if (pos1Sav == 0 && pos2Sav == 1) {
            // (-g, -g, g, H)
            // MG order 0, 1, 2
            order[0] = 0; order[1] = 1; order[2] = 2; order[3] = 0;
          } else if (pos1Sav == 0 && pos2Sav == 2) {
            // (-g, g, -g, H)
            // MG order 0, 2, 1
            order[0] = 0; order[1] = 2; order[2] = 1; order[3] = 1;
          }
        } else if (nGluonTot == 1) {
          // Set couplings
          coupl_yuk_.ghbot[0] = complex( 0.0, yukCoup[abs(id[0])] );
          coupl_yuk_.ghbot[1] = complex( 0.0, yukCoup[abs(id[0])] );
          if (pos1Sav == 1 && pos2Sav == 2) {
            // (q, -g, -qbar, H)
            // MG order 1, 2, 0
            order[0] = 1; order[1] = 2; order[2] = 0; order[3] = 2;
          } else if (pos1Sav == 0 && pos2Sav == 1) {
            // (-q, -g, qbar, H)
            // MG order 1, 0, 2
            order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 3;
          } else if (pos1Sav == 0 && pos2Sav == 2) {
            // (-q, g, -qbar, H)
            // MG order 2, 0, 1
            order[0] = 2; order[1] = 0; order[2] = 1; order[3] = 4;
          }
        }
        // Convert momenta to MadGraph format
        double PNEW[4][4];
        Vec4toMG(p, order[0], order[1], order[2], PNEW);
        if (verbose >= 5) cout << " Using order = " << order[3] << endl;
        // 0) Default assumption: fully unpolarized
        // 1) Sum over H polarization, but explicit polarizations overwise
        if (sumHpol && !avIncome) {
          int polMap[3] = { 4, 2, 1 };
          HELsbp[0] = 1;
          for (int iHel=0; iHel<3; iHel++)
            HELsbp[0] += ((pol[order[iHel]] == -1) ? 0 : polMap[iHel]);
          if (verbose >= 5) printHels(HELsbp);
        }
        // 2) Sum over incoming and H polarizations
        else if (avIncome && isPolarised) {
          int polInt2 = ((pol[order[2]] == -1) ? 0 : 1);
          for (int iHel=0; iHel<4; iHel++) {
            HELsbp[iHel] = 1 + iHel + polInt2;
            polInt2 += 1;
          }
          if (verbose >= 5) printHels(HELsbp);
        }
        // Different color flows
        int icol1 = 1;
        int icol2 = 2;
        if (order[3] == 0)
          sgg2gh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // F[0, 1, 2]
        else if (order[3] == 1)
          sgg2gh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // F[0, 2, 1]
        else if (order[3] == 2)
          sgb2bh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 1, 0]
        else if (order[3] == 3)
          sgbx2bxh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 2, 0]
        else if (order[3] == 4)
          sbbx2gh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 0, 2]
      }

      // pp->jjH
      else if (nOut == 3) {
        // Find order
        int  order[5] = { -1, -1, -1, -1, -1 };
        bool sameFlav = false;
        if (nGluonTot == 4) {
          // Momenta and ids are now given in order g(i) g g g
          if (pos1Sav == 0 && pos2Sav == 1) {
            // (-g, -g, g, g, H)
            // MG order 0, 1, 2, 3
            order[0] = 0; order[1] = 1; order[2] = 2; order[3] = 3; order[4] = 0;
          }
          else if (pos1Sav == 0 && pos2Sav == 2) {
            // (-g, g, -g, g, H)
            // MG order 0, 2, 1, 3
            order[0] = 0; order[1] = 2; order[2] = 1; order[3] = 3; order[4] = 1;
          }
          else if (pos1Sav == 0 && pos2Sav == 3) {
            // (-g, g, g, -g, H)
            // MG order 0, 3, 1, 2
            order[0] = 0; order[1] = 3; order[2] = 1; order[3] = 2; order[4] = 2;
          }
        } else if (nGluonTot == 2) {
          // Momenta and ids are now given in colour order q g g qbar
          // Set couplings
          coupl_yuk_.ghbot[0] = complex( 0.0, yukCoup[abs(id[0])] );
          coupl_yuk_.ghbot[1] = complex( 0.0, yukCoup[abs(id[0])] );
          if (pos1Sav == 1 && pos2Sav == 2) {
            // (q, -g, -g, qbar, H)
            // MG order 1, 2, 0, 3
            order[0] = 1; order[1] = 2; order[2] = 0; order[3] = 3; order[4] = 3;
          }
          else if (pos1Sav == 1 && pos2Sav == 3) {
            // (q, -g, g, -qbar, H)
            // MG order 1, 3, 0, 2
            order[0] = 1; order[1] = 3; order[2] = 0; order[3] = 2; order[4] = 4;
          }
          else if (pos1Sav == 0 && pos2Sav == 1) {
            // (-q, -g, g, qbar, H)
            // MG order 1, 0, 3, 2
            order[0] = 1; order[1] = 0; order[2] = 3; order[3] = 2; order[4] = 5;
          }
          else if (pos1Sav == 0 && pos2Sav == 2) {
            // (-q, g, -g, qbar, H)
            // MG order 2, 0, 3, 1
            order[0] = 2; order[1] = 0; order[2] = 3; order[3] = 1; order[4] = 6;
          }
          else if (pos1Sav == 2 && pos2Sav == 3) {
            // (q, g, -g, -qbar, H)
            // MG order 2, 3, 0, 1
            order[0] = 2; order[1] = 3; order[2] = 0; order[3] = 1; order[4] = 7;
          }
          else if (pos1Sav == 0 && pos2Sav == 3) {
            // (-q, g, g, -qbar, H)
            // MG order 3, 0, 2, 1
            order[0] = 3; order[1] = 0; order[2] = 2; order[3] = 1; order[4] = 8;
          }
        } else if (nGluonTot == 0) {
          // Momenta and ids are now given in order q qbar q qbar
          // Check if all quarks have the same flavour
          if ( (abs(id[1])==abs(id[2])) && (abs(id[1])==abs(id[3])) && 
               (abs(id[1])==abs(id[4])) ) sameFlav = true;
          int posM1 = pos1Sav;
          int posM2 = pos2Sav;
          if (!sameFlav) {
            if ( (pos1Sav==0 && pos2Sav==1) || (pos1Sav==2 && pos2Sav==3) ) { 
              posM1 = pos2Sav; posM2 = pos1Sav;
            } else if (pos1Sav==0 && pos2Sav==3) { posM1 = 0; posM2 = 2; }
            else if (pos1Sav==1 && pos2Sav==2) { posM1 = 2; posM2 = 0; }
          }
          // Set couplings
          coupl_yuk_.ghbot[0] = complex( 0.0, yukCoup[abs(id[posM1])] );
          coupl_yuk_.ghbot[1] = complex( 0.0, yukCoup[abs(id[posM1])] );
          coupl_yuk_.ghcha[0] = complex( 0.0, yukCoup[abs(id[posM2])] );
          coupl_yuk_.ghcha[1] = complex( 0.0, yukCoup[abs(id[posM2])] );
          if (pos1Sav == 1 && pos2Sav == 3) {
            // (q, -qbar, q, -qbar, H)
            // MG order 1, 3, 2, 0
            order[0] = 1; order[1] = 3; order[2] = 2; order[3] = 0; order[4] = 9;
          }
          else if (pos1Sav == 0 && pos2Sav == 2) {
            // (-q, qbar, -q, qbar, H)
            // MG order 0, 2, 3, 1
            order[0] = 0; order[1] = 2; order[2] = 3; order[3] = 1; order[4] = 10;
          }
          else if (pos1Sav == 0 && pos2Sav == 1) {
            // (-q, -qbar, q, qbar, H)
            // MG order 1, 0, 2, 3
            order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 3; order[4] = 11;
          }
          else if (pos1Sav == 2 && pos2Sav == 3) {
            // (q, qbar, -q, -qbar, H)
            // MG order 3, 2, 0, 1
            order[0] = 3; order[1] = 2; order[2] = 0; order[3] = 1; order[4] = 12;
          }
          else if (pos1Sav == 0 && pos2Sav == 3) {
            // (-q, qbar, q, -qbar, H)
            // MG order 3, 0, 2, 1
            order[0] = 3; order[1] = 0; order[2] = 2; order[3] = 1; order[4] = 13;
          }
          else if (pos1Sav == 1 && pos2Sav == 2) {
            // (q, -qbar, -q, qbar, H)
            // MG order 1, 2, 0, 3
            order[0] = 1; order[1] = 2; order[2] = 0; order[3] = 3; order[4] = 14;
          }
        }
        // Convert momenta to MadGraph format
        double PNEW[5][4];
        Vec4toMG(p, order[0], order[1], order[2], order[3], PNEW);
        if (verbose >= 5) cout << " Using order = " << order[4] << endl;
        // 0) Default assumption: fully unpolarized
        // 1) Sum over H polarization, but explicit polarizations overwise
        if (sumHpol && !avIncome) {
          int polMap[4] = { 8, 4, 2, 1 };
          HELsbp[0] = 1;
          for (int iHel=0; iHel<4; iHel++)
            HELsbp[0] += ((pol[order[iHel]] == -1) ? 0 : polMap[iHel]);
          if (verbose >= 5) printHels(HELsbp);
        }
        // 2) Sum over incoming and H polarizations
        else if (avIncome && isPolarised) {
          int polInt2 = ((pol[order[2]] == -1) ? 0 : 2);
          int polInt3 = ((pol[order[3]] == -1) ? 0 : 1);
          for (int iHel=0; iHel<4; iHel++) {
            HELsbp[iHel] = 1 + iHel + polInt2 + polInt3;
            polInt2 += 3;
          }
          if (verbose >= 5) printHels(HELsbp);
        }
        // Different color flows
        int icol1 = 1;
        int icol2 = 2;
        int icol5 = 5;
        if (order[4] == 0)
          sgg2ggh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // F[0, 1, 2, 3]
        else if (order[4] == 1)
          sgg2ggh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // F[0, 2, 1, 3]
        else if (order[4] == 2)
          sgg2ggh_(&colorDepth,&icol5,PNEW,HELsbp,&MGME2); // F[0, 2, 3, 1]
        else if (order[4] == 3)
          sgg2bbxh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 3, 0, 1]
        else if (order[4] == 4)
          sgb2bgh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 1, 0, 3]
        else if (order[4] == 5)
          sgbx2bxgh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 2, 0, 3]
        else if (order[4] == 6)
          sgbx2bxgh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 2, 3, 0]
        else if (order[4] == 7)
          sgb2bgh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 1, 3, 0]
        else if (order[4] == 8)
          sbbx2ggh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 0, 3, 2]
        else if (order[4] == 9) {
          if (sameFlav) sbb2bbh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 1]
          else          sbc2bch_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 1]
        } else if (order[4] == 10) {
          if (sameFlav) sbxbx2bxbxh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 2]T[ 0, 3]
          else          sbxcx2bxcxh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 2]T[ 0, 3]
        } else if (order[4] == 11) {
          if (sameFlav) sbbx2bbxh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0]
          else          sbcx2bcxh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0]
        } else if (order[4] == 12) {
          if (sameFlav) sbbx2bbxh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0]
          else          sbcx2bcxh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0]
        } else if (order[4] == 13) {
          if (sameFlav) sbbx2bbxh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3]
          else          sbbx2ccxh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3]
        } else if (order[4] == 14) {
          if (sameFlav) sbbx2bbxh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3]
          else          sbbx2ccxh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3]
        }
      }

      // pp->jjjH
      else if (nOut == 4) {
        // Find order
        int  order[6] = { -1, -1, -1, -1, -1, -1 };
        bool sameFlav = false;
        if (nGluonTot == 5) {
          // Momenta and ids are now given in order g(i) g g g g
          if (pos1Sav == 0 && pos2Sav == 1) {
            // (-g, -g, g, g, g, H)
            // MG order 0, 1, 2, 3, 4
            order[0] = 0; order[1] = 1; order[2] = 2; order[3] = 3; order[4] = 4; order[5] = 0;
          }
          else if (pos1Sav == 0 && pos2Sav == 2) {
            // (-g, g, -g, g, g, H)
            // MG order 0, 2, 1, 3, 4
            order[0] = 0; order[1] = 2; order[2] = 1; order[3] = 3; order[4] = 4; order[5] = 1;
          }
          else if (pos1Sav == 0 && pos2Sav == 3) {
            // (-g, g, g, -g, g, H)
            // MG order 0, 3, 1, 2, 4
            order[0] = 0; order[1] = 3; order[2] = 1; order[3] = 2; order[4] = 4; order[5] = 2;
          }
          else if (pos1Sav == 0 && pos2Sav == 4) {
            // (-g, g, g, g, -g, H)
            // MG order 0, 4, 1, 2, 3
            order[0] = 0; order[1] = 4; order[2] = 1; order[3] = 2; order[4] = 3; order[5] = 3;
          }
        } else if (nGluonTot == 3) {
          // Momenta and ids are now given in colour order q g g g qbar
          // Set couplings
          coupl_yuk_.ghbot[0] = complex( 0.0, yukCoup[abs(id[0])] );
          coupl_yuk_.ghbot[1] = complex( 0.0, yukCoup[abs(id[0])] );
          if (pos1Sav == 0 && pos2Sav == 4) {
            // (-q, g, g, g, -qbar, H)
            // MG order 4, 0, 1, 2, 3
            order[0] = 4; order[1] = 0; order[2] = 1; order[3] = 2; order[4] = 3; order[5] = 4;
          }
          else if (pos1Sav == 1 && pos2Sav == 2) {
            // (q, -g, -g, g, qbar, H)
            // MG order 1, 2, 0, 4, 3
            order[0] = 1; order[1] = 2; order[2] = 0; order[3] = 4; order[4] = 3; order[5] = 5;
          }
          else if (pos1Sav == 1 && pos2Sav == 3) {
            // (q, -g, g, -g, qbar, H)
            // MG order 1, 3, 0, 4, 2
            order[0] = 1; order[1] = 3; order[2] = 0; order[3] = 4; order[4] = 2; order[5] = 6;
          }
          else if (pos1Sav == 2 && pos2Sav == 3) {
            // (q, g, -g, -g, qbar, H)
            // MG order 2, 3, 0, 4, 1
            order[0] = 2; order[1] = 3; order[2] = 0; order[3] = 4; order[4] = 1; order[5] = 7;
          }
          else if (pos1Sav == 0 && pos2Sav == 1) {
            // (-q, -g, g, g, qbar, H)
            // MG order 1, 0, 4, 2, 3
            order[0] = 1; order[1] = 0; order[2] = 4; order[3] = 2; order[4] = 3; order[5] = 8;
          }
          else if (pos1Sav == 0 && pos2Sav == 2) {
            // (-q, g, -g, g, qbar, H)
            // MG order 2, 0, 4, 1, 3
            order[0] = 2; order[1] = 0; order[2] = 4; order[3] = 1; order[4] = 3; order[5] = 9;
          }
          else if (pos1Sav == 0 && pos2Sav == 3) {
            // (-q, g, g, -g, qbar, H)
            // MG order 3, 0, 4, 1, 2
            order[0] = 3; order[1] = 0; order[2] = 4; order[3] = 1; order[4] = 2; order[5] = 10;
          }
          else if (pos1Sav == 1 && pos2Sav == 4) {
            // (q, -g, g, g, -qbar, H)
            // MG order 1, 4, 0, 2, 3
            order[0] = 1; order[1] = 4; order[2] = 0; order[3] = 2; order[4] = 3; order[5] = 11;
          }
          else if (pos1Sav == 2 && pos2Sav == 4) {
            // (q, g, -g, g, -qbar, H)
            // MG order 2, 4, 0, 1, 3
            order[0] = 2; order[1] = 4; order[2] = 0; order[3] = 1; order[4] = 3; order[5] = 12;
          }
          else if (pos1Sav == 3 && pos2Sav == 4) {
            // (q, g, g, -g, -qbar, H)
            // MG order 3, 4, 0, 1, 2
            order[0] = 3; order[1] = 4; order[2] = 0; order[3] = 1; order[4] = 2; order[5] = 13;
          }
        } else if (nGluonTot == 1) {
          // Momenta and ids are now given in order q qbar q g qbar
          // Check if all quarks have the same flavour
          if ( (abs(id[1])==abs(id[2])) && (abs(id[1])==abs(id[3])) && 
               (abs(id[1])==abs(id[4])) ) sameFlav = true;
          int posM1 = pos1Sav;
          int posM2 = pos2Sav;
          if (!sameFlav) {
            if ( (pos1Sav==0 && pos2Sav==1) || (pos1Sav==2 && pos2Sav==4) ) {
              posM1 = pos2Sav; posM2 = pos1Sav;
            } else if ( (pos1Sav==0 && pos2Sav==4) || (pos1Sav==0 && pos2Sav==3) || 
                        (pos1Sav==3 && pos2Sav==4) ) {
              posM1 = 0; posM2 = 2;
            } else if ( (pos1Sav==1 && pos2Sav==2) || (pos1Sav==1 && pos2Sav==3) || 
                        (pos1Sav==2 && pos2Sav==3) ) { 
              posM1 = 2; posM2 = 0;
            }
          }
          else {
            posM1 = (id[pos1Sav]==21 ? pos2Sav : pos1Sav);
            posM2 = (id[pos2Sav]==21 ? pos1Sav : pos2Sav);
          }
          // Set couplings
          coupl_yuk_.ghbot[0] = complex( 0.0, yukCoup[abs(id[posM1])] );
          coupl_yuk_.ghbot[1] = complex( 0.0, yukCoup[abs(id[posM1])] );
          coupl_yuk_.ghcha[0] = complex( 0.0, yukCoup[abs(id[posM2])] );
          coupl_yuk_.ghcha[1] = complex( 0.0, yukCoup[abs(id[posM2])] );
          if (pos1Sav == 0 && pos2Sav == 1) {
            // (-q, -qbar, q, g, qbar, H)
            // MG order 1, 0, 2, 4, 3
            order[0] = 1; order[1] = 0; order[2] = 2; order[3] = 4; order[4] = 3; order[5] = 14;
          }
          else if (pos1Sav == 0 && pos2Sav == 2) {
            // (-q, qbar, -q, g, qbar, H)
            // MG order 0, 2, 4, 1, 3
            order[0] = 0; order[1] = 2; order[2] = 4; order[3] = 1; order[4] = 3; order[5] = 15;
          }
          else if (pos1Sav == 0 && pos2Sav == 3) {
            // (-q, qbar, q, -g, qbar, H)
            // MG order 0, 3, 4, 2, 1
            order[0] = 0; order[1] = 3; order[2] = 4; order[3] = 2; order[4] = 1; order[5] = 16;
          }
          else if (pos1Sav == 0 && pos2Sav == 4) {
            // (-q, qbar, q, g, -qbar, H)
            // MG order 4, 0, 2, 1, 3
            order[0] = 4; order[1] = 0; order[2] = 2; order[3] = 1; order[4] = 3; order[5] = 17;
          }
          else if (pos1Sav == 1 && pos2Sav == 2) {
            // (q, -qbar, -q, g, qbar, H)
            // MG order 1, 2, 0, 4, 3
            order[0] = 1; order[1] = 2; order[2] = 0; order[3] = 4; order[4] = 3; order[5] = 18;
          }
          else if (pos1Sav == 1 && pos2Sav == 3) {
            // (q, -qbar, q, -g, qbar, H)
            // MG order 1, 3, 2, 0, 4
            order[0] = 1; order[1] = 3; order[2] = 2; order[3] = 0; order[4] = 4; order[5] = 19;
          }
          else if (pos1Sav == 1 && pos2Sav == 4) {
            // (q, -qbar, q, g, -qbar, H)
            // MG order 1, 4, 2, 0, 3
            order[0] = 1; order[1] = 4; order[2] = 2; order[3] = 0; order[4] = 3; order[5] = 20;
          }
          else if (pos1Sav == 2 && pos2Sav == 3) {
            // (q, qbar, -q, -g, qbar, H)
            // MG order 2, 3, 1, 0, 4
            order[0] = 2; order[1] = 3; order[2] = 1; order[3] = 0; order[4] = 4; order[5] = 21;
          }
          else if (pos1Sav == 2 && pos2Sav == 4) {
            // (q, qbar, -q, g, -qbar, H)
            // MG order 4, 2, 0, 1, 3
            order[0] = 4; order[1] = 2; order[2] = 0; order[3] = 1; order[4] = 3; order[5] = 22;
          }
          else if (pos1Sav == 3 && pos2Sav == 4) {
            // (q, qbar, q, -g, -qbar, H)
            // MG order 4, 3, 0, 2, 1
            order[0] = 4; order[1] = 3; order[2] = 0; order[3] = 2; order[4] = 1; order[5] = 23;
          }
        }
        // Convert momenta to MadGraph format
        double PNEW[6][4];
        Vec4toMG(p, order[0], order[1], order[2], order[3], order[4], PNEW);
        if (verbose >= 5) cout << " Using order = " << order[5] << endl;
        // 0) Default assumption: fully unpolarized
        // 1) Sum over H polarization, but explicit polarizations overwise
        if (sumHpol && !avIncome) {
          int polMap[5] = { 16, 8, 4, 2, 1 };
          HELsbp[0] = 1;
          for (int iHel=0; iHel<5; iHel++)
            HELsbp[0] += ((pol[order[iHel]] == -1) ? 0 : polMap[iHel]);
          if (verbose >= 5) printHels(HELsbp);
        }
        // 2) Sum over incoming and H polarizations
        else if (avIncome && isPolarised) {
          int polInt2 = ((pol[order[2]] == -1) ? 0 : 4);
          int polInt3 = ((pol[order[3]] == -1) ? 0 : 2);
          int polInt4 = ((pol[order[4]] == -1) ? 0 : 1);
          for (int iHel=0; iHel<4; iHel++) {
            HELsbp[iHel] = 1 + iHel + polInt2 + polInt3 + polInt4;
            polInt2 += 7;
          }
          if (verbose >= 5) printHels(HELsbp);
        }
        // Different color flows
        int icol1  = 1;
        int icol2  = 2;
        int icol3  = 3;
        int icol4  = 4;
        int icol5  = 5;
        int icol6  = 6;
        int icol18 = 18;
        int icol22 = 22;
        if (order[5] == 0)
          sgg2gggh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // F[0, 1, 2, 3, 4]
        else if (order[5] == 1)
          sgg2gggh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // F[0, 2, 1, 3, 4]
        else if (order[5] == 2)
          sgg2gggh_(&colorDepth,&icol18,PNEW,HELsbp,&MGME2); // F[0, 2, 3, 1, 4]
        else if (order[5] == 3)
          sgg2gggh_(&colorDepth,&icol22,PNEW,HELsbp,&MGME2); // F[0, 2, 3, 4, 1]
        else if (order[5] == 4)
          sbbx2gggh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 0, 2, 3, 4]
        else if (order[5] == 5)
          sgg2bbxgh_(&colorDepth,&icol6,PNEW,HELsbp,&MGME2); // T[ 2, 3, 0, 1, 4]
        else if (order[5] == 6)
          sgg2bbxgh_(&colorDepth,&icol5,PNEW,HELsbp,&MGME2); // T[ 2, 3, 0, 4, 1]
        else if (order[5] == 7)
          sgg2bbxgh_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 2, 3, 4, 0, 1]
        else if (order[5] == 8)
          sgbx2bxggh_(&colorDepth,&icol6,PNEW,HELsbp,&MGME2); // T[ 1, 2, 0, 3, 4]
        else if (order[5] == 9)
          sgbx2bxggh_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 1, 2, 3, 0, 4]
        else if (order[5] == 10)
          sgbx2bxggh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 1, 2, 3, 4, 0]
        else if (order[5] == 11)
          sgb2bggh_(&colorDepth,&icol5,PNEW,HELsbp,&MGME2); // T[ 2, 1, 0, 3, 4] 
        else if (order[5] == 12)
          sgb2bggh_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 2, 1, 3, 0, 4]
        else if (order[5] == 13)
          sgb2bggh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 1, 3, 4, 0]
        else if (order[5] == 14) {
          if (sameFlav) sbbx2bbxgh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 0]T[ 2, 3, 4]
          else          sbcx2bcxgh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 1, 0]T[ 2, 3, 4]
        } else if (order[5] == 15) {
          if (sameFlav) sbxbx2bxbxgh_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 0, 3]T[ 1, 2, 4]
          else          sbxcx2bxcxgh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 0, 3]T[ 1, 2, 4]
        } else if (order[5] == 16) {
          if (sameFlav) sbxg2bxbbxh_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 3, 2, 1]T[ 0, 4]
          else          sbxg2bxccxh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 2, 1]T[ 0, 4]
        } else if (order[5] == 17) {
          if (sameFlav) sbbx2bbxgh_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 2, 0, 4]T[ 1, 3]
          else          sbbx2ccxgh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 0, 4]T[ 1, 3]
        } else if (order[5] == 18) {
          if (sameFlav) sbbx2bbxgh_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3, 4]
          else          sbbx2ccxgh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 0]T[ 1, 3, 4]
        } else if (order[5] == 19) {
          if (sameFlav) sbg2bbbxh_(&colorDepth,&icol3,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 4, 1]
          else          sbg2bccxh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 4, 1]
        } else if (order[5] == 20) {
          if (sameFlav) sbb2bbgh_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 1, 4]
          else          sbc2bcgh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 0]T[ 2, 1, 4]
        } else if (order[5] == 21) {
          if (sameFlav) sbxg2bxbbxh_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 3, 2]T[ 0, 4, 1]
          else          sbxg2bxccxh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 3, 2]T[ 0, 4, 1]
        } else if (order[5] == 22) {
          if (sameFlav) sbbx2bbxgh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0, 4]
          else          sbcx2bcxgh_(&colorDepth,&icol1,PNEW,HELsbp,&MGME2); // T[ 2, 3]T[ 1, 0, 4]
        } else if (order[5] == 23) {
          if (sameFlav) sbg2bbbxh_(&colorDepth,&icol4,PNEW,HELsbp,&MGME2); // T[ 2, 4]T[ 3, 0, 1]
          else          sbg2bccxh_(&colorDepth,&icol2,PNEW,HELsbp,&MGME2); // T[ 2, 4]T[ 3, 0, 1]
        }
      }

      // Set masses back to original values
      coupl_mass_.bmass = bMassSav;
      coupl_mass_.cmass = cMassSav;
      coupl_mass_.smass = sMassSav;
      coupl_mass_.umass = uMassSav;
      coupl_mass_.dmass = dMassSav;

      // Add average over colours in the initial state
      if (id[pos1Sav]==21) MGME2*=1.0/8.0;
      else MGME2*=1.0/3.0;
      if (id[pos2Sav]==21) MGME2*=1.0/8.0;
      else MGME2*=1.0/3.0;

    }
    // H production
    //------------------------------------------------------------------------

    //------------------------------------------------------------------------
    // Z decay
    else if (nIn == 1 && id0 == 23) {

      // Safety
      for (int ihel=0; ihel<4; ihel++) HEL[ihel] = -1;
      // Sum over Z polarizations, for specific final-state polarizations
      bool sumZpol = (pol[0] == 9);
      if (sumZpol) {
        for (int i=1; i<nMom; i++)
          if (pol[i] == 9) sumZpol = false;
      }
      // Fully polarized Z decay
      bool fulPolZ = (!sumZpol);
      if (fulPolZ) {
        for (int i=0; i<nMom; i++)
          if (pol[i] == 9) fulPolZ = false;
      }
      // Average factor for incoming polarization (+-0 for Z)
      nPolIn = (fulPolZ ? 1 : 3);

      // Ensure ME evaluated on shell
      const double mSys = p[0].mCalc();
      coupl_mass_.zmass = mSys;

      // Z->jj
      if (nOut == 2 && id[1]!=21 && id[2]!=21 && abs(id[1])==abs(id[2])) {
        // 0) Default assumption: fully unpolarized Z decay
        // 1) Sum over Z polarizations, for specific final-state polarizations
        if (sumZpol) {
          HEL[0] = (int)(1 + (pol[2] + 1)/2 + (pol[1] + 1));
          HEL[1] = (int)(5 + (pol[2] + 1)/2 + (pol[1] + 1));
          HEL[2] = (int)(9 + (pol[2] + 1)/2 + (pol[1] + 1));
        }
        // 2) Fully polarized Z decay
        else if (fulPolZ) {
          HEL[0] = (int)(1 + (pol[0] + 1)*4 + (pol[2] + 1)/2 + (pol[1] + 1));
        }
        // Convert momenta to MadGraph format
        double PNEW[3][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        // Swap if necessary
        if (abs(id[1]) % 2 == 0) swapZud();
        // Set masses
        coupl_mass_.dmass = m[1];
        sz2ddx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        // Swap back if necessary
        if (abs(id[1]) % 2 == 0) swapZud();
      }

      // Z->jjj
      else if (nOut == 3) {
        // 0) Default assumption: fully unpolarized Z decay
        // 1) Sum over Z polarizations, for specific final-state polarizations
        if (sumZpol) {
          HEL[0] = (int)( 1 + (pol[3] + 1)/2 + (pol[2] + 1) + 2*(pol[1] + 1));
          HEL[1] = (int)( 9 + (pol[3] + 1)/2 + (pol[2] + 1) + 2*(pol[1] + 1));
          HEL[2] = (int)(17 + (pol[3] + 1)/2 + (pol[2] + 1) + 2*(pol[1] + 1));
        }
        // 2) Fully polarized Z decay
        else if (fulPolZ) {
          nPolIn = 1;
          HEL[0] = (int)(1 + (pol[0] + 1)*8 + (pol[3] + 1)/2 + (pol[2] + 1) 
                         + 2*(pol[1] + 1));
        }
        // Convert momenta to MadGraph format
        double PNEW[4][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        // Swap if necessary
        if (abs(id[1]) % 2 == 0) swapZud();
        // Set masses
        coupl_mass_.dmass = m[1];
        sz2dgdx_(&colorDepth,&icol,PNEW,HEL,&MGME2); // T[ 1, 3, 2]
        // Swap back if necessary
        if (abs(id[1]) % 2 == 0) swapZud();
      }

      // Z->jjjj
      else if (nOut == 4) {
        // 0) Default assumption: fully unpolarized Z decay
        // 1) Sum over Z polarizations, for specific final-state polarizations
        if (sumZpol) {
          HEL[0] = (int)( 1 + (pol[4] + 1)/2 + (pol[3] + 1) + 2*(pol[2] + 1) 
                         + 4*(pol[1] + 1));
          HEL[1] = (int)(17 + (pol[4] + 1)/2 + (pol[3] + 1) + 2*(pol[2] + 1) 
                         + 4*(pol[1] + 1));
          HEL[2] = (int)(33 + (pol[4] + 1)/2 + (pol[3] + 1) + 2*(pol[2] + 1) 
                         + 4*(pol[1] + 1));
        }
        // 2) Fully polarized Z decay
        else if (fulPolZ) {
          HEL[0] = (int)(1 + (pol[0] + 1)*16 + (pol[4] + 1)/2 + (pol[3] + 1) 
                         + 2*(pol[2] + 1) + 4*(pol[1] + 1));
        }
        // Convert momenta to MadGraph format
        double PNEW[5][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        // Swap if necessary
        if (abs(id[1]) % 2 == 0) swapZud();
        // Set masses
        coupl_mass_.dmass = m[1];
        coupl_mass_.umass = m[2];
        coupl_mass_.smass = m[2];
        // Different color flows
        int icol1 = 1;
        int icol2 = 2;
        if (nGluonTot == 2) {
          // Z -> q g g qbar
          sz2dggdx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 4, 2, 3]
        } else if (nGluonTot == 0) {
          // Z -> q qbar' q' qbar
          // Check if all quarks have the same flavour / same type (u/d)
          bool sameFlav = false, sameType = false;
          if ( (abs(id[1])==abs(id[2])) && (abs(id[1])==abs(id[3])) &&
               (abs(id[1])==abs(id[4])) ) sameFlav = true;
          if ( (abs(id[1])%2==1) && (abs(id[2])%2==1) && 
               (abs(id[3])%2==1) && (abs(id[4])%2==1) ) sameType = true;
          if ( (abs(id[1])%2==0) && (abs(id[2])%2==0) && 
               (abs(id[3])%2==0) && (abs(id[4])%2==0) ) sameType = true;
               if (!sameFlav && !sameType) sz2duxudx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]
          else if (!sameFlav &&  sameType) sz2dsxsdx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]
          else if ( sameFlav &&  sameType) sz2ddxddx_(&colorDepth,&icol2,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]
        }
        // Swap back if necessary
        if (abs(id[1]) % 2 == 0) swapZud();
      }

      // Z->jjjjj
      else if (nOut == 5) {
        // 0) Default assumption: fully unpolarized Z decay
        // 1) Sum over Z polarizations, for specific final-state polarizations
        if (sumZpol) {
          HEL[0] = (int)( 1 + (pol[5] + 1)/2 + (pol[4] + 1) + 2*(pol[3] + 1) 
                         + 4*(pol[2] + 1) + 8*(pol[1] + 1));
          HEL[1] = (int)(33 + (pol[5] + 1)/2 + (pol[4] + 1) + 2*(pol[3] + 1) 
                         + 4*(pol[2] + 1) + 8*(pol[1] + 1));
          HEL[2] = (int)(65 + (pol[5] + 1)/2 + (pol[4] + 1) + 2*(pol[3] + 1) 
                         + 4*(pol[2] + 1) + 8*(pol[1] + 1));
        }
        // 2) Fully polarized Z decay
        else if (fulPolZ) {
          HEL[0] = (int)(1 + (pol[0] + 1)*32 + (pol[5] + 1)/2 + (pol[4] + 1) 
                         + 2*(pol[3] + 1) + 4*(pol[2] + 1) + 8*(pol[1] + 1));
        }
        // Convert momenta to MadGraph format
        double PNEW[6][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        // Swap if necessary
        if (abs(id[1]) % 2 == 0) swapZud();
        // Set masses
        coupl_mass_.dmass = m[1];
        coupl_mass_.umass = m[2];
        coupl_mass_.smass = m[2];
        if (nGluonTot == 3) {
          // Z -> q g g g qbar
          sz2dgggdx_(&colorDepth,&icol,PNEW,HEL,&MGME2); // T[ 1, 5, 2, 3, 4]
        } else if (nGluonTot == 1) {
          // Z -> q qbar' q' g qbar
          // Check if all quarks have the same flavour / same type (u/d)
          bool sameFlav = false, sameType = false;
          if ( (abs(id[1])==abs(id[2])) && (abs(id[1])==abs(id[3])) &&
               (abs(id[1])==abs(id[5])) ) sameFlav = true;
          if ( (abs(id[1])%2==1) && (abs(id[2])%2==1) && 
               (abs(id[3])%2==1) && (abs(id[5])%2==1) ) sameType = true;
          if ( (abs(id[1])%2==0) && (abs(id[2])%2==0) && 
               (abs(id[3])%2==0) && (abs(id[5])%2==0) ) sameType = true;
               if (!sameFlav && !sameType) sz2duxugdx_(&colorDepth,&icol,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 5, 4]
          else if (!sameFlav &&  sameType) sz2dsxsgdx_(&colorDepth,&icol,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 5, 4]
          else if ( sameFlav &&  sameType) sz2ddxdgdx_(&colorDepth,&icol,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 5, 4]
        }
        // Swap back if necessary
        if (abs(id[1]) % 2 == 0) swapZud();
      }

      // Z->jjjjjj
      else if (nOut == 6) {
        // Find the right order of the partons
        map<int,int> mapVtoMG;
        // Default: order as it is
        for (int i=0; i<nMom; ++i) mapVtoMG[i] = i;
        // Note the different cases
        int caseNow = 0;
        // q g g g g qbar
        if (nGluonTot == 4) caseNow = 1;
        else if (nGluonTot == 2) {
          // q qbar' q' g g qbar
          if ( (id[4]==21) && (id[5]==21) ) caseNow = 2;
          // q g1 qbar' q' g2 qbar --- sort to --- q qbar' q' g1 g2 qbar
          else if ( (id[2]==21) && (id[5]==21) ) {
            caseNow = 3;
            mapVtoMG[2] = 3;
            mapVtoMG[3] = 4;
            mapVtoMG[4] = 2;
          }
        } else if (nGluonTot == 0) {
          vector<int> downType;
          vector<int> upType;
          map<int,int> flavours;
          for (int i=1; i<=5; ++i) flavours[i] = 0;
          for (int i=1; i<nMom; ++i) {
            flavours[abs(id[i])]++;
            if (abs(id[i]) % 2 == 0) upType.push_back(i); 
            else downType.push_back(i);
          }
          // 0 if 0 or 1 quark pair, 1 if 2 quark pairs, 2 if 3 quark pais
          int countSameFlavour = 0;
          for (int i=1; i<=5; ++i)
            countSameFlavour = max(max((flavours[i]-2)/2,0),countSameFlavour);
          // 2 down type quark pairs and 1 up type quark pair or
          // 1 down type quark pair and 2 up type quark pairs
          if ( (downType.size()==4 && upType.size()==2) ||
               (downType.size()==2 && upType.size()==4) ) {
            // d ubar u ubar u dbar
            if (countSameFlavour == 1) {
              caseNow = 4;
              int absFlav = 0; // flavour that appears once
              for (int i=1; i<=5; ++i) if (flavours[i] == 2) absFlav = i;
              pair<int,int> flavourPos; // quark first, antiquark second
              for (int i=1; i<nMom; ++i) if (abs(id[i]) == absFlav) {
                if (id[i] > 0) flavourPos.first = i; 
                else flavourPos.second = i;
              }
              vector<int> order = orderQQBqqbqqb(flavourPos);
              for (int i=0; i<nMom; ++i) mapVtoMG[i] = order[i];
            } 
            // d dbar' d' ubar u dbar
            else {
              caseNow = 5;
              int type = ( (downType.size()==2) ? 1 : 0 ); // type that appears once
              pair<int,int> typePos; // quark first, antiquark second
              for (int i=1; i<nMom; ++i) if (abs(id[i]) % 2 == type) {
                if (id[i] > 0) typePos.first = i; 
                else typePos.second = i;
              }
              vector<int> order = orderQQBqqbqqb(typePos, false);
              for (int i=0; i<nMom; ++i) mapVtoMG[i] = order[i];
            }
          }
          // 3 down type quark pairs or 3 up type quark pairs
          else {
            // d dbar d dbar d dbar
            if (countSameFlavour == 2) {
              caseNow = 6;
              int absFlav = 0; // flavour that appears once
              for (int i=1; i<=5; ++i) if (flavours[i] == 2) absFlav = i;
              pair<int,int> flavourPos; // quark first, antiquark second
              for (int i=1; i<nMom; ++i) if (abs(id[i]) == absFlav) {
                if (id[i] > 0) flavourPos.first = i; 
                else flavourPos.second = i;
              }
              vector<int> order = orderQQBqqbqqb(flavourPos);
              for (int i=0; i<nMom; ++i) mapVtoMG[i] = order[i];
            }
            // d dbar' d' dbar' d' dbar
            else if (countSameFlavour == 1) {
              caseNow = 7;
              int absFlav = 0; // flavour that appears once
              for (int i=1; i<=5; ++i) if (flavours[i] == 2) absFlav = i;
              pair<int,int> flavourPos; // quark first, antiquark second
              for (int i=1; i<nMom; ++i) if (abs(id[i]) == absFlav) {
                if (id[i] > 0) flavourPos.first = i; 
                else flavourPos.second = i;
              }
              vector<int> order = orderQQBqqbqqb(flavourPos);
              for (int i=0; i<nMom; ++i) mapVtoMG[i] = order[i];
            }
            // d dbar' d' dbar'' d'' dbar
            else if (countSameFlavour == 0) {
              caseNow = 8;
              int absFlav = 0; // pick one flavour to make decision
              for (int i=1; i<=5; ++i) if (flavours[i] == 2) absFlav = i;
              pair<int,int> flavourPos; // quark first, antiquark second
              for (int i=1; i<nMom; ++i) if (abs(id[i]) == absFlav) {
                if (id[i] > 0) flavourPos.first = i; 
                else flavourPos.second = i;
              }
              vector<int> order = orderQQBqqbqqb(flavourPos);
              for (int i=0; i<nMom; ++i) mapVtoMG[i] = order[i];
            }
          }
        }
        vector<Vec4>   tmpMom = p;
        vector<int>    tmpId  = id, tmpPol = pol;
        vector<double> tmpM   = m;
        for (int i=0; i<nMom; i++) {
          p[i]   = tmpMom[mapVtoMG[i]];
          id[i]  = tmpId[mapVtoMG[i]];
          pol[i] = tmpPol[mapVtoMG[i]];
          m[i]   = tmpM[mapVtoMG[i]];
        }
        // 0) Default assumption: fully unpolarized Z decay
        // 1) Sum over Z polarizations, for specific final-state polarizations
        if (sumZpol) {
          HEL[0] = (int)(1 + (pol[6] + 1)/2 + (pol[5] + 1) + 2*(pol[4] + 1) 
                         + 4*(pol[3] + 1) + 8*(pol[2] + 1) + 16*(pol[1] + 1));
          HEL[1] = (int)(65 + (pol[6] + 1)/2 + (pol[5] + 1) + 2*(pol[4] + 1) 
                         + 4*(pol[3] + 1) + 8*(pol[2] + 1) + 16*(pol[1] + 1));
          HEL[2] = (int)(129 + (pol[6] + 1)/2 + (pol[5] + 1) + 2*(pol[4] + 1) 
                         + 4*(pol[3] + 1) + 8*(pol[2] + 1) + 16*(pol[1] + 1));
        }
        // 2) Fully polarized Z decay
        else if (fulPolZ) {
          HEL[0] = (int)(1 + (pol[0] + 1)*64 + (pol[6] + 1)/2 + (pol[5] + 1) 
                         + 2*(pol[4] + 1) + 4*(pol[3] + 1) + 8*(pol[2] + 1) 
                         + 16*(pol[1] + 1));
        }
        // Convert momenta to MadGraph format
        double PNEW[7][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        // Different color flows
        int icol1  = 1;
        int icol5  = 5;
        int icol7  = 7;
        int icol11 = 11;
        // Swap if necessary
        if (abs(id[1]) % 2 == 0) swapZud();
        if (caseNow == 1) {
          // Z -> q g g g g qbar
          // Set masses
          coupl_mass_.dmass = m[1];
          sz2dggggdx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 6, 2, 3, 4, 5]
        } else if (caseNow == 2 || caseNow == 3) {
          // Z -> q qbar' q' g g qbar
          // Set masses
          coupl_mass_.dmass = m[1];
          coupl_mass_.umass = m[2];
          coupl_mass_.smass = m[2];
          // Check if all quarks have the same flavour / same type (u/d)
          bool sameFlav = false, sameType = false;
          if ( (abs(id[1])==abs(id[2])) && (abs(id[1])==abs(id[3])) && 
               (abs(id[1])==abs(id[6])) ) sameFlav = true;
          if ( (abs(id[1])%2==1) && (abs(id[2])%2==1) && 
               (abs(id[3])%2==1) && (abs(id[6])%2==1) ) sameType = true;
          if ( (abs(id[1])%2==0) && (abs(id[2])%2==0) && 
               (abs(id[3])%2==0) && (abs(id[6])%2==0) ) sameType = true;
          if (caseNow == 2) {
                 if (!sameFlav && !sameType) sz2duxuggdx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 6, 4, 5]
            else if (!sameFlav &&  sameType) sz2dsxsggdx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 6, 4, 5]
            else if ( sameFlav &&  sameType) sz2ddxdggdx_(&colorDepth,&icol7,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 6, 4, 5]
          } else if (caseNow == 3) {
                 if (!sameFlav && !sameType) sz2duxuggdx_(&colorDepth,&icol5 ,PNEW,HEL,&MGME2); // T[ 1, 2, 4]T[ 3, 6, 5]
            else if (!sameFlav &&  sameType) sz2dsxsggdx_(&colorDepth,&icol5 ,PNEW,HEL,&MGME2); // T[ 1, 2, 4]T[ 3, 6, 5]
            else if ( sameFlav &&  sameType) sz2ddxdggdx_(&colorDepth,&icol11,PNEW,HEL,&MGME2); // T[ 1, 2, 4]T[ 3, 6, 5]
          }
        } else if (caseNow == 4) {
          // Z -> d ubar u ubar u dbar
          // Set masses
          coupl_mass_.dmass = m[1];
          coupl_mass_.umass = m[2];
          sz2duxuuxudx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
        } else if (caseNow == 5) {
          // Z -> d dbar' d' ubar u dbar
          // Set masses
          coupl_mass_.dmass = m[1];
          coupl_mass_.smass = m[2];
          coupl_mass_.umass = m[4];
          sz2dsxsuxudx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
        } else if (caseNow == 6) {
          // Z -> d dbar d dbar d dbar
          // Set masses
          coupl_mass_.dmass = m[1];
          sz2ddxddxddx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
        } else if (caseNow == 7) {
          // Z -> d dbar' d' dbar' d' dbar
          // Set masses
          coupl_mass_.dmass = m[1];
          coupl_mass_.smass = m[2];
          sz2dsxssxsdx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
        } else if (caseNow == 8) {
          // Z -> d dbar' d' dbar'' d'' dbar
          // Set masses
          coupl_mass_.dmass = m[1];
          coupl_mass_.smass = m[2];
          coupl_mass_.umass = m[4];
          sz2dsxsbxbdx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
        }
        // Swap back if necessary
        if (abs(id[1]) % 2 == 0) swapZud();
      }

    }
    // Z decay
    //------------------------------------------------------------------------
    
    //------------------------------------------------------------------------
    // W decay
    else if (nIn == 1 && id0 == 24) {

      // Safety
      for (int ihel=0; ihel<4; ihel++) HEL[ihel] = -1;
      // Sum over W polarizations, for specific final-state polarizations
      bool sumWpol = (pol[0] == 9);
      if (sumWpol) {
        for (int i=1; i<nMom; i++)
          if (pol[i] == 9) sumWpol = false;
      }
      // Fully polarized W decay
      bool fulPolW = (!sumWpol);
      if (fulPolW) {
        for (int i=0; i<nMom; i++)
          if (pol[i] == 9) fulPolW = false;
      }
      // Average factor for incoming polarization (+-0 for W)
      nPolIn = (fulPolW ? 1 : 3);

      // Ensure ME evaluated on shell
      const double mSys = p[0].mCalc();
      coupl_mass_.wmass = mSys;
      
      // In case of W- reverse order of partons, use -id
      if (id[0] < 0) {
        map<int,int> swapMapBasic;
        swapMapBasic[0] = 0;
        for (int i=1; i<nMom; ++i) swapMapBasic[i] = nMom-i;
        if (nOut == 5 && nGluonTot == 1) {
          // Special case for W -> q qbar q g qbar
          swapMapBasic[1] = 2;
          swapMapBasic[2] = 1;
          swapMapBasic[3] = 5;
          swapMapBasic[4] = 4;
          swapMapBasic[5] = 3;
        } else if (nOut == 6 && nGluonTot == 2 && id[4] == 21 && id[5] == 21) {
          // Special case for W -> q qbar q g g qbar
          swapMapBasic[1] = 2;
          swapMapBasic[2] = 1;
          swapMapBasic[3] = 6;
          swapMapBasic[4] = 5;
          swapMapBasic[5] = 4;
          swapMapBasic[6] = 3;
        }
        vector<Vec4> tmpMom = p;
        vector<int>  tmpId  = id, tmpPol = pol;
        for (int i=0; i<nMom; i++) {
          p[i] = tmpMom[swapMapBasic[i]];
          if (tmpId[swapMapBasic[i]] == 21) id[i] = tmpId[swapMapBasic[i]];
          else id[i] = -tmpId[swapMapBasic[i]];
          if (abs(tmpPol[swapMapBasic[i]]) == 1) pol[i] = -tmpPol[swapMapBasic[i]];
          else pol[i] = tmpPol[swapMapBasic[i]];
        }
      }

      // W->jj
      if (nOut == 2 && id[1]!=21 && id[2]!=21) {
        // 0) Default assumption: fully unpolarized W decay
        // 1) Sum over W polarizations, for specific final-state polarizations
        if (sumWpol) {
          HEL[0] = (int)(1 + (pol[2] + 1)/2 + (pol[1] + 1));
          HEL[1] = (int)(5 + (pol[2] + 1)/2 + (pol[1] + 1));
          HEL[2] = (int)(9 + (pol[2] + 1)/2 + (pol[1] + 1));
        }
        // 2) Fully polarized W decay
        else if (fulPolW) {
          HEL[0] = (int)(1 + (pol[0] + 1)*4 + (pol[2] + 1)/2 + (pol[1] + 1));
        }
        // Convert momenta to MadGraph format
        double PNEW[3][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        // Set masses and couplings
        coupl_mass_.umass = m[1];
        coupl_mass_.dmass = m[2];
        setWcouplings(abs(id[1]), abs(id[2]));
        swx2udx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
      }

      // W->jjj
      else if (nOut == 3) {
        // 0) Default assumption: fully unpolarized W decay
        // 1) Sum over W polarizations, for specific final-state polarizations
        if (sumWpol) {
          HEL[0] = (int)( 1 + (pol[3] + 1)/2 + (pol[2] + 1) + 2*(pol[1] + 1));
          HEL[1] = (int)( 9 + (pol[3] + 1)/2 + (pol[2] + 1) + 2*(pol[1] + 1));
          HEL[2] = (int)(17 + (pol[3] + 1)/2 + (pol[2] + 1) + 2*(pol[1] + 1));
        }
        // 2) Fully polarized W decay
        else if (fulPolW) {
          nPolIn = 1;
          HEL[0] = (int)(1 + (pol[0] + 1)*8 + (pol[3] + 1)/2 + (pol[2] + 1) 
                         + 2*(pol[1] + 1));
        }
        // Convert momenta to MadGraph format
        double PNEW[4][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        // Set masses and couplings
        coupl_mass_.umass = m[1];
        coupl_mass_.dmass = m[3];
        setWcouplings(abs(id[1]), abs(id[3]));
        swx2ugdx_(&colorDepth,&icol,PNEW,HEL,&MGME2); // T[ 1, 3, 2]
      }

      // W->jjjj
      else if (nOut == 4) {
        // Find order, resort if 4 quarks present
        vector<int> swapMap;
        if (nGluonTot == 0) swapMap = orderWfourQuarks(id);
        else for (int i=0; i<nMom; i++) swapMap.push_back(i);
        vector<Vec4>   tmpMom = p;
        vector<int>    tmpId  = id, tmpPol = pol;
        vector<double> tmpM   = m;
        for (int i=0; i<nMom; i++) {
          p[i]   = tmpMom[swapMap[i]];
          id[i]  = tmpId[swapMap[i]];
          pol[i] = tmpPol[swapMap[i]];
          m[i]   = tmpM[swapMap[i]];
        }
        // 0) Default assumption: fully unpolarized W decay
        // 1) Sum over W polarizations, for specific final-state polarizations
        if (sumWpol) {
          HEL[0] = (int)( 1 + (pol[4] + 1)/2 + (pol[3] + 1) + 2*(pol[2] + 1) 
                         + 4*(pol[1] + 1));
          HEL[1] = (int)(17 + (pol[4] + 1)/2 + (pol[3] + 1) + 2*(pol[2] + 1) 
                         + 4*(pol[1] + 1));
          HEL[2] = (int)(33 + (pol[4] + 1)/2 + (pol[3] + 1) + 2*(pol[2] + 1) 
                         + 4*(pol[1] + 1));
        }
        // 2) Fully polarized W decay
        else if (fulPolW) {
          HEL[0] = (int)(1 + (pol[0] + 1)*16 + (pol[4] + 1)/2 + (pol[3] + 1) 
                         + 2*(pol[2] + 1) + 4*(pol[1] + 1));
        }
        // Convert momenta to MadGraph format
        double PNEW[5][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        // Set masses and couplings
        coupl_mass_.umass = m[1];
        coupl_mass_.dmass = m[4];
        setWcouplings(abs(id[1]), abs(id[4]));
        // Different color flows
        int icol1 = 1;
        int icol2 = 2;
        if (nGluonTot == 2) {
          // W -> q g g qbar
          swx2uggdx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 4, 2, 3]
        } else if (nGluonTot == 0) {
          // W -> q qbar q qbar
          if (id[1] == id[3]) {
            // W -> q qbar q qbar'
            swx2uuxudx_(&colorDepth,&icol2,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]
          } else if (id[2] == id[4]) {
            // W -> q' qbar q qbar
            swx2udxddx_(&colorDepth,&icol2,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]
          } else {
            // W -> q' qbar q qbar''
            coupl_mass_.smass = m[2];
            swx2usxsdx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]
          }
        }
      }

      // W->jjjjj
      else if (nOut == 5) {
        // Find order, resort if 4 quarks present
        vector<int> swapMap;
        bool gSwap = false;
        if (nGluonTot == 1) swapMap = orderWfourQuarks(id, gSwap);
        else for (int i=0; i<nMom; i++) swapMap.push_back(i);
        vector<Vec4>   tmpMom = p;
        vector<int>    tmpId  = id, tmpPol = pol;
        vector<double> tmpM   = m;
        for (int i=0; i<nMom; i++) {
          p[i]   = tmpMom[swapMap[i]];
          id[i]  = tmpId[swapMap[i]];
          pol[i] = tmpPol[swapMap[i]];
          m[i]   = tmpM[swapMap[i]];
        }
        // 0) Default assumption: fully unpolarized W decay
        // 1) Sum over W polarizations, for specific final-state polarizations
        if (sumWpol) {
          HEL[0] = (int)( 1 + (pol[5] + 1)/2 + (pol[4] + 1) + 2*(pol[3] + 1) 
                         + 4*(pol[2] + 1) + 8*(pol[1] + 1));
          HEL[1] = (int)(33 + (pol[5] + 1)/2 + (pol[4] + 1) + 2*(pol[3] + 1) 
                         + 4*(pol[2] + 1) + 8*(pol[1] + 1));
          HEL[2] = (int)(65 + (pol[5] + 1)/2 + (pol[4] + 1) + 2*(pol[3] + 1) 
                         + 4*(pol[2] + 1) + 8*(pol[1] + 1));
        }
        // 2) Fully polarized W decay
        else if (fulPolW) {
          HEL[0] = (int)(1 + (pol[0] + 1)*32 + (pol[5] + 1)/2 + (pol[4] + 1) 
                         + 2*(pol[3] + 1) + 4*(pol[2] + 1) + 8*(pol[1] + 1));
        }
        // Convert momenta to MadGraph format
        double PNEW[6][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        // Set masses and couplings
        coupl_mass_.umass = m[1];
        coupl_mass_.dmass = m[5];
        setWcouplings(abs(id[1]), abs(id[5]));
        // Different color flows
        int icol1 = 1;
        int icol2 = 2;
        int icol3 = 3;
        int icol4 = 4;
        if (nGluonTot == 3) {
          // W -> q g g g qbar
          swx2ugggdx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 5, 2, 3, 4]
        } else if (nGluonTot == 1) {
          // W -> q qbar q g qbar
          if (id[1] == id[3]) {
            // W -> q qbar q g qbar'
            if (gSwap) swx2uuxugdx_(&colorDepth,&icol4,PNEW,HEL,&MGME2); // T[ 3, 5]T[ 1, 2, 4]
            else       swx2uuxugdx_(&colorDepth,&icol3,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 5, 4]
          } else if (id[2] == id[5]) {
            // W -> q' qbar q g qbar
            if (gSwap) swx2udxdgdx_(&colorDepth,&icol4,PNEW,HEL,&MGME2); // T[ 3, 5]T[ 1, 2, 4]
            else       swx2udxdgdx_(&colorDepth,&icol3,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 5, 4]
          } else {
            // W -> q' qbar q g qbar''
            coupl_mass_.smass = m[2];
            if (gSwap) swx2usxsgdx_(&colorDepth,&icol2,PNEW,HEL,&MGME2); // T[ 3, 5]T[ 1, 2, 4]
            else       swx2usxsgdx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 5, 4]
          }
        }
      }

      // W->jjjjjj
      else if (nOut == 6) {
        // Find order, resort if 4 or 6 quarks present
        vector<int> swapMap;
        bool gSwap = false, gg = false, flav2ordChange = false;
        map<int,int> flavours;
        if (nGluonTot == 2) swapMap = orderWfourQuarks(id, gSwap, gg);
        else if (nGluonTot == 0) swapMap = orderWsixQuarks(id, flavours,
          flav2ordChange);
        else for (int i=0; i<nMom; i++) swapMap.push_back(i);
        vector<Vec4>   tmpMom = p;
        vector<int>    tmpId  = id, tmpPol = pol;
        vector<double> tmpM   = m;
        for (int i=0; i<nMom; i++) {
          p[i]   = tmpMom[swapMap[i]];
          id[i]  = tmpId[swapMap[i]];
          pol[i] = tmpPol[swapMap[i]];
          m[i]   = tmpM[swapMap[i]];
        }
        // Get absolute values of ids
        vector<int> idAbs = id;
        for (int i=0; i<(int)idAbs.size(); i++) idAbs[i] = abs(idAbs[i]);
        // 0) Default assumption: fully unpolarized W decay
        // 1) Sum over W polarizations, for specific final-state polarizations
        if (sumWpol) {
          HEL[0] = (int)(1 + (pol[6] + 1)/2 + (pol[5] + 1) + 2*(pol[4] + 1) 
                         + 4*(pol[3] + 1) + 8*(pol[2] + 1) + 16*(pol[1] + 1));
          HEL[1] = (int)(65 + (pol[6] + 1)/2 + (pol[5] + 1) + 2*(pol[4] + 1) 
                         + 4*(pol[3] + 1) + 8*(pol[2] + 1) + 16*(pol[1] + 1));
          HEL[2] = (int)(129 + (pol[6] + 1)/2 + (pol[5] + 1) + 2*(pol[4] + 1) 
                         + 4*(pol[3] + 1) + 8*(pol[2] + 1) + 16*(pol[1] + 1));
        }
        // 2) Fully polarized W decay
        else if (fulPolW) {
          HEL[0] = (int)(1 + (pol[0] + 1)*64 + (pol[6] + 1)/2 + (pol[5] + 1) 
                         + 2*(pol[4] + 1) + 4*(pol[3] + 1) + 8*(pol[2] + 1) 
                         + 16*(pol[1] + 1));
        }
        // Convert momenta to MadGraph format
        double PNEW[7][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        // Set masses and couplings
        coupl_mass_.umass = m[1];
        coupl_mass_.dmass = m[6];
        setWcouplings(idAbs[1], idAbs[6]);
        // Different color flows
        int icol1  = 1;
        int icol2  = 2;
        int icol3  = 3;
        int icol4  = 4;
        int icol6  = 6;
        int icol7  = 7;
        int icol10 = 10;
        int icol12 = 12;
        if (nGluonTot == 4) {
          // W -> q g g g g qbar
          swx2uggggdx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 6, 2, 3, 4, 5]
        } else if (nGluonTot == 2) {
          // W -> q qbar q g qbar
          if (id[1] == id[3]) {
            // W -> q qbar q g g qbar'
            if (gg) {
              if (gSwap) swx2uuxuggdx_(&colorDepth,&icol12,PNEW,HEL,&MGME2); // T[ 3, 6]T[ 1, 2, 4, 5]
              else       swx2uuxuggdx_(&colorDepth,&icol7 ,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 6, 4, 5]
            } else {
              swx2uuxuggdx_(&colorDepth,&icol10,PNEW,HEL,&MGME2); // T[ 1, 2, 4]T[ 3, 6, 5]
            }
          } else if (id[2] == id[6]) {
            // W -> q' qbar q g g qbar
            if (gg) {
              if (gSwap) swx2udxdggdx_(&colorDepth,&icol12,PNEW,HEL,&MGME2); // T[ 3, 6]T[ 1, 2, 4, 5]
              else       swx2udxdggdx_(&colorDepth,&icol7 ,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 6, 4, 5]
            } else {
              swx2udxdggdx_(&colorDepth,&icol10,PNEW,HEL,&MGME2); // T[ 1, 2, 4]T[ 3, 6, 5]
            }
          } else {
            // W -> q' qbar q g g qbar''
            coupl_mass_.smass = m[2];
            if (gg) {
              if (gSwap) swx2usxsggdx_(&colorDepth,&icol6,PNEW,HEL,&MGME2); // T[ 3, 6]T[ 1, 2, 4, 5]
              else       swx2usxsggdx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 6, 4, 5]
            } else {
              swx2usxsggdx_(&colorDepth,&icol4,PNEW,HEL,&MGME2); // T[ 1, 2, 4]T[ 3, 6, 5]
            }
          }
        } else if (nGluonTot == 0) {
          // W -> q qbar q qbar q qbar
          int flavCount = 0;
          for (int i=1; i<=5; ++i) if (flavours[i] > 0) flavCount++;
          if (flavCount == 4) {
            // W -> q'' qbar q qbar' q' qbar'''
            coupl_mass_.smass = m[2];
            coupl_mass_.cmass = m[4];
            swx2usxscxcdx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
          } else if (flavCount == 3) {
            if (flavours[idAbs[1]] > 1) {
              if (idAbs[1] == idAbs[2] && idAbs[1] == idAbs[3]) {
                // W -> q qbar q qbar' q' qbar''
                coupl_mass_.smass = m[4];
                swx2uuxusxsdx_(&colorDepth,&icol3,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
              } else if (idAbs[1] == idAbs[3] && idAbs[1] == idAbs[4]) {
                // W -> q qbar' q qbar q' qbar''
                // Swap 12 with 34 -> q qbar q qbar' q' qbar''
                double PTMP[7][4];
                for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PTMP[i]);
                for (int j=0; j<4; ++j) {
                  PNEW[1][j] = PTMP[3][j];
                  PNEW[2][j] = PTMP[4][j];
                  PNEW[3][j] = PTMP[1][j];
                  PNEW[4][j] = PTMP[2][j];
                }
                coupl_mass_.umass = m[3];
                coupl_mass_.smass = m[2];
                setWcouplings(idAbs[3], idAbs[6]);
                swx2uuxusxsdx_(&colorDepth,&icol3,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
              } else if (idAbs[1] == idAbs[4] && idAbs[1] == idAbs[5]) {
                // W -> q qbar' q' qbar q qbar''
                // Swap 23 with 45 -> q qbar q qbar' q' qbar''
                double PTMP[7][4];
                for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PTMP[i]);
                for (int j=0; j<4; ++j) {
                  PNEW[2][j] = PTMP[4][j];
                  PNEW[3][j] = PTMP[5][j];
                  PNEW[4][j] = PTMP[2][j];
                  PNEW[5][j] = PTMP[3][j];
                }
                coupl_mass_.smass = m[2];
                swx2uuxusxsdx_(&colorDepth,&icol4,PNEW,HEL,&MGME2); // T[ 1, 4]T[ 5, 2]T[ 3, 6]
              } else if (verbose >= 3)
                printOut("MG4interface::ME","sorting problem in W -> 6q with 3 "
                "different flavours and #flav[1] > 1)");
            } else if (flavours[idAbs[6]] > 1) {
              if (idAbs[6] == idAbs[2] && idAbs[6] == idAbs[3]) {
                // W -> q'' qbar q qbar' q' qbar
                coupl_mass_.smass = m[4];
                swx2udxdsxsdx_(&colorDepth,&icol4,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
              } else if (idAbs[6] == idAbs[3] && idAbs[6] == idAbs[4]) {
                // W -> q'' qbar' q qbar q' qbar
                // Swap 2 with 4 -> q'' qbar q qbar' q' qbar
                double PTMP[7][4];
                for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PTMP[i]);
                for (int j=0; j<4; ++j) {
                  PNEW[2][j] = PTMP[4][j];
                  PNEW[4][j] = PTMP[2][j];
                }
                coupl_mass_.smass = m[2];
                swx2udxdsxsdx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 4]T[ 3, 2]T[ 5, 6]
              }
              else if (idAbs[6] == idAbs[4] && idAbs[6] == idAbs[5]) {
                // W -> q'' qbar' q' qbar q qbar
                // Swap 23 with 45 -> q'' qbar q qbar' q' qbar
                double PTMP[7][4];
                for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PTMP[i]);
                for (int j=0; j<4; ++j) {
                  PNEW[2][j] = PTMP[4][j];
                  PNEW[3][j] = PTMP[5][j];
                  PNEW[4][j] = PTMP[2][j];
                  PNEW[5][j] = PTMP[3][j];
                }
                coupl_mass_.smass = m[2];
                swx2udxdsxsdx_(&colorDepth,&icol2,PNEW,HEL,&MGME2); // T[ 1, 4]T[ 5, 2]T[ 3, 6]
              } else if (verbose >= 3)
                printOut("MG4interface::ME","sorting problem in W -> 6q with 3 "
                  "different flavours and #flav[6] > 1)");
            } else if (idAbs[2] == idAbs[3] && idAbs[2] == idAbs[4] &&
              idAbs[2] == idAbs[5]) {
              // W -> q' qbar q qbar q qbar''
              coupl_mass_.smass = m[4];
              swx2usxssxsdx_(&colorDepth,&icol2,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
            } else if (verbose >= 3)
              printOut("MG4interface::ME","sorting problem in W -> 6q with 3 "
                "different flavours");
          } else {
            if (idAbs[1] == idAbs[2] && idAbs[1] == idAbs[3] &&
              idAbs[1] == idAbs[4] && idAbs[1] == idAbs[5]) {
              // W -> q qbar q qbar q qbar'
              swx2uuxuuxudx_(&colorDepth,&icol6,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
            } else if (idAbs[6] == idAbs[2] && idAbs[6] == idAbs[3] &&
              idAbs[6] == idAbs[4] && idAbs[6] == idAbs[5]) {
              // W -> q' qbar q qbar q qbar
              swx2udxddxddx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
            } else if (idAbs[1] == idAbs[2] && idAbs[1] == idAbs[3]) {
              // q qbar q qbar' q' qbar'
              if (flav2ordChange)
                swx2uuxudxddx_(&colorDepth,&icol3,PNEW,HEL,&MGME2); // T[ 1, 4]T[ 5, 2]T[ 3, 6]
              else
                swx2uuxudxddx_(&colorDepth,&icol6,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
            } else if (verbose >= 3)
              printOut("MG4interface::ME","sorting problem in W -> 6q with 2 "
                "different flavours");
          }
        }
      }

    }
    // W decay
    //------------------------------------------------------------------------

    //------------------------------------------------------------------------
    // H decay
    // NOTE If the quarks are massless, the pre-calculated Yukawa couplings
    // are used. If they are massive, the coupling gets re-calculated
    else if (nIn == 1 && id0 == 25) {

      // Safety
      for (int ihel=0; ihel<4; ihel++) HEL[ihel] = -1;
      // Fully polarized H decay
      bool fulPolH = true;
      for (int i=1; i<nMom; i++)
        if (pol[i] == 9) fulPolH = false;
      // Average factor for incoming polarization (- for H)
      nPolIn = 1;

      // Ensure ME evaluated on shell
      const double mSys = p[0].mCalc();
      coupl_mass_.hmass = mSys;

      // H->jj
      if (nOut == 2) {
        // 0) Default assumption: fully unpolarized H decay
        // 1) Fully polarized H decay
        if (fulPolH) {
          HEL[0] = (int)(1 + (pol[2] + 1)/2 + (pol[1] + 1));
        }
        // Convert momenta to MadGraph format
        double PNEW[3][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        if (nGluonTot == 2) {
          // H -> g g
          sh2gg_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        } else {
          // H -> q qbar
          // Set masses and couplings
          setYukawa(abs(id[1]), m[1]);
          sh2bbx_(&colorDepth,&icol,PNEW,HEL,&MGME2);
        }
      }

      // H->jjj
      else if (nOut == 3) {
        // 0) Default assumption: fully unpolarized H decay
        // 1) Fully polarized H decay
        if (fulPolH) {
          HEL[0] = (int)(1 + (pol[3] + 1)/2 + (pol[2] + 1) 
                         + 2*(pol[1] + 1));
        }
        // Convert momenta to MadGraph format
        double PNEW[4][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        if (nGluonTot == 3) {
          // H -> g g g
          sh2ggg_(&colorDepth,&icol,PNEW,HEL,&MGME2); // F[1, 2, 3]
        } else {
          // H -> q g qbar
          // Set masses and couplings
          setYukawa(abs(id[1]), m[1]);
          sh2bgbx_(&colorDepth,&icol,PNEW,HEL,&MGME2); // T[1, 3, 2]
        }
      }

      // H->jjjj
      else if (nOut == 4) {
        // 0) Default assumption: fully unpolarized H decay
        // 1) Fully polarized H decay
        if (fulPolH) {
          HEL[0] = (int)(1 + (pol[4] + 1)/2 + (pol[3] + 1) 
                         + 2*(pol[2] + 1) + 4*(pol[1] + 1));
        }
        // Convert momenta to MadGraph format
        double PNEW[5][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        // Different color flows
        int icol1 = 1;
        int icol2 = 2;
        if (nGluonTot == 4) {
          // H -> g g g g
          sh2gggg_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // F[1, 2, 3, 4]
        } else if (nGluonTot == 2) {
          // H -> q g g qbar
          // Set masses and couplings
          setYukawa(abs(id[1]), m[1]);
          sh2bggbx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[1, 4, 2, 3]
        } else if (nGluonTot == 0) {
          // H -> q qbar' q' qbar
          // Check if all quarks have the same flavour
          bool sameFlav = false;
          if ( (abs(id[1])==abs(id[2])) && (abs(id[1])==abs(id[3])) &&
               (abs(id[1])==abs(id[4])) ) sameFlav = true;
          // Set masses and couplings
          setYukawa(abs(id[1]), abs(id[2]), m[1], m[2]);
          if (sameFlav) sh2bbxbbx_(&colorDepth,&icol2,PNEW,HEL,&MGME2); // T[ 3, 4]T[ 1, 2]
          else          sh2bcxcbx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 3, 4]T[ 1, 2]
        }
      }

      // H->jjjjj
      else if (nOut == 5) {
        // 0) Default assumption: fully unpolarized H decay
        // 1) Fully polarized H decay
        if (fulPolH) {
          HEL[0] = (int)(1 + (pol[5] + 1)/2 + (pol[4] + 1) 
                         + 2*(pol[3] + 1) + 4*(pol[2] + 1) + 8*(pol[1] + 1));
        }
        // Convert momenta to MadGraph format
        double PNEW[6][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        if (nGluonTot == 5) {
          // H -> g g g g g
          sh2ggggg_(&colorDepth,&icol,PNEW,HEL,&MGME2); // F[1, 2, 3, 4, 5]
        } else if (nGluonTot == 3) {
          // H -> q g g g qbar
          // Set masses and couplings
          setYukawa(abs(id[1]), m[1]);
          sh2bgggbx_(&colorDepth,&icol,PNEW,HEL,&MGME2); // T[ 1, 5, 2, 3, 4]
        } else if (nGluonTot == 1) {
          // H -> q qbar' q' g qbar
          // Check if all quarks have the same flavour
          bool sameFlav = false;
          if ( (abs(id[1])==abs(id[2])) && (abs(id[1])==abs(id[3])) && 
               (abs(id[1])==abs(id[5])) ) sameFlav = true;
          // Set masses and couplings
          setYukawa(abs(id[1]), abs(id[2]), m[1], m[2]);
          if (sameFlav) sh2bbxbgbx_(&colorDepth,&icol,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 5, 4]
          else          sh2bcxcgbx_(&colorDepth,&icol,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 5, 4]
        }
      }

      // H->jjjjjj
      else if (nOut == 6) {
        // Find the right order of the partons
        map<int,int> mapVtoMG;
        // Default: order as it is
        for (int i=0; i<nMom; ++i) mapVtoMG[i] = i;
        // Note the different cases
        int caseNow = 0;
        // g g g g g g
        if (nGluonTot == 6) caseNow = 0;
        // q g g g g qbar
        else if (nGluonTot == 4) caseNow = 1;
        else if (nGluonTot == 2) {
          // q qbar' q' g g qbar
          if ( (id[4]==21) && (id[5]==21) ) caseNow = 2;
          // q g1 qbar' q' g2 qbar --- sort to --- q qbar' q' g1 g2 qbar
          else if ( (id[2]==21) && (id[5]==21) ) {
            caseNow = 3;
            mapVtoMG[2] = 3;
            mapVtoMG[3] = 4;
            mapVtoMG[4] = 2;
          }
        } else if (nGluonTot == 0) {
          map<int,int> flavours;
          for (int i=1; i<=5; ++i) flavours[i] = 0;
          for (int i=1; i<nMom; ++i) flavours[abs(id[i])]++;
          // 0 if 0 or 1 quark pair, 1 if 2 quark pairs, 2 if 3 quark pais
          int countSameFlavour = 0;
          for (int i=1; i<=5; ++i)
            countSameFlavour = max(max((flavours[i]-2)/2,0),countSameFlavour);
          // q qbar q qbar q qbar
          if (countSameFlavour == 2) caseNow = 4;
          // q' qbar q qbar q qbar'
          else if (countSameFlavour == 1) {
            caseNow = 5;
            int absFlav = 0; // flavour that appears once
            for (int i=1; i<=5; ++i) if (flavours[i] == 2) absFlav = i;
            pair<int,int> flavourPos; // quark first, antiquark second
            for (int i=1; i<nMom; ++i) if (abs(id[i]) == absFlav) {
              if (id[i] > 0) flavourPos.first = i;
              else flavourPos.second = i;
            }
            vector<int> order = orderQQBqqbqqb(flavourPos);
            for (int i=0; i<nMom; ++i) mapVtoMG[i] = order[i];
          }
          // q'' qbar' q' qbar q qbar'''
          else if (countSameFlavour == 0) {
            caseNow = 6;
            // pick one quark pair at ramdom for order, use first quark
            pair<int,int> randomPos; // quark first, antiquark second
            randomPos.first = 1;
            for (int i=2; i<nMom; ++i) if (abs(id[i]) == abs(id[1]))
              randomPos.second = i;
            vector<int> order = orderQQBqqbqqb(randomPos);
            for (int i=0; i<nMom; ++i) mapVtoMG[i] = order[i];
          }
        }
        vector<Vec4>   tmpMom = p;
        vector<int>    tmpId  = id, tmpPol = pol;
        vector<double> tmpM   = m;
        for (int i=0; i<nMom; i++) {
          p[i]   = tmpMom[mapVtoMG[i]];
          id[i]  = tmpId[mapVtoMG[i]];
          pol[i] = tmpPol[mapVtoMG[i]];
          m[i]   = tmpM[mapVtoMG[i]];
        }
        // 0) Default assumption: fully unpolarized H decay
        // 1) Fully polarized H decay
        if (pol[1] != 9 && pol[2] != 9 && pol[3] != 9 
                        && pol[4] != 9 && pol[5] != 9 && pol[6] != 9) {
          HEL[0] = (int)(1 + (pol[6] + 1)/2 + (pol[5] + 1) 
                         + 2*(pol[4] + 1) + 4*(pol[3] + 1) + 8*(pol[2] + 1) 
                         + 16*(pol[1] + 1));
        }
        // Convert momenta to MadGraph format
        double PNEW[7][4];
        for (int i=0; i<nMom; ++i) Vec4toMG(p[i],PNEW[i]);
        // Different color flows
        int icol1 = 1;
        int icol5  = 5;
        int icol7  = 7;
        int icol9  = 9;
        int icol10 = 10;
        if (caseNow == 0) {
          // H -> g g g g g g
          sh2gggggg_(&colorDepth,&icol9,PNEW,HEL,&MGME2); // F[1, 2, 3, 4, 5, 6]
        } else if (caseNow == 1) {
          // H -> q g g g g qbar
          // Set masses and couplings
          setYukawa(abs(id[1]), m[1]);
          sh2bggggbx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 6, 2, 3, 4, 5]
        } else if (caseNow == 2 || caseNow == 3) {
          // H -> q qbar' q' g g qbar
          // Check if all quarks have the same flavour
          bool sameFlav = false;
          if ( (abs(id[1])==abs(id[2])) && (abs(id[1])==abs(id[3])) &&
               (abs(id[1])==abs(id[6])) ) sameFlav = true;
          // Set masses and couplings
          setYukawa(abs(id[1]), abs(id[2]), m[1], m[2]);
          if (caseNow == 2) {
            if (sameFlav) sh2bbxbggbx_(&colorDepth,&icol7,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 6, 4, 5]
            else          sh2bcxcggbx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 6, 4, 5]
          } else if (caseNow == 3) { // q g qbar' q' g qbar ( was this )
            if (sameFlav) sh2bbxbggbx_(&colorDepth,&icol10,PNEW,HEL,&MGME2); // T[ 1, 2, 4]T[ 3, 6, 5]
            else          sh2bcxcggbx_(&colorDepth,&icol5 ,PNEW,HEL,&MGME2); // T[ 1, 2, 4]T[ 3, 6, 5]
          }
        }
        else if (caseNow == 4) {
          // H -> q qbar q qbar q qbar
          // Set masses and couplings
          setYukawa(abs(id[1]), 0, m[1], 0.0);
          sh2bbxbbxbbx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
        } else if (caseNow == 5) {
          // H -> q' qbar q qbar q qbar'
          // Set masses and couplings
          setYukawa(abs(id[1]), abs(id[2]), m[1], m[2]);
          sh2bcxccxcbx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
        } else if (caseNow == 6) {
          // H -> q'' qbar' q' qbar q qbar'''
          // Set masses and couplings
          setYukawa(abs(id[1]), abs(id[2]), abs(id[4]), m[1], m[2], m[4]);
          sh2bcxcsxsbx_(&colorDepth,&icol1,PNEW,HEL,&MGME2); // T[ 1, 2]T[ 3, 4]T[ 5, 6]
        }
      }

    }
    // H decay
    //------------------------------------------------------------------------

    if (std::isnan(MGME2) && verbose >= 1) {
      printOut("MG4interface::ME2","Error! ME2 = nan. Printing momenta:");
      for (int i=0; i<int(p.size()); ++i) {
        cout<<"   p"<<i<<p[i];
      }
    }

    // Return |ME|^2 value (or -1 if no ME found)
    if (abs(MGME2+1) < TINY) return -1.;
    else return MGME2/max(1,nPolIn);

  }

}
